/* Tables used by TradingService Attribute Database  */

CREATE TABLE OF_Attributes
(
   Name VARCHAR (36) NOT NULL, /* name of the attribute */
   PID INTEGER NOT NULL, /* process id of this trader */
   Object INTEGER NOT NULL, /* value of the attribute */
   PRIMARY KEY (Name, PID)
)
go

/* Tables used by TradingService Service Database */

CREATE TABLE OF_Services
(
   Incarnation NUMERIC NOT NULL, /* consecutive integer id number */
   PID INTEGER NOT NULL, /* process id of this trader */
   ServiceTypeIndex INTEGER NOT NULL, /* index of service type name */
   Masked INTEGER NOT NULL, /* boolean value, whether service type is masked */
   InterfaceTypeIndex INTEGER NOT NULL, /* service interface name index */
   PRIMARY KEY (Incarnation, PID)
)
go

CREATE TABLE OF_ServiceProperties
(
   Incarnation NUMERIC NOT NULL, /* consecutive integer id number */
   PID INTEGER NOT NULL, /* process id of this trader */
   NameIndex INTEGER NOT NULL, /* property name index */
   PMode INTEGER NOT NULL, /* propertymode as an integer */
   TypeCode IMAGE NOT NULL /* serialized typecode */
)
go

CREATE TABLE OF_SuperTypes
(
   Incarnation NUMERIC NOT NULL, /* consecutive integer id number */
   PID INTEGER NOT NULL, /* process id of this trader */
   ServiceType NUMERIC NOT NULL  /* incarnation number of a supertype */
)
go

/* Tables used by TradingService Offer Database */

CREATE TABLE OF_Offers
(
   OfferID VARCHAR (36), /* offer id (UUID) */
   PID INTEGER, /* process id of this trader */
   IsProxy INTEGER, /* boolean, true if offer is a proxy, false otherwise */
   ServiceTypeIndex INTEGER, /* service type name index */
   OfferData IMAGE NOT NULL, /* object reference as a string */
   MatchAll INTEGER NOT NULL, /*boolean, true if IF_MATCH_ALL is true */
   PRIMARY KEY (OfferID)
)
go

CREATE TABLE OF_OfferProps
(
   Name VARCHAR(50),  /* property name */
   Value VARCHAR(200),        /* property value */
   ServiceTypeIndex INTEGER NOT NULL,  /* service type name index */
   PID INTEGER NOT NULL,  /* process id of this trader */
   NumVal NUMERIC, /* numeric property value */
   OfferID VARCHAR(36) NOT NULL,  /* offer id ( UUID ) */
   ComplexValue INTEGER NOT NULL /* boolean, true if property value is dynamically evaluated, or is not a primative */
)
go

CREATE PROCEDURE of_insert_attribute
   @nm VARCHAR(36),
   @pd INTEGER,
   @obj INTEGER
AS
BEGIN
   INSERT INTO OF_Attributes
      (Name, PID, Object)
      VALUES
      (@nm, @pd, @obj)
   RETURN 0
END
go

CREATE PROCEDURE of_update_attribute
   @nm VARCHAR(36),
   @pd INTEGER,
   @obj INTEGER
AS
BEGIN
   UPDATE OF_Attributes
      SET Object = @obj
      WHERE Name = @nm AND PID = @pd
   RETURN 0
END
go

CREATE PROCEDURE of_restore_attribute
   @nm VARCHAR(36),
   @pd INTEGER,
   @obj INTEGER output
AS
BEGIN
   SELECT @obj = Object FROM OF_Attributes
      WHERE Name = @nm AND PID = @pd
   IF @@rowcount = 0
   BEGIN
      SELECT @obj = 0
   END
   RETURN 0
END
go

CREATE PROCEDURE of_attribute_exists
   @nm VARCHAR(36),
   @pd INTEGER,
   @obj INTEGER output
AS
BEGIN
   SELECT @obj = (SELECT Object FROM OF_Attributes
      WHERE Name = @nm AND PID = @pd)
   RETURN 0
END
go

CREATE PROCEDURE of_servicetype
   @inc INTEGER,
   @pd INTEGER,
   @idx INTEGER output
AS
BEGIN
   SELECT @idx = (SELECT ServiceTypeIndex FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd)
   RETURN 0
END
go

CREATE PROCEDURE of_incarnation
   @inx INTEGER,
   @pd INTEGER,
   @inc INTEGER output
AS
BEGIN
   SELECT @inc = Incarnation FROM OF_Services
      WHERE ServiceTypeIndex = @inx AND PID = @pd
   IF @@rowcount = 0
   BEGIN
      SELECT @inc = -1
   END
   RETURN 0
END
go

CREATE PROCEDURE of_max_incarnation
   @pd INTEGER,
   @inc INTEGER output
AS
BEGIN
   SELECT @inc = (SELECT MAX (Incarnation) FROM OF_Services
      WHERE  PID = @pd)
   IF @inc = NULL
   BEGIN
      SELECT @inc = 0
   END
   RETURN 0
END
go

CREATE PROCEDURE of_describe_services
   @inc INTEGER,
   @pd INTEGER,
   @msk INTEGER output,
   @int INTEGER output
AS
BEGIN
   SELECT @msk = (SELECT Masked
      FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd)
   SELECT @int = (SELECT InterfaceTypeIndex
      FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd)
   RETURN 0
END
go


CREATE PROCEDURE of_update_mask
   @inc INTEGER,
   @pd INTEGER,
   @mk INTEGER,
   @found INTEGER output
AS
BEGIN
   SELECT @found = (SELECT COUNT (*) FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd)
   IF @found > 0
   BEGIN
      UPDATE OF_Services
         SET Masked = @mk
         WHERE Incarnation = @inc AND PID = @pd
   END
   RETURN 0
END
go

CREATE PROCEDURE of_insert_services
   @inc INTEGER,
   @pd INTEGER,
   @idx INTEGER,
   @mk INTEGER,
   @int INTEGER
AS
BEGIN
   INSERT INTO OF_Services
      (Incarnation, PID, ServiceTypeIndex, Masked, InterfaceTypeIndex)
      VALUES (@inc, @pd, @idx, @mk, @int)
   RETURN 0
END
go

CREATE PROCEDURE of_insert_supertypes
   @inc INTEGER,
   @pd INTEGER,
   @st INTEGER
AS
BEGIN
   INSERT INTO OF_SuperTypes
      (Incarnation, PID, ServiceType)
      VALUES (@inc, @pd, @st)
   RETURN 0
END
go

CREATE PROCEDURE of_delete_services
   @inc INTEGER,
   @pd INTEGER
AS
BEGIN
   DELETE FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd
   RETURN 0
END
go

CREATE PROCEDURE of_delete_service_all
  @inc INTEGER,
  @pd INTEGER
AS
BEGIN
   DELETE FROM OF_Services
      WHERE Incarnation = @inc AND PID = @pd
   DELETE FROM OF_ServiceProperties
      WHERE Incarnation = @inc AND PID = @pd
   DELETE FROM OF_SuperTypes
      WHERE Incarnation = @inc AND PID = @pd
   RETURN 0
END
go

CREATE PROCEDURE of_delete_sproperties
   @inc INTEGER,
   @pd INTEGER
AS
BEGIN
   DELETE FROM OF_ServiceProperties
      WHERE Incarnation = @inc AND PID = @pd
   RETURN 0
END
go


CREATE PROCEDURE of_delete_supertypes
   @inc INTEGER,
   @pd INTEGER
AS
BEGIN
   DELETE FROM OF_SuperTypes WHERE Incarnation = @inc AND PID = @pd
   RETURN 0
END
go


CREATE PROCEDURE of_delete_offer
   @oid VARCHAR(36),
   @pd INTEGER,
   @px INTEGER,
   @res INTEGER output
AS
BEGIN
   DECLARE @kn INTEGER
   DECLARE @isp INTEGER
   SELECT @res = 0
   SELECT @kn = (SELECT COUNT (*)
      FROM OF_Offers
      WHERE OfferID = @oid AND PID = @pd)
   IF @kn = 0
   BEGIN
      SELECT @res = 2
      RETURN 0
   END
   SELECT @isp = (SELECT IsProxy
      FROM OF_Offers
      WHERE OfferID = @oid AND PID = @pd)
   IF @isp != @px
   BEGIN
      SELECT @res = 1
      RETURN 0
   END
   DELETE FROM OF_Offers
      WHERE OfferID = @oid AND PID = @pd
   RETURN 0
END
go


CREATE INDEX of_offerPropsIndexValue ON OF_OfferProps(Value, Name, ServiceTypeIndex, PID)
go

CREATE INDEX of_offerPropsIndexNumValue ON OF_OfferProps(NumVal, Name, ServiceTypeIndex, PID)
go

CREATE INDEX of_offerPropsIndexComplex ON OF_OfferProps(Name, ComplexValue)
go

CREATE INDEX of_offersIndex ON OF_Offers (ServiceTypeIndex, PID, MatchAll)
go
