/*************************************************************************
 *                           - COPYRIGHT NOTICE -                        *
 *                                                                       *
 * The information contained herein and which follows is proprietary     *
 * information and is the property of PrismTech Corporation.  This       *
 * information is not to be divulged, released, copied, reproduced, or   *
 * conveyed to unauthorized personnel, companies, or other institutions  *
 * without the direct and expressed approval in writing of PrismTech     *
 * Corporation.                                                          *
 *                                                                       *
 *                                                                       *
 *                           - DISCLAIMER -                              *
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech")     *
 * are under any obligation to support the use of this code which is     *
 * provided without warranty of any kind. PrismTech shall (except as     *
 * otherwise prohibited by law) have no liability whatsoever in relation *
 * to its use.                                                           *
 *                                                                       *
 ************************************************************************/

/**
 * @author PrismTech
 * @version Version 1.0
 */

package com.prismt.cos.CosNotification.examples.Benchmark;

import java.util.*;
import java.io.*;

import org.omg.CosNotification.*;
import org.omg.CosNotifyComm.*;
import org.omg.CosNotifyChannelAdmin.*;
import org.omg.CosNotifyFilter.*;

import com.prismt.orb.ObjectAdapter;
import com.prismt.orb.util.AnyUtil;


/**
 * The basic functionality that is common to both suppliers and consumers.
 */
public abstract class BenchmarkBaseClient implements java.io.Serializable
{
   /**
    * Indicates the nature of the client (supplier vs consumer) that extends
    * this class.
    */
   protected boolean isSupplier = true;

   /**
    * Indicates the nature of the client (structured vs sequence vs untyped)
    * that extends this class.
    */
   protected ClientType clientType = ClientType.STRUCTURED_EVENT;

   /** Timing related information. */
   protected long startTime = -1;
   protected long currentTime;

   /** The time at which the last report was generated. */
   protected long lastReportTime = 0;

   /** The total time spent in deliberate delays. */
   protected long totalDelayDuration = 0;

   /** The deliberate delays up to the time when the last report was generated. */
   protected long delayAtLastReport = 0;

   /** The CORBA object for the client. */
   protected org.omg.CORBA.Object thisRef = null;

   /** The QoS properties to set on the proxy. */
   protected Property[] m_qos = null;

   /** The filters to add to the proxy. */
   protected Vector m_filters = null;

   /** Notification Service references. */
   protected org.omg.CORBA.ORB orb = null;
   protected EventChannelFactory factory = null;
   protected EventChannel channel = null;
   protected SupplierAdmin suppAdmin = null;
   protected ConsumerAdmin consAdmin = null;

   /** Notification Service resolve name. */
   protected String resolveName = null;
   protected String resolveNameDefault = "NotificationService";

   /** Channel ID. */
   protected int channelId = -1;
   protected int channelIdDefault = 0;

   /** Event Persistence.  There is no actual default. */
   protected boolean persistence = false;
   protected boolean persistenceSpecified = false;

   /** Number of events between reports: 0 to disable reporting. */
   protected int reportThreshold = 0;
   protected int reportThresholdDefault = 0;

   /** AutoSequence QoS. */
   protected int autoSequenceBatchSize = 0;
   protected int autoSequenceBatchSizeDefault = 200;
   protected long autoSequenceTimeout = 0;
   protected long autoSequenceTimeoutDefault = 200;

   /**
    * The batch size.  This is the number of events that the sequence supplier
    * will send in each array of events, and the number of events that the
    * sequence consumer specifies via the MaximumBatchSize QoS property.
    */
   protected int batchSize = 0;
   protected int batchSizeDefault = 10;

   /** The pacing interval for the sequence consumer's proxy. */
   protected long pacingInterval = (long) 0;
   protected long pacingIntervalDefault = (long) 0;

   /** The number of events at which the last report was issued. */
   protected int lastReportEventCount = 0;

   /**
    * File for report information.  The default is only used to detect if the
    * parameter wasn't specified.
    */
   protected String reportFilename = "";
   protected String reportFilenameDefault = "<NOFILE>";
   protected FileWriter fw = null;
   protected BufferedWriter bw = null;

   /** Quick reference boolean for file output. */
   protected boolean reportToFile = false;

   /** Number of events to send: 0 for unlimited. */
   protected int eventLimit = 0;
   protected int eventLimitDefault = 10000;

   /** Number of events between delays: 0 to disable delays. */
   protected int eventDelayThreshold = 0;
   protected int eventDelayThresholdDefault = 0;

   /** Length of time (in ms) to wait during the above delays: 0 to disable. */
   protected int eventDelay = 0;
   protected int eventDelayDefault = 0;

   /** Boolean calculated from the above for easy reference. */
   protected boolean performDelays = false;

   /** For consumer, whether or not to check for sequential event IDs. */
   protected boolean checkOrdering = false;
   protected boolean checkOrderingDefault = false;

   /** For consumer, whether or not to display the event details. */
   protected boolean displayReceivedEvents = false;
   protected boolean displayReceivedEventsDefault = false;

   /** Number of rejected events. */
   protected int numRejectedEvents = 0;

   /** Delay value if an event is rejected. */
   protected long rejectedEventDelay = 1000;

   /** The EventFactory to use. */
   protected EventFactory eventFactory = null;
   
   /** The ID of this client. */
   protected String m_id = "";

   /** Whether or not to display additional information. */
   protected boolean m_showAdditionalInfo = true;

   /** Indicates if the proxy reference has been set externally. */
   protected boolean m_proxyReferenceSet = false;

   /** Separator used in displaying received events. */
   private String m_spacer = "  ";


   protected BenchmarkBaseClient ()
   {
   }
   

   public String toString ()
   {
      return m_id;
   }


   /**
    * Sets the ID for this client.
    * @param id   New identifier
    */
   public void setID (String id)
   {
      m_id = id;
   }


   /**
    * Sets the flag that determines if various additional output (i.e. that not
    * directly requested such as event rate reports) is sent to the console.
    * @param state    The new state for the flag
    */
   public void setShowAdditionalInfo (boolean state)
   {
      m_showAdditionalInfo = state;
   }


   /**
    * Prints command-line usage.
    */
   protected void showUsage ()
   {
      // General parameters.
      System.out.println ("Command-line parameters are as follows:");
      System.out.println (" -r name         Use <name> as the resolve name for the notification service");
      System.out.println ("                 Default: " + resolveNameDefault);
      System.out.println (" -id name        Set client name to <name>");
      System.out.println ("                 Default: " + m_id);
      System.out.println (" -cid number     Use channel with ID <number>");
      System.out.println ("                 Default: " + channelIdDefault);
      System.out.println (" -dt number      Pause after every <number> of events (0 to disable delays)");
      System.out.println ("                 Default: " + eventDelayThresholdDefault);
      System.out.println (" -dl number      Duration of above pause in milliseconds (0 to disable delays)");
      System.out.println ("                 Default: " + eventDelayDefault);
      System.out.println (" -rt number      Report event rate every <number> events (0 to disable)");
      System.out.println ("                 Default: " + reportThresholdDefault);
      System.out.println (" -rf file        File for report data (optional)");
      System.out.println (" -ef name        Fully qualified name of Event Factory to use. Omit this parameter " +
                          "entirely to use default factory.");

      // Supplier-specific parameters.
      if (isSupplier)
      {
         if (clientType != ClientType.ANY_EVENT)
         {
            System.out.println (" -p true|false   Setting for EventReliability in the event header (override channel setting)");
            System.out.println ("                 Default: do not override (use channel setting)");
         }
         System.out.println (" -c number       Send <number> of events (0 for infinite events)");
         System.out.println ("                 Default: " + eventLimitDefault);

         if (clientType == ClientType.SEQUENCE_EVENT)
         {
            System.out.println (" -bs             Number of structured events to send in each batch");
            System.out.println ("                 Default: " + batchSize);
         }
      }

      // Consumer-specific parameters.
      if (! isSupplier)
      {
         if (clientType == ClientType.STRUCTURED_EVENT)
         {
            System.out.println (" -co true|false  Confirm event IDs are received in FIFO order");
            System.out.println ("                 Default: " + checkOrderingDefault);
         }

         if (clientType == ClientType.SEQUENCE_EVENT)
         {
            System.out.println (" -bs             Value to set for MaximumBatchSize QoS property");
            System.out.println ("                 Default: " + batchSizeDefault);
            System.out.println (" -pi             Value to set for PacingInterval QoS property");
            System.out.println ("                 Default: " + pacingIntervalDefault);
         }

         System.out.println (" -de             Whether or not to display the received events");
         System.out.println ("                 Default: " + displayReceivedEventsDefault);
      }

      // Other parameters.
      if (clientType == ClientType.STRUCTURED_EVENT)
      {
         System.out.println (" -asbatch        Value for AutoSequenceBatchSize");
         System.out.println ("                 Default: " + autoSequenceBatchSizeDefault);
         System.out.println (" -astimeout      Value for AutoSequenceTimeout");
         System.out.println ("                 Default: " + autoSequenceTimeoutDefault);
      }
   }


   /**
    * Parses command-line arguments; displays relevant information.
    * @param args  The command line arguments to parse.
    */
   protected void parseArguments (String[] args)
   {
      // Show usage if required.
      if (args.length != 0)
      {
         String firstArg = args [0];
         if (firstArg.equalsIgnoreCase ("?")    ||
             firstArg.equalsIgnoreCase ("-?")   ||
             firstArg.equalsIgnoreCase ("help") ||
             firstArg.equalsIgnoreCase ("-help"))
         {
            showUsage ();
            System.exit (-1);
         }
      }

      m_id = BenchmarkUtils.extractArg (args, "-id", m_id);

      rejectedEventDelay = BenchmarkUtils.extractArg(args, "-rd", rejectedEventDelay);

      channelId = BenchmarkUtils.extractArg (args, "-cid", channelIdDefault);

      /* In order to avoid problems with VisiBroker when registering the
         name NotificationService, rename it */
      if ((com.prismt.openfusion.Version.getORB().toUpperCase().startsWith("VISIBROKER-52")) || (com.prismt.openfusion.Version.getORB().toUpperCase().startsWith("VISIBROKER-60")))
      {
         resolveNameDefault = "OFNotificationService";
      }
      
      resolveName = BenchmarkUtils.extractArg (args, "-r", resolveNameDefault);
      /* End of specific code to deal with VisiBroker and registering
       NotificationService name problem*/

      int persCode = BenchmarkUtils.extractBooleanArg (args, "-p");
      if (persCode == BenchmarkUtils.TRUE_SPECIFIED)
      {
         persistence = true;
         persistenceSpecified = true;
      }
      else if (persCode == BenchmarkUtils.FALSE_SPECIFIED)
      {
         persistence = false;
         persistenceSpecified = true;
      }
      else
      {
         persistenceSpecified = false;
      }

      autoSequenceBatchSize = BenchmarkUtils.extractArg (args, "-asbatch",
                                                         autoSequenceBatchSizeDefault);
      autoSequenceTimeout   = BenchmarkUtils.extractArg (args, "-astimeout",
                                                         autoSequenceTimeoutDefault);

      reportThreshold =
            BenchmarkUtils.extractArg (args, "-rt", reportThresholdDefault);
      reportFilename =
            BenchmarkUtils.extractArg (args, "-rf", reportFilenameDefault);
      if ((reportThreshold == 0) ||
          (reportFilename.equals (reportFilenameDefault)))
      {
         reportToFile = false;
      }
      else
      {
         reportToFile = true;
      }

      eventLimit = BenchmarkUtils.extractArg (args, "-c",  eventLimitDefault);
      eventDelayThreshold =
            BenchmarkUtils.extractArg (args, "-dt", eventDelayThresholdDefault);
      eventDelay = BenchmarkUtils.extractArg (args, "-dl", eventDelayDefault);
      checkOrdering =
            BenchmarkUtils.extractArg (args, "-co", checkOrderingDefault);
      batchSize = BenchmarkUtils.extractArg (args, "-bs", batchSizeDefault);
      pacingInterval =
            BenchmarkUtils.extractArg (args, "-pi", pacingIntervalDefault);
      displayReceivedEvents =
            BenchmarkUtils.extractArg (args, "-de", displayReceivedEventsDefault);
      
      String defaultName = BenchmarkUtils.extractArg (args, "-ef", "" );
      
      String eventFactoryInfo = null;
      if (defaultName.length() > 0)
      {
         try
         {
            eventFactory = (EventFactory)(Class.forName (defaultName)).newInstance();
            eventFactoryInfo = "using EventFactory: " + defaultName;
         }
         catch (Throwable t)
         {
            eventFactory = new DefaultEventFactory ();
            eventFactoryInfo = "could not instantiate EventFactory, using default (" + t + ")";
         }
      }
      else
      {
         eventFactory = new DefaultEventFactory();
         eventFactoryInfo = "using default EventFactory";
      }

      if ((eventDelayThreshold == 0) || (eventDelay == 0))
      {
         performDelays = false;
      }
      else
      {
         performDelays = true;
      }

      if (m_showAdditionalInfo)
      {
         System.out.println (this + ": service resolve name: " + resolveName);
         System.out.println (this + ": channel ID: " + channelId);
         System.out.println (this + ": " + eventFactoryInfo);

         if ( (isSupplier) && (clientType != ClientType.ANY_EVENT))
         {
            String persText = null;
            if (persistenceSpecified)
            {
               if (persistence)
               {
                  persText = "override channel setting to Persistent";
               }
               else
               {
                  persText = "override channel setting to BestEffort";
               }
            }
            else
            {
               persText = "do not override channel setting";
            }
            System.out.println (this +": event persistence: " + persText);
         }

         System.out.println (this + ": report threshold: " +
                             (reportThreshold == 0 ? "no reporting" : "" + reportThreshold));
         if (reportToFile)
         {
            System.out.println (this +": report file: " + reportFilename);
         }

         // Event limit is for supplier only.
         if (isSupplier)
         {
            System.out.println (this + ": event limit: " +
                                (eventLimit == 0 ? "unlimited" : "" + eventLimit));
         }

         if (! isSupplier)
         {
            // Order checking is for structured consumer only.
            if (clientType == ClientType.STRUCTURED_EVENT)
            {
               System.out.println (this + ": check event ordering: " +
                                   (checkOrdering ? "yes" : "no"));
            }

            // Event displaying is for consumers only.
            System.out.println (this + ": display received events: " +
                                (displayReceivedEvents ? "yes" : "no"));
         }

         if (performDelays)
         {
            System.out.println (this + ": delay interval: " + eventDelayThreshold + " events");
            System.out.println (this + ": delay duration: " + eventDelay + "ms");
         }

         if (clientType == ClientType.SEQUENCE_EVENT)
         {
            if (isSupplier)
            {
               System.out.println (this + ": events per sequence: " + batchSize);
            }
            else
            {
               System.out.println (this + ": value for MaximumBatchSize QoS: " + batchSize);
               System.out.println (this + ": value for PacingInterval QoS: " + pacingInterval);
            }
         }
         System.out.println (this + ": rejected event delay: " + rejectedEventDelay + " ms");

         if (clientType == ClientType.STRUCTURED_EVENT)
         {
            System.out.println (this + ": value for AutoSequenceBatchSize QoS: " + autoSequenceBatchSize);
            System.out.println (this + ": value for AutoSequenceTimeout QoS: " + autoSequenceTimeout);
         }
      }
   }


   /**
    * Obtains the reference to the ORB.
    * @param args  The command line arguments.
    */
   protected void getOrbReference (String[] args)
   {
      try
      {
         if (orb == null)
         {
            orb = BenchmarkUtils.getORB (args, this);
         }
      }
      catch (Exception ex)
      {
         System.err.println (this + ": failed to get ORB reference");
         ex.printStackTrace ();
         System.exit (-1);
      }
   }


   /**
    * Directly sets the reference to the ORB.  This method is used in the case
    * when the benchmark client is being used by code that has already obtained
    * an ORB reference.
    * @param orb   The ORB reference
    */
   public void setOrbReference (org.omg.CORBA.ORB orb)
   {
      this.orb = orb;
   }


   /**
    * Obtains the reference to the CORBA object for the client.
    */
   protected void getThisReference ()
   {
      thisRef = ObjectAdapter.createTransient (this);
   }


   /**
    * Obtains the reference to the event channel factory.
    */
   protected void getEventChannelFactoryReference ()
   {
      try
      {
         if (factory == null)
         {
            factory = BenchmarkUtils.getEventChannelFactory (orb, resolveName, this);
         }
      }
      catch (Exception ex)
      {
         System.out.println (this + ": failed to get event channel factory reference");
         ex.printStackTrace ();
         System.exit (-1);
      }
   }


   /**
    * Directly sets the reference to the event channel factory.  This method is
    * used in the case when the benchmark client is being used by code that has
    * already obtained a event channel factory reference, and so the BenchmarkUtil
    * method is not needed.
    * @param factory   The event channel factory reference
    */
   public void setEventChannelFactoryReference (EventChannelFactory factory)
   {
      this.factory = factory;
   }


   /**
    * Obtains the reference to the event channel.
    */
   protected void getEventChannelReference ()
   {
      try
      {
         if (channel == null)
         {
            channel = BenchmarkUtils.getEventChannel (orb, factory, channelId, this);
         }
      }
      catch (Exception ex)
      {
         System.out.println (this + ": failed to get event channel " + channelId);
         ex.printStackTrace ();
         System.exit (-1);
      }
   }


   /**
    * Directly sets the reference to the event channel.  This method is used in
    * the case when the benchmark client is being used by code that has already
    * obtained a channel reference, and so the BenchmarkUtil method is not
    * needed.
    * @param channel   The event channel reference
    */
   public void setEventChannelReference (EventChannel channel)
   {
      this.channel = channel;
   }


   /**
    * Directly sets the proxy reference that this client should use.
    * @param proxy   Proxy reference
    */
   abstract public void setProxyReference (org.omg.CORBA.Object proxy);


   /**
    * Directly sets the supplier admin reference.
    * @param suppAdmin   Supplier admin reference.
    */
   public void setSupplierAdminReference (SupplierAdmin suppAdmin)
   {
      this.suppAdmin = suppAdmin;
   }


   /**
    * Directly sets the consumer admin reference.
    * @param consAdmin   Consumer admin reference.
    */
   public void setConsumerAdminReference (ConsumerAdmin consAdmin)
   {
      this.consAdmin = consAdmin;
   }


   /**
    * Obtains the references to the admins.
    */
   protected void getAdminReferences ()
   {
      getAdminReferences (false);
   }


   /**
    * Obtains the references to the admins.  The default admins are used on the
    * channel.  If a default admin does not exist, a new admin will be created
    * if createAdmins is true.
    * @param createAdmins   Boolean: true to create new admins if the default
    *                       admins do not exist.
    */
   protected void getAdminReferences (boolean createAdmins)
   {
      if (suppAdmin == null)
      {
         suppAdmin = channel.default_supplier_admin ();
         if (suppAdmin == null)
         {
            if (m_showAdditionalInfo)
            {
               System.out.println (this + ": Could not find default supplier admin");
            }

            if (createAdmins)
            {
               org.omg.CosNotifyChannelAdmin.InterFilterGroupOperator filtOp =
                  org.omg.CosNotifyChannelAdmin.InterFilterGroupOperator.OR_OP;
               org.omg.CORBA.IntHolder idHolder = new org.omg.CORBA.IntHolder ();
               suppAdmin = channel.new_for_suppliers (filtOp, idHolder);
               if (m_showAdditionalInfo)
               {
                  System.out.println (this + ": Created new supplier admin: ID is " + idHolder.value);
               }
            }
         }
      }
      
      if (consAdmin == null)
      {
         consAdmin = channel.default_consumer_admin ();
         if (consAdmin == null)
         {
            if (m_showAdditionalInfo)
            {
               System.out.println (this + ": Could not find default consumer admin");
            }

            if (createAdmins)
            {
               org.omg.CosNotifyChannelAdmin.InterFilterGroupOperator filtOp =
                  org.omg.CosNotifyChannelAdmin.InterFilterGroupOperator.OR_OP;
               org.omg.CORBA.IntHolder idHolder = new org.omg.CORBA.IntHolder ();
               consAdmin = channel.new_for_consumers (filtOp, idHolder);
               if (m_showAdditionalInfo)
               {
                  System.out.println (this + ": Created new consumer admin: ID is " + idHolder.value);
               }
            }
         }
      }
   }


   /**
    * Initialises the QoS property array with any needed values.  By default
    * no properties are used.
    */
   protected void prepareQoSProperties ()
   {
   }


   /**
    * Performs the general initialisation for the client, including the
    * connection to the event channel.
    * @param args  The command line arguments.
    */
   public void initialise (String[] args)
   {
      parseArguments (args);
      getOrbReference (args);
      ObjectAdapter.ready (false);
      getThisReference ();
      getEventChannelFactoryReference ();
      getEventChannelReference ();
      getAdminReferences ();
      openReportFile ();
      prepareQoSProperties ();
   }


   /**
    * Opens the report file for writing.
    */
   protected void openReportFile ()
   {
      if (reportToFile)
      {
         try
         {
            //File f = new File (reportFilename);
            //fw = new FileWriter (f, false);
            fw = new FileWriter (reportFilename, false);

            bw = new BufferedWriter (fw);

            // Write the header.
            bw.write ("Elapsed\tCount\tEv/Sec (Start)\tEv/Sec (Last Period)");
            bw.newLine ();
            bw.flush ();
         }
         catch (Exception ex)
         {
            System.err.println (this + ": Could not open the report file for writing: " + ex);
            ex.printStackTrace ();
            System.err.println (this + ": Operation will continue, but reports will only be sent to the console");
            reportToFile = false;
         }
      }
   }


   /**
    * Closes the report file.
    */
   protected void closeReportFile ()
   {
      if (reportToFile)
      {
         try
         {
            if (bw != null)
            {
               bw.flush ();
               bw.close ();
            }
            if (fw != null)
            {
               fw.close ();
            }
         }
         catch (Exception ex)
         {
            System.err.println (this + ": could not close the report file: " + ex);
            ex.printStackTrace ();
         }
      }
   }


   /**
    * Set the QoS properties that will be applied to the proxy on connection.
    * @param qos   Property array with QoS properties.
    */
   public void setQoSProperties (Property[] qos)
   {
      m_qos = qos;
   }
   
   
   /**
    * Sets the filters that will be added to the proxy on connection.
    * @param filters   Vector with the filters.
    */
   public void setFilters (Vector filters)
   {
      m_filters = filters;
   }
   

   /**
    * Connects this client to the event channel.
    */
   public abstract void connect ();


   /**
    * Applies the QoS properties and filters to the proxy.
    * @param proxy   Proxy CORBA object reference.
    */
   protected void applyQoSFilters (org.omg.CORBA.Object proxy)
      throws Exception
   {
      if ((m_qos != null) && (m_qos.length != 0))
      {
         QoSAdmin admin = QoSAdminHelper.narrow (proxy);
         admin.set_qos (m_qos);
      }

      if ((m_filters != null) && (m_filters.size () != 0))
      {
         FilterAdmin admin = FilterAdminHelper.narrow (proxy);
         Enumeration enum = m_filters.elements ();
         Filter filter = null;
         while (enum.hasMoreElements ())
         {
            filter = (Filter) enum.nextElement ();
            admin.add_filter (filter);
         }
      }
   }


   /**
    * Disconnects this client from the event channel.
    */
   public abstract void disconnect ();


   /**
    * Performs the delay if required.
    * @param numEvents  The number of events that have been sent/received so far.
    */
   protected void delayIfNeeded (int numEvents)
   {
      if (performDelays)
      {
         if (numEvents % eventDelayThreshold == 0)
         {
            delay (numEvents);
         }
      }
   }


   /**
    * Performs the delay if required.  This variant is for sequence proxies,
    * where the batch size added to the current total will not necessarily
    * be exact multiples of the delay threshold.
    * @param currentTotal  The current total of events.
    * @param batchSize     The batch size just sent/recieved.
    */
   protected void delayIfNeeded (int currentTotal, int batchSize)
   {
      if (performDelays)
      {
         int threshCountBefore = (int) (currentTotal / eventDelayThreshold);
         int threshCountAfter  = (int) ((currentTotal+batchSize) / eventDelayThreshold);

         if (threshCountAfter > threshCountBefore)
         {
            delay (currentTotal + batchSize);
         }
      }
   }


   /**
    * Performs the required delay.
    * @param numEvents  The number of events that have been sent/received so far.
    */
   protected void delay (int numEvents)
   {
      if (m_showAdditionalInfo)
      {
         String prefix = (isSupplier ? "Sent " : "Received ");
         System.out.println (this + ": " + prefix + numEvents +
                             " events, pausing for " + eventDelay + "ms...");
      }

      // Sleep using the default delay.
      sleep (-1);
   }


   /**
    * Performs a Thread.sleep() operation.
    * @param delay   The delay in ms, to use default delay specify -1
    */
   protected void sleep (long delay)
   {
      if (delay == -1)
      {
         delay = (long) eventDelay;
      }
      
      try
      {
         Thread.currentThread ().sleep ((long) delay);
         totalDelayDuration += eventDelay;
      }
      catch (InterruptedException ie)
      {
         System.err.println (this + ": delay interrupted");
      } 
   }


   /**
    * Reports the current event rate if required.
    * @param numEvents  The number of events that have been sent/received so far.
    */
   protected void reportIfNeeded (int numEvents)
   {
      if (reportThreshold != 0)
      {
         if (numEvents % reportThreshold == 0)
         {
            report (numEvents);
         }
      }
   }


   /**
    * Reports the current event rate if required.  This variant is for sequence
    * proxies, where the batch size added to the current total will not
    * necessarily be exact multiples of the reporting threshold.
    * @param currentTotal  The current total of events.
    * @param batchSize     The batch size just sent/recieved.
    */
   protected void reportIfNeeded (int currentTotal, int batchSize)
   {
      if (reportThreshold != 0)
      {
         int threshCountBefore = (int) (currentTotal / reportThreshold);
         int threshCountAfter  = (int) ((currentTotal+batchSize) / reportThreshold);

         if (threshCountAfter > threshCountBefore)
         {
            report (currentTotal + batchSize);
         }
      }
   }


   /**
    * Reports the current event rate.
    * @param numEvents  The number of events that have been sent/received so far.
    */
   protected void report (int numEvents)
   {
      // Event rate / time elapsed: since client was started.
      long eventRateSinceStart = 0;
      long elapsedTimeSinceStart = 0;

      // Event rate / time elapsed: since last report was produced.
      long eventRateLastPeriod = 0;
      long elapsedTimeLastPeriod = 0;

      // Calculate the elapsed time since the start, taking into account all
      // delays.
      elapsedTimeSinceStart = (currentTime - startTime - totalDelayDuration);

      // Calculate the elapsed time since the last report, taking into account
      // any delays since then.  The extra check here is to ensure that the
      // very first report is correct.
      if (lastReportTime == 0)
      {
         elapsedTimeLastPeriod = elapsedTimeSinceStart;
      }
      else
      {
         elapsedTimeLastPeriod = (currentTime - lastReportTime) -
                                 (totalDelayDuration - delayAtLastReport);
      }

      // Perform calculations: event rate since start.
      String textSinceStart = null;
      if (elapsedTimeSinceStart <= 0)
      {
         textSinceStart = "cannot be determined (time < 1ms)";
      }
      else if (numEvents <= 0)
      {
         textSinceStart = "cannot be determined: no events appear to have been " +
                          (isSupplier ? "sent" : "received");
      }
      else
      {
         eventRateSinceStart = (1000 * (long) numEvents) / elapsedTimeSinceStart;
         textSinceStart = "" + eventRateSinceStart + "ev/sec";
      }

      // Perform calculations: event rate since last report period.
      String textLastPeriod = null;
      long numEventsLastPeriod = numEvents - lastReportEventCount;
      if (elapsedTimeLastPeriod <= 0)
      {
         textLastPeriod = "cannot be determined (time < 1ms)";
      }
      else if (numEventsLastPeriod <= 0)
      {
         textLastPeriod = "cannot be determined: no events appear to have been " +
                          (isSupplier ? "sent" : "received");
      }
      else
      {
         eventRateLastPeriod = (1000 * (long) numEventsLastPeriod) / elapsedTimeLastPeriod;
         textLastPeriod = "" + eventRateLastPeriod + "ev/sec";
      }

      // Output to console.
      String rateText = (isSupplier ? "sent " : "received ") + numEvents + " event(s): " +
                        "rate since start "         + textSinceStart +
                        ", rate since last report " + textLastPeriod +
                        " num rejects: " + numRejectedEvents;
      System.out.println (this + ": " + rateText);

      // Output to file.
      if (reportToFile)
      {
         StringBuffer sb = new StringBuffer ();
         sb.append (elapsedTimeSinceStart);
         sb.append ("\t");
         sb.append (numEvents);
         sb.append ("\t");
         sb.append (eventRateSinceStart);
         sb.append ("\t");
         sb.append (eventRateLastPeriod);
         try
         {
            bw.write (sb.toString ());
            bw.newLine ();
            bw.flush ();
         }
         catch (Exception ex)
         {
            System.err.println (this + ": error writing to the report file: future " +
                                "reports will be sent to the console only");
            reportToFile = false;
         }
      }

      // Update counters.
      lastReportEventCount = numEvents;
      lastReportTime       = currentTime;
      delayAtLastReport    = totalDelayDuration;
   }


   /**
    * Produces a string representation of an Any.
    * @param theAny  The Any.
    */
   protected String anyToString (org.omg.CORBA.Any theAny)
   {
      return AnyUtil.createAnyString (theAny);
   }


   /**
    * Displays the details of an Any.
    * @param The Any.
    */
   protected void displayAny (org.omg.CORBA.Any theAny)
   {
      // We try to extract a StructuredEvent from the Any to take into account
      // the automatic event translation; if this fails, we just display the
      // Any as it is.

      try
      {
         StructuredEvent event = StructuredEventHelper.extract (theAny);
         System.out.println (this + ": <Any contained the following StructuredEvent>");
         displayStructuredEvent (event);
      }
      catch (Exception ex)
      {
         System.out.println (anyToString (theAny));
      }
   }


   /**
    * Displays the details of a structured event.
    * @param The structured event.
    */
   protected void displayStructuredEvent (StructuredEvent event)
   {
      // Fixed event header.
      FixedEventHeader fixedHeader = event.header.fixed_header;
      EventType eventType = fixedHeader.event_type;
      String eventName = fixedHeader.event_name;
      System.out.println (this + m_spacer + "*** FIXED HEADER ***");
      System.out.println (this + m_spacer + m_spacer + "Domain Name: " + eventType.domain_name);
      System.out.println (this + m_spacer + m_spacer + "Type Name: " + eventType.type_name);
      System.out.println (this + m_spacer + m_spacer + "Event Name: " + eventName);

      // Variable event header.
      Property[] variableHeader = event.header.variable_header;
      System.out.println (this + m_spacer + "*** VARIABLE HEADER ***");
      if (variableHeader.length == 0)
      {
         System.out.println (this + m_spacer + m_spacer + "<no properties>");
      }
      else
      {
         displayProperties (variableHeader);
      }

      // Filterable data.
      Property[] filterableData = event.filterable_data;
      System.out.println (this + m_spacer + "*** FILTERABLE DATA ***");
      if (filterableData.length == 0)
      {
         System.out.println (this + m_spacer + m_spacer + "<no properties>");
      }
      else
      {
         displayProperties (filterableData);
      }

      // Event body.
      org.omg.CORBA.Any eventBody = event.remainder_of_body;
      System.out.println (this + m_spacer + "*** REMAINDER-OF-BODY ***");
      System.out.println (this + m_spacer + m_spacer + anyToString (eventBody));

      System.out.println ("--------------------------------------------");
   }


   /**
    * Displays the details of an array of properties.
    * @param props  The array of properties.
    */
   protected void displayProperties (Property[] props)
   {
      Property prop = null;
      for (int i = 0 ; i < props.length ; i++)
      {
         prop = props[i];
         System.out.println (this + m_spacer + m_spacer + i + ": " + prop.name + " -- " +
                             AnyUtil.createAnyString (prop.value));
      }
   }


   /**
    * No-operation implementations of the supplier and consumer
    * operations that we are not interested in.
    */
   public void disconnect_push_consumer ()
   {
   }
   public void disconnect_push_supplier ()
   {
   }
   public void disconnect_structured_push_consumer ()
   {
   }
   public void disconnect_structured_push_supplier ()
   {
   }
   public void disconnect_sequence_push_consumer ()
   {
   }
   public void disconnect_sequence_push_supplier ()
   {
   }
   public void offer_change (EventType[] added, EventType[] removed)
   {
   }
   public void subscription_change (EventType[] added, EventType[] removed)
   {
   }
}


// EOF
