/*************************************************************************
 *                           - COPYRIGHT NOTICE -                        *
 *                                                                       *
 * The information contained herein and which follows is proprietary     *
 * information and is the property of PrismTech Corporation.  This       *
 * information is not to be divulged, released, copied, reproduced, or   *
 * conveyed to unauthorized personnel, companies, or other institutions  *
 * without the direct and expressed approval in writing of PrismTech     *
 * Corporation.                                                          *
 *                                                                       *
 *                                                                       *
 *                           - DISCLAIMER -                              *
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech")     *
 * are under any obligation to support the use of this code which is     *
 * provided without warranty of any kind. PrismTech shall (except as     *
 * otherwise prohibited by law) have no liability whatsoever in relation *
 * to its use.                                                           *
 *                                                                       *
 ************************************************************************/

/**
 * @author PrismTech
 * @version Version 1.0
 */

package com.prismt.cos.CosNotification.examples.News;

import org.omg.CosNotification.*;
import org.omg.CosNotifyComm.*;
import org.omg.CosNotifyChannelAdmin.*;
import org.omg.CosNotifyFilter.*;

import com.prismt.orb.ObjectAdapter;

/**
 * Listens for events containing stock quote information.
 */
public class Analyst implements ListenerOperations
{
   private org.omg.CORBA.ORB orb;
   private StructuredProxyPushSupplier proxy;
   private java.util.Vector history = new java.util.Vector ();
   private EventChannel channel = null;

   /**
    * Creates a new instance
    * @param orb the ORB to use
    */
   public Analyst (org.omg.CORBA.ORB orb)
   {
      this.orb = orb;

      System.out.print ("Readying object adapter...");
      ObjectAdapter.ready (false);
      System.out.println ("Success");

      // Get an event channel
      try
      {
         this.channel = ChannelUtil.getChannel (this.orb);
      }
      catch (org.omg.CORBA.ORBPackage.InvalidName in)
      {
         System.out.println ("Could not resolve notification service!");
         System.exit (1);
      }

      // Obtain a proxy supplier and connect to it
      System.out.print ("Obtaining proxy...");
      ClientType ct = ClientType.STRUCTURED_EVENT;
      org.omg.CORBA.IntHolder id = new org.omg.CORBA.IntHolder ();
      ConsumerAdmin ca = this.channel.default_consumer_admin ();
      ProxySupplier ps = null;
      try
      {
         ps = ca.obtain_notification_push_supplier (ct,id);
      }
      catch (AdminLimitExceeded ale)
      {
         System.out.println
            ("Admin proxy limit exceeded - could not create new proxy");
         System.exit (1);
      }
      this.proxy = StructuredProxyPushSupplierHelper.narrow (ps);
      System.out.println ("Success");

      System.out.print ("Creating consumer...");
      org.omg.CORBA.Object thisRef = ObjectAdapter.createTransient (this);
      Listener l = ListenerHelper.narrow (thisRef);

      try
      {
         this.proxy.connect_structured_push_consumer (l);
         System.out.println ("Success");
      }
      catch(Exception e)
      {
         System.out.println ("Caught exception: "+e);
         e.printStackTrace ();
         System.exit (1);
      }
      select ("Wall Street");
   }

   /**
    * Attach a filter to the proxy supplier that will limit events pushed
    * to this instance to those from a given bureau
    * @param bureau the bureau from which events will be accepted
    */
   public void select (String bureau)
   {
      /*
       * Specify the event type so that events that are not stock quotes will
       * not be pushed to this instance
       */
      EventType[] types = new EventType[1];
      types[0] = new EventType ("News", "StockQuote");

      // Only accept events whose priority is 5 or more
      String constraint = "$bureau == \'" + bureau + "\' and $Priority >= 5";

      // Remove existing filters
      this.proxy.remove_all_filters ();

      // Create the new filter
      FilterFactory factory = this.channel.default_filter_factory ();
      String grammar = "EXTENDED_TCL";
      Filter filter = null;
      try
      {
         filter = factory.create_filter (grammar);
         int id = this.proxy.add_filter (filter);
      }
      catch (InvalidGrammar ig)
      {
         System.out.println ("Invalid grammar!");
      }

      // Add the filter constraint specified above
      try
      {
         ConstraintExp[] exp = new ConstraintExp[1];
         exp[0] = new ConstraintExp (types, constraint);

         ConstraintInfo[] info = filter.add_constraints (exp);
      }
      catch (InvalidConstraint ic)
      {
         System.out.println ("Invalid Constraint!");
      }
   }

   /**
    * Prints out any events recieved since the last invocation of this method
    */
   public void print ()
   {
      if (history.size () > 0)
      {
         System.out.println ("Received news:");
      }

      java.util.Enumeration e = this.history.elements ();

      while(e.hasMoreElements ())
      {
         System.out.println (e.nextElement ());
      }
      history.removeAllElements ();
   }

   /**
    * Pushes a structured event to this instance
    * @param event the event
    */
   public synchronized void push_structured_event (StructuredEvent event)
   {
      String news = event.remainder_of_body.extract_string ();
      System.out.println
         (event.header.variable_header[1].value.extract_short ());

      history.addElement (news);
   }

   public void offer_change (EventType[] added, EventType[] removed)
   {
   }

   public void disconnect_structured_push_consumer ()
   {
   }

   public static void main (String[] args)
   {
      System.out.print ("Initialising the orb...");

      org.omg.CORBA.ORB orb = ObjectAdapter.init (args);
      System.out.println ("Success");
      Analyst a = new Analyst (orb);

      while (true)
      {
         a.print ();

         try
         {
            Thread.sleep (5000);
         }
         catch(InterruptedException ie)
         {
         }
      }
   }
}
