/*************************************************************************
 *                           - COPYRIGHT NOTICE -                        *
 *                                                                       *
 * The information contained herein and which follows is proprietary     *
 * information and is the property of PrismTech Corporation.  This       *
 * information is not to be divulged, released, copied, reproduced, or   *
 * conveyed to unauthorized personnel, companies, or other institutions  *
 * without the direct and expressed approval in writing of PrismTech     *
 * Corporation.                                                          *
 *                                                                       *
 *                                                                       *
 *                           - DISCLAIMER -                              *
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech")     *
 * are under any obligation to support the use of this code which is     *
 * provided without warranty of any kind. PrismTech shall (except as     *
 * otherwise prohibited by law) have no liability whatsoever in relation *
 * to its use.                                                           *
 *                                                                       *
 ************************************************************************/

/**
 * @author PrismTech
 * @version Version 1.0
 */

package com.prismt.cos.CosNotification.examples.News;

import org.omg.CosNotification.*;
import org.omg.CosNotifyComm.*;
import org.omg.CosNotifyChannelAdmin.*;
import org.omg.CosNotifyFilter.*;

import com.prismt.orb.ObjectAdapter;

/**
 * Receives structured events containing weather reports
 */
public class WeatherMan implements ListenerOperations
{
   private org.omg.CORBA.ORB orb;
   EventChannel channel;
   StructuredProxyPushSupplier proxy;
   private java.util.Vector history = new java.util.Vector ();

   /**
    * Creates a new instance
    * @param orb the ORB to use
    */
   public WeatherMan (org.omg.CORBA.ORB orb)
   {
      this.orb = orb;

      System.out.print ("Readying orb...");
      ObjectAdapter.ready (false);
      System.out.println ("Success");

      // Get an event channel
      try
      {
         this.channel = ChannelUtil.getChannel (this.orb);
      }
      catch (org.omg.CORBA.ORBPackage.InvalidName in)
      {
         in.printStackTrace ();
      }

      // Get a structured proxy push supplier
      ConsumerAdmin ca = this.channel.default_consumer_admin ();
      ClientType ct = ClientType.STRUCTURED_EVENT;
      org.omg.CORBA.IntHolder id = new org.omg.CORBA.IntHolder ();
      ProxySupplier ps = null;
      try
      {
         ps = ca.obtain_notification_push_supplier (ct, id);
      }
      catch(AdminLimitExceeded ale)
      {
         System.out.println ("Admin Limit exceeded!");
         System.exit (1);
      }
      this.proxy = StructuredProxyPushSupplierHelper.narrow (ps);

      System.out.print ("Creating consumer...");
      org.omg.CORBA.Object thisObj = ObjectAdapter.createTransient (this);

      Listener consumer = null;

      // Cpnnect to the proxy
      try
      {
         consumer = ListenerHelper.narrow (thisObj);
         this.proxy.connect_structured_push_consumer (consumer);
         System.out.println ("Success");
      }
      catch (org.omg.CosEventChannelAdmin.AlreadyConnected ac)
      {
         System.out.println ("Failed: " + ac);
         System.exit (1);
      }
      catch(Exception e)
      {
         System.out.println ("Failed: " + e);
         System.exit (1);
      }

      // Add a filter for the appropriate events
      select ("Station North");
   }

   /**
    * Adds a filter to the proxy that ensures that only those events this
    * instance is intereeted in are delivered
    * @param bureau the name of the weather bureau from which to receive
    * events.
    */
   public void select (String bureau)
   {
      EventType[] types = new EventType[1];

      // Only receive weather events
      types[0] = new EventType ("News","Weather");

      // Only receive events whose bureau name matches that which is required
      String constraint = "$bureau == \'"+bureau+"\'";

      FilterFactory factory = this.channel.default_filter_factory ();
      String grammar = "EXTENDED_TCL";
      Filter filter = null;

      // Create the filter
      try
      {
         filter = factory.create_filter (grammar);
         int id = this.proxy.add_filter (filter);
      }
      catch(InvalidGrammar ig)
      {
         System.out.println("Invalid Grammar: " + ig);
         System.exit (1);
      }

      // Add the constraint
      try
      {
         ConstraintExp[] ce = new ConstraintExp[1];
         ce[0] = new ConstraintExp (types,constraint);

         ConstraintInfo[] info = filter.add_constraints (ce);
      }
      catch(InvalidConstraint ic)
      {
         System.out.println ("Invalid Constraint: " + ic);
      }
   }

   /**
    * Prints out any events recieved since the last invocation of this method
    */
   public void print ()
   {
      java.util.Enumeration e = this.history.elements ();
      while (e.hasMoreElements ())
      {
         System.out.println ((String) e.nextElement ());
      }
      history.removeAllElements ();
   }

   /**
    * Pushes a structured event to this instance
    * @param event the event
    */
   public void push_structured_event (StructuredEvent event)
   {
      String message = event.filterable_data[1].value.extract_string() +
         " - " + event.remainder_of_body.extract_string ();

      System.out.println ("Got message");
      this.history.add (message);
   }

   public void offer_change (EventType[] added, EventType[] removed)
   {
   }

   public void disconnect_structured_push_consumer ()
   {
   }

   public static void main (String args[])
   {
      System.out.print ("Initialising the orb...");

      org.omg.CORBA.ORB orb = ObjectAdapter.init (args);
      System.out.println ("Success");
      WeatherMan wm = new WeatherMan (orb);

      while (true)
      {
         wm.print ();
         try
         {
            Thread.sleep (5000);
         }
         catch (InterruptedException ie)
         {
         }
      }
   }
}
