/*************************************************************************
 *                           - COPYRIGHT NOTICE -                        *
 *                                                                       *
 * The information contained herein and which follows is proprietary     *
 * information and is the property of PrismTech Corporation.  This       *
 * information is not to be divulged, released, copied, reproduced, or   *
 * conveyed to unauthorized personnel, companies, or other institutions  *
 * without the direct and expressed approval in writing of PrismTech     *
 * Corporation.                                                          *
 *                                                                       *
 *                                                                       *
 *                           - DISCLAIMER -                              *
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech")     *
 * are under any obligation to support the use of this code which is     *
 * provided without warranty of any kind. PrismTech shall (except as     *
 * otherwise prohibited by law) have no liability whatsoever in relation *
 * to its use.                                                           *
 *                                                                       *
 ************************************************************************/

/**
 * @author PrismTech
 * @version Version 1.0
 */

package com.prismt.cos.CosNotification.examples.News;

import org.omg.CosNotification.*;
import org.omg.CosNotifyComm.*;
import org.omg.CosNotifyChannelAdmin.*;

import com.prismt.orb.ObjectAdapter;

import java.util.Random;

/**
 * Creates structured events representing weather reports and pushes them to
 * an event channel.
 */
public class WeatherStation implements BureauOperations
{
   org.omg.CORBA.ORB orb;
   EventChannel channel;
   StructuredProxyPushConsumer proxy;

   /**
    * All the possible elements of weather information that may go into a report
    */
   String[][] allNews =
   {
      {"San Francisco", "80 miles/hour storm"    },
      {      "Houston", "4 inches of snow"       },
      {     "New York", "Clouds and 43 degrees"  },
      {      "Seattle", "A bit of rain"          },
      {      "Atlanta", "Sun and 89 degrees"     },
      {    "Newcastle", "Rain and mist"          },
      {   "Louisianna", "Twister destroys homes" }
   };
   private boolean stopped = false;

   /**
    * Create a new instance
    * @param orb the ORB to use
    */
   public WeatherStation (org.omg.CORBA.ORB orb)
   {
      this.orb = orb;

      System.out.print ("Readying orb...");
      ObjectAdapter.ready (false);
      System.out.println ("Success");

      // Get the event channel
      try
      {
         System.out.print ("Getting channel...");
         this.channel = ChannelUtil.getChannel (this.orb);

         if (channel != null)
         {
            System.out.println ("Success");
         }
         else
         {
            System.out.println ("Error getting channel!");
            System.exit (1);
         }
      }
      catch (org.omg.CORBA.ORBPackage.InvalidName in)
      {
         System.out.println ("Could not resolve Notification Service!");
         System.exit (1);
      }

      // Get a structured proxy push consumer
      SupplierAdmin sa = this.channel.default_supplier_admin ();
      ClientType ct = ClientType.STRUCTURED_EVENT;
      org.omg.CORBA.IntHolder id = new org.omg.CORBA.IntHolder ();
      ProxyConsumer pc = null;
      System.out.print ("Getting proxy...");

      try
      {
         pc = sa.obtain_notification_push_consumer (ct,id);
      }
      catch (AdminLimitExceeded ale)
      {
         System.out.println
            ("Could not obtain new proxy - AdminLimitExceeded: " + ale);

         System.exit (1);
      }

      this.proxy = StructuredProxyPushConsumerHelper.narrow (pc);
      System.out.println ("Success");

      // Create a CORB object for this instance
      System.out.print ("Creating supplier...");
      org.omg.CORBA.Object thisObj = ObjectAdapter.createTransient (this);
      System.out.println ("Success");
      Bureau b = BureauHelper.narrow (thisObj);

      // Connect to the proxy structured push consumer
      try
      {
         System.out.print ("Connecting to proxy...");
         this.proxy.connect_structured_push_supplier (b);
         System.out.println ("Success");
      }
      catch (org.omg.CosEventChannelAdmin.AlreadyConnected ac)
      {
         System.out.println ("Could not connect to proxy!");
         System.exit (1);
      }
   }

   /**
    * Publish structured events representing weather reports
    */
   public void broadcast ()
   {
      while (!this.stopped)
      {
         StructuredEvent event = getNextEvent ();
         try
         {
            this.proxy.push_structured_event (event);
            System.out.println
               (
                "Sent event: " +
                event.filterable_data[1].value.extract_string() + " - " +
                event.remainder_of_body.extract_string ()
               );

            Thread.sleep (Math.abs ((new Random ()).nextInt () % 5000));
         }
         catch (InterruptedException ie)
         {
         }
         catch (org.omg.CosEventComm.Disconnected d)
         {
            d.printStackTrace ();
            break;
         }
      }
   }

   public void stop ()
   {
      this.stopped = true;
   }

   /**
    * Create a new structured event representing a randomly generated weather
    * report.
    * @return the structured event
    */
   private StructuredEvent getNextEvent ()
   {
      Random r = new Random ();
      String city = allNews[Math.abs (r.nextInt () % allNews.length)][0];
      String news = allNews[ Math.abs(r.nextInt () % allNews.length)][1];
      StructuredEvent event = new StructuredEvent ();
      EventType type = new EventType ("News", "Weather");
      FixedEventHeader feh = new FixedEventHeader (type, "");

      event.header = new EventHeader ();
      event.header.variable_header = new Property[0];
      event.header.fixed_header = feh;

      Property[] filterable = new Property[2];

      filterable[0] = new Property();
      filterable[0].name = "bureau";
      filterable[0].value = this.orb.create_any ();
      filterable[0].value.insert_string ("Station North");

      filterable[1] = new Property ();
      filterable[1].name = "city";
      filterable[1].value = this.orb.create_any ();
      filterable[1].value.insert_string (city);

      event.filterable_data = filterable;

      event.remainder_of_body = this.orb.create_any ();
      event.remainder_of_body.insert_string (news);

      return event;
   }

   public void disconnect_structured_push_supplier ()
   {
   }

   public void subscription_change (EventType[] added, EventType[] removed)
   {
   }

   public static void main (String args[])
   {
      System.out.print ("Initialising the orb...");

      org.omg.CORBA.ORB orb = ObjectAdapter.init (args);
      System.out.println ("Success");
      java.io.BufferedReader br = new java.io.BufferedReader
         (new java.io.InputStreamReader (System.in));

      WeatherStation ws = new WeatherStation (orb);
      System.out.println ("Ready to send events. Press enter to start.");

      try
      {
         br.readLine ();
      }
      catch(java.io.IOException ioe)
      {
      }
      ws.broadcast ();
   }
}
