package org.jacorb.imrutility.imr;

import java.io.Serializable;

import org.jacorb.util.Debug;

/**
 * @author cj
 * This class holds the name to be used when registering servers/POAs. 
 * The name key may consist of only poa names or it may consist of an
 * implementation name and poa names.  This definition is required to 
 * cater for the registration of non-jacorb objects as the ultimate
 * aim is that servers will be able to register with any Implementation
 * Repository e.g TAO servers may register with the JacORB ImR.
 */
public class POANameKey implements Serializable
{
    /**The implementation name - this may be blank*/
    private String implName;
    
    /**An array of POA names */
    private String [] poas;
    
    /** The concatenation of the <code>implName</code> and the 
     * <code>poas</code>
     */
    private String stringified;

    private String pstringified;
    
    /**
     * Empty constructor
     */
    public POANameKey()
    {
    }
    
    /**
     * Constructor with details already populated
     * @param <code>iname</code> the implementation name
     * @param <code>ipoas</code> an array of poa names
     */
    public POANameKey(String iname, String [] ipoas)
    {
        implName = iname;
        poas = ipoas;
        stringified = implName;
        for (int i = 0; i < ipoas.length; i++)
        {
           stringified += "/";
           stringified += ipoas[i];           
           
           pstringified += "/";
           pstringified += ipoas[i];           

        }           
    }
    
    /**
     * Constructor taking only an implementation name
     * @param <code>iname</code> the implementation name
     */
    public POANameKey(String iname)
    {
        implName = iname;
        poas = new String[0];
        stringified = implName;
    }
    
    /**
     * Constructor taking only an array of poa names
     * @param <code>ipoas</code> an array of names
     */
    public POANameKey(String [] ipoas)
    {
        poas = ipoas;
        implName = "";
        
        for (int i = 0; i < ipoas.length; i++)
        {
           stringified += "/";
           stringified += ipoas[i];

           pstringified += "/";
           pstringified += ipoas[i];           

        }
    }
    
    /**
     * Returns the implementaion name
     */
    public String getImplName()
    {
        return implName;
    }
    
    /**
     * Returns the poa names
     */
    public String [] getPOAS()
    {
        return poas;
    }
    
    /**
     * Returns the POA names stringified
     */
    public String getPOASStringified()
    {
        return pstringified;
    }
    
    /**
     * Returns this key as a {@link ServerName}
     */
    public ServerName keyToServerName()
    {
        ServerName sname = new ServerName();
        String [][] spoas = new String [1] [poas.length];
        
        if (implName.equals(""))
        {
            for (int i = 0; i < poas.length; i++)
            {
                spoas[0][i] = poas[i];
            }
            
            sname.poas(spoas);
        }  
        else
        {
            sname.impl(implName);
        } 
        
        return sname;
    }
    
    /**
     * Returns an array of <code>POANameKey</code> created from 
     * a {@link ServerName}     
     * @param <code>sname</code> the <code>ServerName</code>
     * @return an array of keys derived from the <code>ServerName</code>
     */
    public static POANameKey [] serverNameToKeys(ServerName sname)
    {
        POANameKey [] keys;
        String [] [] poaKeys;
        
        if (sname.discriminator())
        {   
            keys =  new POANameKey[1];
            keys[0] = new POANameKey(sname.impl());
        }
        else
        {
            int noOfKeys = sname.poas().length;
            keys = new POANameKey[noOfKeys];
            poaKeys = sname.poas();
            
            for (int i = 0; i < noOfKeys; i++)
            {
                keys[i] = new POANameKey(poaKeys[i]);
            }
        }
     
        return keys;
    }
    
    /**
     * Returns a hash code for this key
     */    
    public int hashCode()
    {
        return stringified.hashCode();      
    }
    
    /**
     * Returns a <code>String</code> representing the concatenation of the
     * <code>implName</code> and the <code>poas</code>
     */
    public String toString()
    {
        return stringified;
    }

    /**
     * Method for matching keys
     */    
    public boolean equals(Object key)
    {
        if (key instanceof POANameKey)
        {
            return this.toString().equals(((POANameKey) key).toString());
        }
        
        return false;
    }
 
    /**
     * Method to check whether this key has an impl name
     */   
    public boolean hasImpl()
    {
        return implName != null && !implName.equals("");
    }
}

