package org.jacorb.imrutility.imr.persistence;


import java.io.*;
import java.util.HashMap;

import org.jacorb.util.Debug;
import org.jacorb.util.Environment;

import org.apache.avalon.framework.logger.*;

/**
 * This class is responsible for the persistence of data added to and
 * removed from the JacORB Implementation Repository 
 * @see org.jacorb.imrutility.imr.RepositoryImpl where the persistence
 * mechanism is file.  The persistence mechanism will be configurable
 */ 
public class FilePersistence 
{  
    /** 
     * The output file to which data is to be written.  As a number of writes
     * may occur in the life of the Repository the file initialisation should
     * be performed only once and a reference kept to the file.
     */ 
    private File outputFile;

    /** 
     * Logger to output any logging info
     */
    protected Logger logger = Debug.getNamedLogger("jacorb.imr");
        
    /**
     * Reads data from the specified file - only occurs at start up of the
     * ImR
     */ 
    public RepositoryData read()
    {
        String dataFile = Environment.getProperty("jacorb.imr.imr_data_file");
        RepositoryData repData = null;
        ObjectInputStream ois = null;
        
        if (dataFile != null)
        {            
            File f = new File(dataFile);
            
            try
            {
                if (f.exists())
                {
                    ois = new ObjectInputStream(new FileInputStream(f));
        
                    repData = (RepositoryData) ois.readObject();
                                
                    ois.close();
                }
            }
            catch (IOException ioe)
            {
                if (logger.isDebugEnabled())
                {
                    logger.warn
                        ("WARNING: Unable to read ImR data file " 
                            + dataFile +"! ");
                }            
            }
            catch (ClassNotFoundException cnfe)
            {
                if (logger.isDebugEnabled())
                {
                    logger.warn
                        ("WARNING: ImR data file " + dataFile  
                            + " does not contain valid object" +"! ");                    
                }
            }
        }
        else
        {
            if (logger.isDebugEnabled())
            {
                logger.warn("WARNING: ImR data file name not configured! ");                     
            }            
        } 
        
        if (repData == null)
        {
            repData = new RepositoryData();
        }                       

        ois = null;
                
        return repData;
    }
    
    /**
     * Instigates persistence of data to file
     */
    public void write(RepositoryData repData)
    {  
        if (outputFile == null)
        {
            String dataFile = Environment.getProperty("jacorb.imr.imr_data_file");
        
            if (dataFile == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.warn
                    ( 
                        "WARNING: ImR data file name not configured"
                            + " creating \"data.dat\" in current directory, " 
                    );
                    
                    dataFile = "data.dat";
                }            
            }

            outputFile = new File(dataFile);        
        
            if (!outputFile.exists())
            {            
                try
                {
                    outputFile.createNewFile();    
                }
                catch (IOException ioe)
                {
                    if (logger.isDebugEnabled())               
                    {
                        logger.error
                        (
                            "Unable to create ImR data file " + dataFile,
                            ioe
                        );
                
                        System.exit(-1);
                    }
                }
            }
        }
        
        write(outputFile, repData);
    }

    /**
     * Used at shutdown to create a backup of the data file - only occurs at 
     * shutdown.
     */    
    public void backup(RepositoryData repData)
    {
        File f;
        String dataFile = Environment.getProperty("jacorb.imr.imr_backup_file");
        boolean write = true;
        
        if (dataFile != null)
        {
            f = new File(dataFile);        
        
            if (!f.exists())
            {
                try
                {                    
                    f.createNewFile();    
                }
                catch (IOException ioe)
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.warn
                        (
                            "Unable to create ImR backup data file " + dataFile,
                            ioe 
                        );
                        write = false;
                    }
                }      
            }
            
            if (write)
            {
                write(f, repData);  
            }
        }
    }
    
    /**
     * Responsible for the writing of the data to the specified file
     */
    private void write(File file, RepositoryData repData)
    {  
        ObjectOutputStream oos = null;
        
        repData.lock.getLock(DataLock.EXCLUSIVE);
              
        try
        {
            oos = new ObjectOutputStream(new FileOutputStream(file));
            oos.writeObject(repData);     
            oos.flush();
            oos.close();
        }
        catch (IOException ioe)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn
                (
                    "Unable to write ImR data to file " + file.getName(),
                    ioe
                );            
            }
        }
        
        repData.lock.releaseLock(DataLock.EXCLUSIVE);
        
        oos = null;
        
    }
}
   
