package org.jacorb.imrutility.imr.util;

import java.util.ArrayList;
import java.util.Iterator;
import org.jacorb.imrutility.imr.Repository;
import org.jacorb.imrutility.imr.RepositoryHelper;
import org.jacorb.orb.CDRInputStream;
import org.jacorb.orb.CDROutputStream;
import org.jacorb.util.Debug;
import org.jacorb.util.Environment;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.ORB;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAPackage.ServantNotActive;
import org.omg.PortableServer.POAPackage.WrongPolicy;

import org.apache.avalon.framework.logger.*;

/**
 * <code>IMRLocator</code> is a utility class used to determine if there
 * are any Implementation Repository instances running anywhere in the domain.
 *
 * @author cj
 * @version 1.0
 */
public class IMRLocator
{
    /**
     * <code>udpPort</code> is the port number on which the MulticastServer will
     * be listening.
     */
    private int udpPort;

    /**
     * <code>imrs</code> is a list of any IMRs that respond to the request.
     */
    private ArrayList imrs;

    /**
     * The ORB
     */
    private static ORB orb;

    /**
     * Logs output
     */
    private Logger logger = Debug.getNamedLogger("jacorb.imr.locator");
    
    /**
     * Construct a new locator. Used by main only.
     *
     * @param orb_ an <code>ORB</code> value. Used to retrieve the POA.
     * @param port_ an <code>int</code> value, the UDP port.
     */
    public IMRLocator (ORB orb_, int port_)
        throws Exception
    {        
        orb  = orb_;

        udpPort = port_;
    }


    /**
     * Construct a new Locator. Used by ORB.
     *
     * @param poa a <code>POA</code> value
     * @param port_ an <code>int</code> value, the UDP port.
     * @exception ServantNotActive if an error occurs
     * @exception WrongPolicy if an error occurs
     */
    public IMRLocator (POA poa, int port_) throws ServantNotActive, WrongPolicy
    {
        udpPort = port_;
    }


    /**
     * <code>main</code> used to list any IMRs running.
     *
     * @param args a <code>String[]</code> value
     */
    public static void main (String [] args)
    {
        IMRLocator locator = null;
        int        udpPort = 0;

        System.setProperty( "jacorb.use_imr", "off" );

        orb = org.omg.CORBA.ORB.init (args, null);
        
        if (args.length > 0)
        {
            if (args[0].equals( "-h" ) || args[0].equals("--help"))
            {
                usage();
            }
            else if (args.length == 2)
            {
                try
                {
                    udpPort = Integer.parseInt (args[1]);
                }
                catch (NumberFormatException nfe)
                {
                    System.out.println("The UDP Port " + args[1] + " is not valid \n");
                    usage();
                }
            }
        }
        
        if (udpPort == 0)
        {
            udpPort = Environment.getIntPropertyWithDefault ("jacorb.imr.udp_port", 0);

            if (udpPort == 0)
            {
                System.out.println("UDP port number == " + udpPort
                    + " - unable to process request\n");
                usage();
            }
        }


        try
        {
            locator = new IMRLocator(orb, udpPort);
            
            locator.findIMRs();

            System.out.println ("Waiting 3 seconds for ImR repsonses");
            Thread.sleep (3000);

            /* If IMRs have responded print out their details */
            if (locator.getIMRs().size() > 0)
            {
                locator.printIMRs();
            }
            else
            {
                System.out.println ("No IMRs found");
            }
        }
        catch (Exception ie)
        {
            System.err.println( "Failed to find IMRs\n");
            ie.printStackTrace();
        }
        
        System.exit (0);
    }


    /**
     * <code>findIMRs</code> is called to make a request to the network using
     * a MulticastClient to find any IMRs.
     *
     * @exception Exception if an error occurs
     */
    public void findIMRs() throws Exception
    {
        byte[] [] returnedData = null;
        MulticastClient mcClient = new MulticastClient(udpPort);;
        byte [] imrData = null;
        CDRInputStream serverResponse = null;
        String name = null;
        Repository imr = null;                
                   
        try
        {        
            /* Set the name to be used when issuing the request */
            CDROutputStream clientRequest = new CDROutputStream();
            clientRequest.beginEncapsulatedArray();
            clientRequest.write_string("ImplementationRepository");

            byte[] inputPacket = clientRequest.getBufferCopy();
            
            clientRequest.close();            
            mcClient.multicastSend(inputPacket);
                 
            /* Give the client time to receive replies*/
            Thread.sleep(2000);
                    
            returnedData = mcClient.getReturnedData();             
                                         
            if (returnedData != null)
            {
                imrs = new ArrayList();
            
                for (int i = 0; i < returnedData.length; i++)
                {
                    imrData = returnedData[i]; 
                                         
                    serverResponse = new CDRInputStream(orb, imrData);
                    serverResponse.openEncapsulatedArray();
                    name = serverResponse.read_string();
                    
                    imr = RepositoryHelper.read(serverResponse);                    
                    imrs.add(imr); 

                    serverResponse.close();                                                                               
                }
            }                
        }
        catch(Exception ex)
        {
            if (logger.isDebugEnabled())
            {   
                logger.error("IMRLocator got an exception finding ImRs", ex);
            }
            
            mcClient.close();        
            throw ex;
        }
        
        mcClient.close();
    }


    /**
     * <code>getIMRs</code> returns the ArrayList containing any IMRs that
     * have been found.
     *
     * @return an <code>ArrayList</code> value, non-null.
     */
    public ArrayList getIMRs()
    {        
        if (imrs == null)
        {
            imrs = new ArrayList();
        }
        return imrs;
    }

    /**
     * Internal method to display details of the IMRs that are running
     * in the domain.  This method will only be used when running the 
     * IMRLocator from the command line
     */
    private void printIMRs()
    {
        Iterator iter = imrs.iterator();
        org.omg.CORBA.Object obj;
        Repository rep;

        System.out.println("IMRs found on: \n" + " #### " + imrs.size());

        while (iter.hasNext())
        {
            try
            {
                obj =  (org.omg.CORBA.Object) iter.next();

                rep = RepositoryHelper.narrow (obj);
                
                org.jacorb.orb.ParsedIOR imrIOR = 
                    ((org.jacorb.orb.Delegate)
                        ((org.omg.CORBA.portable.ObjectImpl)rep)
                            ._get_delegate()).getParsedIOR();

                System.out.println("Running at address - " 
                        + imrIOR.getEffectiveProfile().toString() + "\n");
            }
            catch (Exception ex)
            {
                if (logger.isDebugEnabled())
                {
                    logger.error("Exception getting host name for ImR", ex);
                }
            }
        }
    }

    /**
     * Displays the possible command line properties for this tool
     */
    private static void usage()
    {
        System.out.println (" Usage : To locate any ImRs running the UDP port number must be");
        System.out.println ("         provided either using the command line option below or");
        System.out.println ("         setting the jacorb.imr.udp_port property              ");
        System.out.println ("                                                               ");
        System.out.println ("[ -u | --udp_port ] <port number>: The port number on which the");
        System.out.println ("                                   MultiCast server for the ImR");
        System.out.println ("                                   is running                  ");
        System.out.println ("---------------------------------------------------------------");
        System.out.println ("[ -h | --help ]                    displays this help          ");
        System.out.println ("---------------------------------------------------------------");
        System.exit (0);
    }
}
