package org.jacorb.imrutility.imr.util;


import java.net.MulticastSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.util.Vector;

import org.jacorb.util.Debug;

import org.apache.avalon.framework.logger.Logger;

/**
 * @author cj
 *
 * This class is used to locate processes running anywhere on the network.
 * The specified process will create a <code>MulticastServer</code> giving it
 * a port number on which to run.  The port number is configurable and will be
 * specific to the main process.  When a client sends a request addressed to this
 * port the <code>MulticastServer</code> receives the request and notifies any
 * listeners, i.e. the main process, the main process then returns any required
 * details that are then passed back to the requestor 
 */
public class MulticastServer
{
    /** The main processes that are linked to this <code>MulticastServer</code>*/
    private Vector listeners = new Vector();

    /** The socket for this object */
    private MulticastSocket s = null;

    /** The network address to be used */
    private InetAddress address = null;

    /** controls the polling loop */
    private boolean poll = true;

    /** MCast listener event thread */
    private ServerThread st;

    /**
     * Logs output
     */
    private Logger logger;

    /**
     * Constructor
     * @param port the port number that this <code>MulticastServer</code> is to
     * run on.
     */
    public MulticastServer (int port)
        throws Exception
    {
        logger = Debug.getNamedLogger("jacorb.imr");
        
        /* Create the socket and join the specified group */
        s = new MulticastSocket (port);
        address = InetAddress.getByName ("230.0.0.1");
        s.joinGroup (address);

        st = new ServerThread();
        st.start();
    }

    /**
     * The object that creates this <code>MulticastServer</code> must be added as a
     * listener so when a packet is received the listener can be notified and taken
     * the required action.
     * @param ml the listening
     */
    public void addMulticastListener (MulticastListener ml)
    {
        synchronized (listeners)
        {
            if (!listeners.contains (ml))
            {
                listeners.addElement (ml);
            }
        }
    }

     
    /**
     * Remove the listener
     * @param ml the listener to be removed
     */
    public void removeMulticastListener (MulticastListener ml)
    {
        synchronized (listeners)
        {
            if (listeners.contains (ml))
            {
                listeners.remove (ml);
            }
        }
    }

    /**
     * When finished ( if this can be achieved ) the object that created the
     * <code>MulticastServer</code> should close it.  This allows the socket to
     * remove itself from the group and also to close.
     */
    public void close()
    {
        try
        {
            s.leaveGroup (address);
        }
        catch (Exception ex)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn
                    ("Exception in MulticastServer.close(): " + ex.getMessage());
            }
        }

        poll = false;
        st   = null;
        
        // ### What is the difference between close and disconnect?
        s.close();
        s.disconnect();
    }



    /* Internal thread that runs indefinitely*/
    private class ServerThread extends Thread
    {
        long FIVE_SECONDS = 5000;

        public void run()
        {
            try
            {
                byte [] buf;
                byte [] returnData;
                DatagramPacket packet;
                DatagramPacket returnPacket;
                MulticastEvent me;
                
                /* loop until closed */
                while (poll)
                {
                    returnData = null;
                    returnPacket = null;
                    me = null;
                    
                    buf = new byte [1024];
                    packet = new DatagramPacket (buf, buf.length);

                    s.receive (packet);

                    if (packet.getData().length > 0)
                    {
                        me = new MulticastEvent (packet.getData());
                                                
                        synchronized (listeners)
                        {
                            for (int i = 0; i < listeners.size(); i++)
                            {
                                returnData =
                                ( (MulticastListener)
                                  listeners.get(i)).multicastEvent (me);

                                /* Check that data has been returned to cater
                                 * for potential situations where there may
                                 * be nothing to return
                                 */
                                if (returnData != null)
                                {                                 
                                    returnPacket = 
                                        new DatagramPacket
                                        (
                                            returnData,
                                            returnData.length,
                                            packet.getAddress(),
                                            packet.getPort()
                                        );
                                    
                                    s.send(returnPacket);                                           
                                }
                            }
                        }
                    }
                    
                    sleep ( (long) (Math.random() * FIVE_SECONDS));
                }
            }
            catch (Exception ex)
            {
                if (poll && logger.isDebugEnabled())
                {
                    logger.error
                        ("Exception in MulticastServer$ServerThread.run(): ",
                          ex);
                }
            }
        }
    }
}
