package org.jacorb.test.imrutility.imr;


import junit.framework.*;

import org.jacorb.imrutility.imr.*;

import java.io.*;


public class MultiplesAlivenessTest extends PersistentServerTestCase
{
    private String serverIOR;
    private final String ior = "server";
    private final String home = System.getProperty("user.home");
    private final String separator = System.getProperty("file.separator");
    private RepositoryAdmin rep;
    private Activator act;
    
    private final int FIFO = 2;
    private final int LIFO = 3;
    private final int ROUND_ROBIN = 1;
    private String name;
            
    public MultiplesAlivenessTest(String name, PersistentServerSetup setup)
    {
        super(name, setup);
        this.name = name;
    }

    public static Test suite()
    {
        TestSuite suite = new TestSuite ("Multiples Aliveness Test");
                       
        PersistentServerSetup setup =
            new PersistentServerSetup
            ( 
                suite,
                "org.jacorb.test.imrutility.imr.Server"
            );
 
        suite.addTest(new MultiplesAlivenessTest("testMultiplesAFIFO", setup));
        suite.addTest(new MultiplesAlivenessTest("testMultiplesALIFO", setup));
        suite.addTest(new MultiplesAlivenessTest("testMultiplesARandom", setup));                   
        suite.addTest(new MultiplesAlivenessTest("testMultiplesARoundRobin", setup));   
         
        // Return the setup, not the suite!
        return setup;
    }
 
    protected void tearDown()
    {
        System.err.println("********** tearDown for " + name);
        
        cleanUp();                 
    }
       
    public void testMultiplesAFIFO()
    {
        testMultiples(LoadBalancingPolicy.from_int(2), 2);
    }   
 
    public void testMultiplesALIFO()
    {
        testMultiples(LoadBalancingPolicy.from_int(3), 2);
    }   

    public void testMultiplesARandom()
    {
        testMultiples(LoadBalancingPolicy.from_int(0), 3);
    }

    public void testMultiplesARoundRobin()
    {
        testMultiples(LoadBalancingPolicy.from_int(1), 3);        
    }

    private void testMultiples(LoadBalancingPolicy lbpolicy, int servers)
    {
        String [] args = null;
        ServerName sname = null;
        StringBuffer command = null;
        BufferedReader br = null;
        TestServer test = null;
        TestServer [] tests = new TestServer[servers];
        Process [] procs = new Process[servers];
        Process proc = null;
        String [] iors = new String[servers];
        
        ActivationMode mode = ActivationMode.from_int(0);
                                                        
        try
        {        
            /* Delete all files so no old data hanging around*/
            TestUtility.removeIMRFiles();
            
            /* Wait for files to be removed*/
            Thread.sleep(1000);
        }
        catch (Exception ex)
        {
            /* Don't do anything as this is not a problem*/
        }

        //Start the IMR        
        try
        {                                 
            /* Set fixed port and also aliveness policy of ON_LOOKUP */
            args = new String[2];
            args[0] = "-a";
            args[1] = "1";
            
            rep = TestUtility.startImR(args, null);
            
            assertTrue("Failed to get ImR", rep != null);
            
            sname= new ServerName();
            sname.impl("test_imr");
            
            rep.addImpl
                (new ImplInfo(sname, true, lbpolicy, mode), null);                            
        }
        catch(Exception ex)
        {
            fail("Exception getting ImR" + ex.getMessage());
        }

        /* get command to start server, start then stop server - gets IOR file*/        
        try
        {
            for (int i = 0; i < servers; i++)
            {
                command = setup.getStartCommand();
                command.append(" -i ");
                serverIOR = home + separator + ior + i + ".ior";
                
                iors[i] = serverIOR;
                
                command.append(serverIOR);
                        
                proc = TestUtility.startServer(command);
                        
                procs[i] = proc;
            }
                               
            assertTrue("Server not registered", rep.listServers(sname).length == servers);
        }
        catch (IOException ioe)
        {
            fail("Got IOException starting server " + ioe.getMessage());
        }
        catch (Exception ex)
        {
            fail("Got Exception starting server " + ex.getMessage());
        }   

        for (int i = 0; i < iors.length; i++)
        {
            try
            {        
                br = new BufferedReader(new FileReader(iors[i]));

                // get object reference from command-line argument file
                org.omg.CORBA.Object obj =
                   TestUtility.getImRORB().string_to_object(br.readLine());

                br.close();

                //narrow to correct interface
                test = TestServerHelper.narrow(obj);

                tests[i] = test;    
                           
                test.ping();                     
            }
            catch (Exception ex)
            {
                fail
                    ("Unexpected exception pinging server " 
                        + i + " " + ex.getMessage());      
            }
        }        
                
        switch(lbpolicy.value())
        {
            case FIFO :
            {
                procs[0].destroy();
                break;
            }
            case LIFO :
            {
                procs[procs.length - 1].destroy();
                break;
            }    
            default :
            {
                procs[1].destroy();
                break;
            }
        }                    
                           
        int j = 0;            
                
        /* Try to ping the servers - the ping may or may not fail depending
         * on the load balancing policy
         */        
        while (j < 5)
        {
            for (int i = 0; i < tests.length; i++)
            {                       
                try
                {
                    tests[i].ping();                
                }
                catch (org.omg.CORBA.COMM_FAILURE cf)
                {
                    // Do nothing - a COMM_FAILURE may occur depending on timing
                }                
                catch (Exception ex)
                {
                    fail
                       ("Unexpected exception - failover should be successful " 
                        + ex.getMessage());
                }       
            }
            
            j++;
        }            
       
        switch(lbpolicy.value())
        {
            case FIFO :
            {
                procs[procs.length - 1].destroy();
                break;
            }
            case LIFO :
            {
                procs[0].destroy();
                break;
            }    
            default :
            {
                procs[0].destroy();
                break;
            }
        }                  
    
      
        j = 0;
        
        while (j < 5)
        {
            for (int i = 0; i < tests.length; i++)
            {                       
                try
                {
                    tests[i].ping();                
                }
                catch (Exception ex)
                {
                    switch (lbpolicy.value())
                    {
                        case FIFO :
                        case LIFO :

                        // Do nothing all pings should fail as no servers running 
                        break;
                        
                        default :
                        
                        if (ex instanceof org.omg.CORBA.COMM_FAILURE)
                        {
                        }
                        else
                        {
                            fail("Unexpected exception - should have failed over");
                        }
                        break;
                    }
                }       
            }
            
            j++;
        }          
         
        for (int i = 0; i < procs.length; i++)
        {
            try
            {
                procs[i].destroy();
            }
            catch (Exception ex)
            {
            }            
        }  
        
        try
        {
            Thread.sleep(3000);
            
            while (true)
            {
                tests[0].ping();            
            }
        }
        catch (InterruptedException ie)
        {
        }
        catch (Exception ex)
        {
            // Do nothing - there should be no servers running now    
        }
        
        try
        {            
            assertTrue
            (
                "Servers still registered", 
                rep.listServers(sname).length == 0
            );  
        }
        catch (NotFound nf)
        {
        }

        
        for (int i = 0; i < iors.length; i++)
        {
            try
            {
                TestUtility.removeFile(iors[i]);
            }
            catch (Exception ex)
            {
                // Do nothing
            }
        }
    }
    
    private void cleanUp()
    {
        try
        {         
            if (rep != null)
            {       
                rep.shutdown(false);
            }
        
            TestUtility.removeIMRFiles();
            TestUtility.destroyImR();
            TestUtility.removeActivatorFiles();
            TestUtility.stopActivator();
            Thread.sleep(1000);
        }
        catch (Exception ex)
        {
            /* don't do anything */
        }                   
    }
}
