package org.jacorb.test.poa;

/*
 *        JacORB  - a free Java ORB
 *
 *   Copyright (C) 1997-2001  Gerald Brose.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.Properties;
import junit.framework.Test;
import junit.framework.TestSuite;
import org.jacorb.test.common.ClientServerSetup;
import org.jacorb.test.common.ClientServerTestCase;
import org.jacorb.test.BasicServer;
import org.jacorb.test.BasicServerHelper;
import org.omg.CORBA.*;
import org.omg.PortableServer.*;


/**
 * <code>BiDirDelegateTest</code> is a test for JAC5; BiDir connections get
 * BAD_INV_ORDER/Delegate not set.
 *
 * @author <a href="mailto:rnc@prismtechnologies.com"></a>
 * @version 1.0
 */
public class BiDirDelegateTest extends ClientServerTestCase
{
    private BasicServer server;

    /**
     * <code>ping</code> is used to denote whether the ping thread
     * successfully completed.
     */
    private static boolean ping = false;


    public BiDirDelegateTest(String name, ClientServerSetup setup)
    {
        super(name, setup);
    }

    public void setUp() throws Exception
    {
        server = BasicServerHelper.narrow( setup.getServerObject() );
    }


    public static Test suite()
    {
        TestSuite suite = new TestSuite( "Basic client/server tests" );

        Properties client_props = new Properties();
        Properties server_props = new Properties();

        client_props.setProperty ("org.omg.PortableInterceptor.ORBInitializerClass.bidir_init", "org.jacorb.orb.giop.BiDirConnectionInitializer");
        server_props.setProperty ("org.omg.PortableInterceptor.ORBInitializerClass.bidir_init", "org.jacorb.orb.giop.BiDirConnectionInitializer");

        // Note that I have activated BiDir connections through the above
        // property and am using my own ClientServerSetup class.
        BiDirDelegateClientServerSetup setup =
            new BiDirDelegateClientServerSetup( suite,
                                   "org.jacorb.test.orb.BasicServerImpl",
                                   client_props,
                                   server_props);

        suite.addTest( new BiDirDelegateTest( "test_ping", setup ));

        return setup;
    }

    /**
     * <code>test_ping</code> calls the ping method on BasicServer.
     *
     * We have to use a thread as otherwise if the test fails the harness will
     * hang. This allows the test to detect a broken server.
     */
    public void test_ping()
    {
        Thread thread = new Thread("BiDirDelegateTest Ping Thread")
        {
            public void run()
            {
                server.ping();
                ping = true;
            }

        };
        thread.start();
        try
        {
            thread.join(3000);
        }
        catch (InterruptedException e) {}

        if (ping == false)
        {
            fail("Did not manage to complete ping call");
        }
    }


    /**
     * <code>main</code> is reimplemented here so that we can start
     * the server with a child POA and use servant_to_id.
     *
     * @param args a <code>String[]</code> value
     */
    public static void main (String[] args)
    {
        try
        {
            //init ORB
            ORB orb = ORB.init( args, null );

            //init POA
            POA rootPoa =
                POAHelper.narrow( orb.resolve_initial_references( "RootPOA" ));
            rootPoa.the_POAManager().activate();

            // create POA
            Policy policies[] = new Policy[2];
            policies[0] = rootPoa.create_id_uniqueness_policy(
                org.omg.PortableServer.IdUniquenessPolicyValue.MULTIPLE_ID);
            policies[1] = rootPoa.create_implicit_activation_policy(
                ImplicitActivationPolicyValue.IMPLICIT_ACTIVATION );

            POA poa = rootPoa.create_POA
                ("childPOA1", rootPoa.the_POAManager(), policies);
            poa.the_POAManager().activate();

            String className = args[0];
            Class servantClass = Class.forName (className);
            Servant servant = ( Servant ) servantClass.newInstance();

            // Get the id
            byte []oid = poa.servant_to_id (servant);

            // create the object reference
            org.omg.CORBA.Object obj = poa.id_to_reference (oid);

            System.out.println ("SERVER IOR: " + orb.object_to_string(obj));
            System.out.flush();

            System.out.println ("Entering ORB event loop for BiDirDelegateTest");

            // wait for requests
            orb.run();
        }
        catch( Exception e )
        {
            e.printStackTrace();
            System.out.println ("Caught error " + e);
        }
    }


    /**
     * <code>BiDirDelegateClientServerSetup</code> overrides ClientServerSetup
     * so that we can provide a different main.
     */
    public static class BiDirDelegateClientServerSetup extends ClientServerSetup
    {
        public BiDirDelegateClientServerSetup ( Test test, String servantName )
        {
            super(test, servantName);
        }

        public BiDirDelegateClientServerSetup( Test test,
                                               String servantName,
                                               Properties clientOrbProperties,
                                               Properties serverOrbProperties )
        {
            super(test, servantName, clientOrbProperties, serverOrbProperties);
        }

        /**
         * <code>getTestServerMain</code> overrides getTestServerMain in common.
         * This is so we can provide our own, using a child poa and servant_to_id.
         *
         * @return a <code>String</code> value
         */
        public String getTestServerMain()
        {
            return "org.jacorb.test.poa.BiDirDelegateTest";
        }
    }
}
