package org.jacorb.imrutility.imr;

// Imports
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintWriter;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;
import org.jacorb.imrutility.imr.policy.LBPolicy;
import org.jacorb.imrutility.imr.persistence.FilePersistence;
import org.jacorb.imrutility.imr.persistence.RepositoryData;
import org.jacorb.imrutility.imr.util.IMRLocator;
import org.jacorb.imrutility.imr.util.MulticastEvent;
import org.jacorb.imrutility.imr.util.MulticastListener;
import org.jacorb.imrutility.imr.util.MulticastServer;
import org.jacorb.orb.CDRInputStream;
import org.jacorb.orb.CDROutputStream;
import org.jacorb.orb.ParsedIOR;
import org.jacorb.util.Debug;
import org.jacorb.util.Environment;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.LongHolder;
import org.omg.PortableServer.*;

import org.apache.avalon.framework.logger.*;

/**
 * @author cj
 *
 * This class holds details of persistent POAs.  The intention is that
 * POAs from any ORB will be able to register with this ImR.  To cater
 * for this eventuality the Repository has been written to handle this.
 * Implementation information such as multiplicity and Load Balancing 
 * policy can be pre-registered.  The information will be registered 
 * against an impl name or POA names but not both.  When a POA is 
 * registered it may be registered with an impl name and POA names or it 
 * may be the POA names alone. 
 */
public class RepositoryImpl extends RepositoryAdminPOA
    implements MulticastListener
{
    /** Holds repository data in memory */
    protected RepositoryData repData = null;

    /** The ORB */
    protected static org.omg.CORBA.ORB orb;

    /** The location of the ior file*/
    protected static String iorfile;

    /** If true the persisted data is not read in at start up */    
    protected static boolean newFile = false;

    /** Holds the time of the last activity for each active server */
    protected HashMap activityTable = new HashMap();

    /** The length of time a server can be inactive */
    private int activeTimeout = 120000;

    /** Flag indicating whether auto registration is allowed */
    private boolean allowAuto = false;

    /** Flag indicating whether the IMR is to check if the server is alive */
    private boolean checkAlive = false;

    /** flag to indicate whether other ImRs should be checked for registered 
     * servers
     */
    private boolean checkOtherImRs = true;

    /** The policy to be applied in pinging registered servers*/
    private AlivenessPolicy aliveness;

    /** The id of this ImR */
    private long id;

    /** MS server */
    private MulticastServer ms;

    /** Mapping of impl/poa to an Activator ior*/
    private HashMap activators = new HashMap();
    
    /** Mapping of a server by id to server details*/
    private HashMap runningServers = new HashMap();

    /** Mapping of poa to registration and server id*/
    private HashMap runningPOAs = new HashMap();

    /** 
     * Timer to handle scheduling of tasks that ping and check server activity
     */
    private Timer timer;

    /** Flag to indicate whether a PingingTask has been scheduled */
    private boolean pinging = false;

    /** The interval to be applied if the aliveness policy is PING*/
    private int heartbeat;

    /** The port that the MulticastServer will listen on*/
    private int udpPort = 0;

    /**
     * The endpoints for the ImR which replace the servers endpoints when
     * it is registered
     */
    private EndpointProfile[] imrEndpoints;

    /** 
     * Maximum number of pings that can fail before assuming the server is
     * no longer running
     */
    private int maxPings = 0;

    /**
     * A thread that writes to persistence when changes are made to the registered
     * data
     */
    private Write writeThread;

    /** Flag to indicate that there are further writes to be made*/   
    private boolean writePending;

    /** Shuts down the ImR */
    private Shutdown shutdownThread;

    /** Logs general ImR information*/    
    private Logger logger;
    
    /** Logs changes to ImR state e.g. registrations/removals*/
    private Logger stateLogger;
    
    /** Logs location request information*/
    private Logger locateLogger; 

    /** Used to generate a long to be used as the id of a Server*/
    private Random idGenerator = new Random();
        
    /**
     * Object reference to this Repository
     */
    private org.omg.CORBA.Object imrObj;
        
    /**
     * Persistence object - change when other persistence mechanisms introduced
     */    
    private FilePersistence persistence = new FilePersistence();
    
    /**
     * Constructor
     */
    public RepositoryImpl() throws Exception
    {
        logger = Debug.getNamedLogger("jacorb.imr");
        stateLogger = Debug.getNamedLogger("jacorb.imr.state");
        locateLogger = Debug.getNamedLogger("jacorb.imr.locate");        
        
        init();
        
        // Prevent attempts to register with ourself
        Environment.isImR(true);

        org.jacorb.poa.POA rootPOA = (org.jacorb.poa.POA) POAHelper.narrow
           (orb.resolve_initial_references("RootPOA"));

        org.omg.CORBA.Policy[] policies = new org.omg.CORBA.Policy[2];

        policies[0] =
            rootPOA.create_lifespan_policy(LifespanPolicyValue.PERSISTENT);

        policies[1] =
            rootPOA.create_id_assignment_policy(
                IdAssignmentPolicyValue.SYSTEM_ID);

        POA imrPOA =
            rootPOA.create_POA("IMRPOA", rootPOA.the_POAManager(), policies);

        for (int i = 0; i < policies.length; i++)
        {
            policies[i].destroy();
        }

        imrPOA.activate_object(this);

        
        PrintWriter pw =
            new PrintWriter(new FileOutputStream(new File(iorfile)));

        imrObj = imrPOA.servant_to_reference(this);

        rootPOA.the_POAManager().activate();
        
        String objString = orb.object_to_string(imrObj);
        
        ( (org.jacorb.orb.ORB) orb).addObjectKey("ImplementationRepository", objString);
        
        pw.println(objString);

        pw.flush();
        pw.close();

        imrEndpoints = rootPOA.getEndpoints();

        orb.run();
    }

    /**
     * This method initialises the variables used by the ImR
     */
    protected void init() throws Exception
    {
        String temp = null;
        
        timer = new Timer();

        shutdownThread = new Shutdown();
        shutdownThread.setDaemon(true);
        shutdownThread.setName("IMR Shutdown Thread");
        
        Runtime.getRuntime().addShutdownHook(shutdownThread);

        writeThread = new Write();
        writeThread.setName ("IMR Write Thread");
        writeThread.setDaemon (true);

        AlivenessPolicy ap =
            AlivenessPolicy.from_int(Environment.getAlivenessPolicy());

        if (ap.value() == AlivenessPolicy._PING)
        {
            temp = Environment.getProperty("jacorb.imr.heartbeat");

            if (temp == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.error(
                        "AlivenessPolicy is PING but heartbeat is not set");
                }

                throw new INTERNAL(
                    "AlivenessPolicy is PING but heartbeat is not set");
            }
            else
            {
                heartbeat = Integer.parseInt(temp);
            }
        }
        else
        {
            if (ap.value() == AlivenessPolicy._ON_EXPIRY)
            {
                temp =
                    Environment.getProperty(
                        "jacorb.imr.active_server_timeout");

                if (temp != null && temp.length() > 0)
                {
                    activeTimeout = Integer.parseInt(temp);
                }
            }
        }

        aliveness(ap);

        temp = Environment.getProperty("jacorb.imr.allow_auto_register");

        if (temp != null)
        {
            allowAuto = temp.equalsIgnoreCase("on");
        }

        udpPort =
            Environment.getIntPropertyWithDefault("jacorb.imr.udp_port", 0);

        if (udpPort > 0)
        {
            initMulticast();
        }
               
                 
        if (!newFile)
        {
            populateData();        

            /* After the data has been populated we may have discarded some
             * of the details so write the file out with the new details
             */
            writeThread.start ();
            
            try
            {
                synchronized (writeThread)
                {
                    writeThread.notifyAll ();
                }
            }
            catch (Exception e)
            {                
            }            
        }
        else
        {
            repData = new RepositoryData();            
            writeThread.start();
        }
        
        temp = null;
        ap = null;
    }

    /**
     * Instigates the population of data from persistence
     */
    private void populateData()
    {
        populateServers();
        populatePOAs();
        populateActivators();
    }
    
    /**
     * Populate server information from persistence
     */
    private void populateServers()
    {
        Iterator iter;
        ServerRecord srecord;
        RunningServer rserver;
        POARecord precord;
        RunningPOA rpoa;
        Server srv;
        Map.Entry entry;
        Long sID;
        long lid;
        ArrayList alist = null;
               
        repData = persistence.read();
        
        Map servers = repData.getServers();
   
        iter = servers.entrySet().iterator();
            
        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();
            sID = (Long) entry.getKey();
            srecord = (ServerRecord) entry.getValue();
            
            srv = ServerHelper.narrow(orb.string_to_object(srecord.srv));
       
            try
            {
                /* Pinging of the server - should this be configurable*/
                srv.ping(0);
               
                rserver = 
                    new RunningServer
                    (
                        srv, 
                        System.currentTimeMillis(),
                        srecord.pnKey.getImplName()
                    );
                    
                runningServers.put(sID, rserver);
                activityTable.put(sID, new Long(System.currentTimeMillis()));
            }
            catch (Exception ex)
            {                
                if (alist == null)
                {
                    alist = new ArrayList();
                }
                
                alist.add(sID);
                break;
            }

            lid = sID.longValue();
                        
        }
        
        /**
         * If there are servers to be removed remove them now
         */
        if (alist != null)
        {
            repData.removeServers(alist);
        }      
    }
    
    /**
     * Populate POA information from persistence
     */
    public void populatePOAs()
    {
        Iterator iter;
        Iterator siter;
        POARecord prec;
        Map.Entry entry;
        POANameKey pnkey = null;
        HashSet set;
        ArrayList pset = null;
        RunningPOA rpoa;
        RegistrationInfo reg;
        RunningServer rsrv;
        EndpointProfile [] endpoints;
        ArrayList alist = null;
        
        Map poas = repData.getPOAs();
        
        iter = poas.entrySet().iterator();
                
        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();
            pnkey = (POANameKey) entry.getKey();
            set = (HashSet) entry.getValue();
            pset = null;    
            
            siter = set.iterator();
            
            while (siter.hasNext())
            {
                prec = (POARecord) siter.next();
                rsrv = (RunningServer) runningServers.get(new Long(prec.id));
                
                if (rsrv != null)
                {
                    endpoints = prec.endpoints;
                    
                    if (endpoints != null)
                    {
                        reg = new RegistrationInfo
                        (
                            pnkey.getImplName(), 
                            endpoints, 
                            rsrv.srv
                        );
                    
                        rpoa = new RunningPOA(reg, prec.id);
                    
                        if (pset == null)
                        {
                            pset = new ArrayList();
                        }        
                    
                        pset.add(rpoa);
                    }
                }                
            }                              
        
            if (pset != null)
            {                
                runningPOAs.put(pnkey, pset);
            }
            else
            {
                if (alist == null)
                {
                    alist = new ArrayList();
                }
                
                alist.add(pnkey);
            }            
        }

        if (alist != null)
        {
            repData.removePOAs(alist);
        }       
    } 
                
    /**
     * Populate Activator details from persistence
     */
    public void populateActivators()
    {
        Map.Entry entry;
        Iterator iter;    
        Map acts = repData.getActivators();
        Long key;
        Activator act;
        String actIOR;
        Map map = null;
                
        iter = acts.entrySet().iterator();
        
        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();
            key = (Long) entry.getKey();
            actIOR = (String) entry.getValue();    
                   
            try
            {
                act = ActivatorHelper.narrow
                    (orb.string_to_object(actIOR));
                activators.put(key, act);
            }
            catch (Exception ex)
            {
                if (map == null)
                {
                    map = new HashMap();
                }
                
                map.put(key, actIOR);
            }                                                                    
        }
        
        if (map != null)
        {
            repData.removeActivators(map);
        }           
    }
    
    /**
     * Method to initialise the <code>MulticastServer</code> object for
     * this IMR.
     */
    private void initMulticast()
    {
        try
        {
            ms = new MulticastServer(udpPort);
            ms.addMulticastListener(this);
        }
        catch (Exception ex)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn("Failed to initialise MulticastServer", ex);
            }
        }
    }

    /* RepositoryOperations */

    /**
     * Returns the timeout to be used when the Aliveness policy is ON_EXPIRY
     */
    public int activeTimeout()
    {
        return activeTimeout;
    }

    /**
     * Sets the timeout to be used when the Aliveness policy is ON_EXPIRY
     */
    public void activeTimeout(int timeout)
    {
        activeTimeout = timeout;
    }

    /**
     * Returns the value of the flag indicating whether auto registration is
     * allowed
     */
    public boolean allowAuto()
    {
        return allowAuto;
    }

    /**
     * Sets the flag that indicates whether auto registration is allowed
     */
    public void allowAuto(boolean auto)
    {
        allowAuto = auto;
    }

    /**
     * Flag indicating whether the ImR is to check whether a server is alive 
     * ON_LOOKUP
     */
    public boolean checkAlive()
    {
        return checkAlive;
    }

    /**
     * Sets the flag indicating whether the ImR is to check if the server 
     * is alive
     */
    public void checkAlive(boolean check)
    {
        checkAlive = check;
    }

    /**
     * Removes an implementation.  For the initial release there will only
     * be 1 Activator per Implementation and the link to the Activator is
     * in the ImplRecord which is removed.  The Activator cannot be 
     * removed as it may be linked to other Implementations.
     * @param name the name of the implementation to be removed
     * @throws ImplActive if the specified implementation is still active
     * @throws NotFound if the specified implementation is not registered
     */
    public void removeImpl(ServerName name, Activator act)
    throws ImplActive, NotFound
    {
        POANameKey[] pnKey = POANameKey.serverNameToKeys(name);
        ImplRecord irec;
        Iterator iter;
        RunningServer rsrv;
        Activator regAct;
                
        /* For each key in the name check whether there is a POA running.  The 
         * POA key is made up of the POA names and the impl name if the latter
         * exists.  The ServerName will only be the implname OR the POA names
         * so the key may not be found in the runningPOAs
         */
        for (int i = 0; i < pnKey.length; i++)
        {
            if (runningPOAs.containsKey(pnKey[i]))
            {
                throw new ImplActive();
            }            
            else
            {
                /* If the server has an impl name */
                if (name.discriminator())
                {
                    synchronized (runningServers)
                    {
                        iter = runningServers.values().iterator();
                    
                        while (iter.hasNext())
                        {
                            rsrv = (RunningServer) iter.next();
                        
                            if (rsrv.impl.equals(name.impl()))
                            {
                                throw new ImplActive();
                            }
                        }                       
                    }
                }
            }

            irec = (ImplRecord) repData.removeImpl(pnKey[i]);

            if (irec == null)
            {
                throw new NotFound
                    (pnKey[i].toString() + " not found");
            }
            
            if (stateLogger.isDebugEnabled())
            {
                stateLogger.debug("IMR :: removing poa " + pnKey[i]);
            }            
        }

        writePending = true;    
        
        try
        {
            synchronized (writeThread)
            {
                writeThread.notifyAll();
            }
        }
        catch (Exception e)
        {                
        }
    }

    /**
     * Adds an implementation to the repository.  This method is used to 
     * pre-register servers.  The registration may be for a jacORB object
     * in which case and implementation name is provided or it may be for
     * a foreign object in which case the name may be made up of the POA
     * names. It will never be made up of both.
     * @param <code>info</code> the <code>ImplInfo</code> to be registered
     * @param <code>act</code> the <code>Activator</code> for this implementation
     * @throws AlreadyRegistered if already registered
     */
    public void addImpl(ImplInfo info, Activator act) throws AlreadyRegistered
    {
        /* This will result in a key made up of either an impl name or a number
         * of keys made up from POA Names but not both impl and poa names
         */
        POANameKey[] pnKeys = POANameKey.serverNameToKeys(info.name);

        ImplRecord irec = 
            new ImplRecord
            (
                info.multiples, 
                info.lbPolicy.value(), 
                info.activation.value()
            );
         
        for (int i = 0; i < pnKeys.length; i++)
        {
            if (repData.getImpl(pnKeys[i]) != null)
            {
                throw new AlreadyRegistered(info.name);
            }
                        
            if (act != null)
            {                 
                long aid = 0;
              
                aid = act.id();
                
                /* If the Activator doesn't have an ID get it one*/
                if (aid == 0)
                {
                    aid = nextID();                    
                }
                
                act.id(aid);
                irec.activatorID = aid;                                   
                
                Long actID = new Long(aid);
                                
                synchronized (activators)
                {
                    activators.put(actID, act);
                }

                repData.addActivator
                (
                    actID,
                    orb.object_to_string(act)
                );
            }
            
            repData.addImpl
            (
                pnKeys[i],
                irec
            );
        }
        
        writePending = true;
         
        /* Only write when all keys have been added*/           
        try
        {
            synchronized (writeThread)
            {
                writeThread.notifyAll();
            }
        }
        catch (Exception e)
        {                
        }
    }

    /**
     * Deregisters all poas for a given server
     * @id the id of the server
     * @throws NotFound if the server is not found
     */
    public void deregisterAll(long lid) throws NotFound
    {
        Map.Entry entry;
        Iterator iter;
        Iterator siter;
        ArrayList activeServers;
        POANameKey pnkey;
        RunningPOA rpoa;
        boolean found = false;
        
        iter = runningPOAs.entrySet().iterator();
        
        synchronized (runningPOAs)
        {
           while (iter.hasNext())
           {
               entry = (Map.Entry) iter.next();
               pnkey = (POANameKey) entry.getKey();
               activeServers = (ArrayList) ( (ArrayList) entry.getValue()).clone();
                
               siter = activeServers.iterator();
            
               while (siter.hasNext() && !found)
               {
                   rpoa = (RunningPOA) siter.next();
                

                   if (rpoa.id == lid)
                   {
                       siter.remove();
                       found = true;
                   }
               }
               
               if (found)
               {
                   runningPOAs.put(pnkey, activeServers);
                   
               }
           }                    
        }        
        
        synchronized (runningServers)
        {            
            if (runningServers.remove(new Long(lid)) != null)
            {
                found = true;    
            }
        }

        if (found)
        {
            writePending = true;
  
            repData.removeServer(lid);
         
            try
            {
                synchronized (writeThread)
                {
                    writeThread.notifyAll();
                }
            }
            catch (Exception e)
            {                
            }
        }
        else
        {
            throw new NotFound();
        }
    }

    /**
     * Method that locates the server for the given object.  If the object is
     * not found then null is returned.  If the server is located then the
     * object is updated with the new host and port and is returned.
     * @param obj the object to be located
     * @return org.omg.CORBA.Object the object if located else null.
     */
    public EndpointProfile[] locate(String impl, String[] poas)
    {
        ArrayList activeServers = null;
        EndpointProfile[] endpoints = null;
        long oldID = 0;
        boolean startProcess = false;
        boolean swap = false;
        POANameKey pnKey;
        RunningPOA prec = null;
        Map.Entry map;
        boolean tryStart = false;
        
        pnKey = new POANameKey(impl, poas);

        synchronized (runningPOAs)
        {
            activeServers = (ArrayList) runningPOAs.get(pnKey);
        }

        if (activeServers != null && activeServers.size() > 0)
        {
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug
                (
                    "Found "
                    + activeServers.size()
                    + " servers for "
                    + pnKey.toString()
                );
            }

            /*
             * Determine which server to use applying load balancing policy
             * if multiples are found
             */
            if (activeServers.size() == 1)
            {
                prec = (RunningPOA) activeServers.get(0);
                oldID = prec.id;
                
                if (checkAlive())
                {
                    try
                    {
                        ( (RunningServer) runningServers.get
                            (new Long(prec.id))).srv.ping(0);                            
                        
                        swap = true;
                    }
                    catch (Exception ex)
                    {                        
                        
                        // Remove the server details as the server wasn't running
                        // and has not restarted - in a Thread maybe?
                        try
                        {
                            deregisterAll(oldID);
                        }
                        catch (NotFound nf)
                        {
                            /* shouldn't get here but we have to catch the 
                             * exception
                             */
                        }

                        /* If checkOtherImRs is true then this is not a locate
                         * request from another ImR and we can consider 
                         * attempting to start the server
                         */
                        if (checkOtherImRs)
                        {
                            tryStart = true;
                        }
                    }
                }
                else
                {
                    /* If the server is in the running table but we aren't 
                     * configured to ping on lookup we can only assume that 
                     * it is still running so return the endpoints
                     */
                    swap = true;
                }
            }
            else
            {
                try
                {
                    prec = applyLoadBalancing(activeServers, pnKey);
                    swap = true;
                }
                catch (StartupFailed sf)
                {
                    if (locateLogger.isDebugEnabled())
                    {
                        locateLogger.debug(sf.reason);
                    }
                }
                catch (Exception ex)
                {
                    /* Do nothing - other ImRs will be called if the required
                     * object is not found by this ImR
                     */
                }
            }
        }
        else
        {
            /* If there are no servers registered then we can try starting
             * an instance if the Activator is running, the implementation is
             * known and can be auto activated
             */
             tryStart = true;
        }


        /* If we haven't found a server on this ImR try other ImRs - if the
         * locate request didn't come from another ImR in the first place. 
         * This is a bit chicken and egg - do we try other ImRs before we
         * try to start a process or vice versa
         */
        if (!swap && checkOtherImRs)
        {
            /* If we haven't found an active server see if any other IMR 
             * has one 
             */
            endpoints = checkOtherImRs(impl, poas);
        }

        /* If we haven't found a local server or a remote server and earlier
         * processing indicates we should try to start a server e.g. no servers
         * recorded locally or only 1 and it's not running
         */               
        if (!swap && endpoints == null && tryStart)
        {
            /* Configured to ping on lookup but the ping failed so
             * try to start the process
             */
            try
            {
                prec = startProcess(pnKey);
            }
            catch (Exception ex2)
            {
                //Do nothing - if start up fails then we remove the 
                //server details, 
            }

            if (prec != null)
            {
                swap = true;
            }
        }

        /* If swap is true then we have found an active server */
        if (swap)
        {
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug("Found Server id " 
                    + prec.id + " - getting endpoints");
            }

            endpoints = prec.reg.endpoints;
        }


        return endpoints;
    }

    /**
     * This method is essentially a request from  another ImR that has not found
     * the required object registered locally.  A flag is set to prevent the main
     * locate method from trying other ImRs if it doesn't find the required object.
     * @param <code>obj</code> the object to be located.
     * @return the result of the locate request - could be null
     */
    public EndpointProfile[] imrLocateRequest(String impl, String[] poas)
    {
        if (locateLogger.isDebugEnabled())
        {
            locateLogger.debug("Received request from other IMR");
            
        }
        /* Set flag to false because we don't want this ImR to try other ImRs 
         * if it is another ImR that has made the initial locate request
         */
        checkOtherImRs = false;

        EndpointProfile[] endpoints = locate(impl, poas);

        /*
         * Set flag back to true so that any local "locate" request try other 
         * ImRs if the initial search is unsuccessful
         */
        checkOtherImRs = true;

        return endpoints;
    }

    /* RegistryOperations */

    /**
     * This method is called when a persistent POA is activated on a server.  If
     * the server has not already been registerd and the allowAuto flag is not
     * true a NotFound exception is thrown.  If the autoAllow flag is true an
     * implementation is registered for the server.  If this is a new
     * registration the server information is passed to the Repository in the
     * Registration parameter.  If the server is already registered the
     * Registration attribute is empty.
     * @param id the unique id allocated to the server
     * @param poas a list of poa names associated with the server
     * @param reg the Registration object containing Server information
     * @throws NotFound if autoAllow is not true and the server is not already
     * registered
     * @throws ImplActive if an active server exists and multiples are not allowed
     */
    public void registerPOA
    (
        LongHolder lid,
        String[] poas,
        Registration reg,
        EndpointProfileSeqHolder endpoints
    )
    throws NotFound, ImplActive, AlreadyRegistered
    {
        POANameKey pnKey = null;
        RegistrationInfo regInfo = null;
        ImplRecord irec = null;
        ArrayList regServers = null;
        POANameKey ipnKey = null;
        POARecord prec;
        RunningPOA rpoa;
        Map.Entry entry;
        RunningServer rsrv;
        boolean registerNew = true;
        
        long longID = lid.value;
                                        
        endpoints.value = imrEndpoints;
        
        /* If the id is populated then this may be a re-registration.  We 
         * cannot rely on the reg.discriminator alone because the registration 
         * could be populated if it is a re-registration and the id will 
         * also be > 0.  If the id is populated but there is no registration
         * info then we need to look for the registration info in the running
         * poas
         */
        if (!reg.discriminator())  // Registration info has been provided 
        {
            pnKey = new POANameKey(reg.reg().impl, poas);
            regInfo = reg.reg();
        }
        else if (longID != -1 && longID != 0)
        {            
            rsrv = (RunningServer) runningServers.get(lid);
            
            /* RunningServer found */
            if (rsrv != null)
            {
                 pnKey = new POANameKey(rsrv.impl, poas);                    
            }
            else
            {
                /* Create key using only the poas as we have not found an impl name*/
                pnKey = new POANameKey("", poas);
            }
                          
            regServers = (ArrayList) runningPOAs.get(pnKey);
                
            if (regServers != null)
            {
                Iterator iter = regServers.iterator();

                while (iter.hasNext())
                {
                    rpoa = (RunningPOA) iter.next();

                    if (rpoa.id == lid.value)
                    {
                        regInfo = rpoa.reg;
                        break;
                    }
                }                                                          
            }                
        }    

        
        if (regInfo == null)
        {
            throw new NotFound
            (
                "Registration details not held for "
                + pnKey.toString()
            );
        }
               
        if (stateLogger.isDebugEnabled())
        {
            stateLogger.debug
                ("RegisterPOA and registering " + pnKey.toString());
        }

        /* Get the ImplRecord details using a simplified key - the key for
         * activators and impl information will consist only of an implname OR
         * POA Names but not both
         */
        ipnKey = getIPOANameKey(pnKey);
 
        irec = repData.getImpl(ipnKey);

        if (irec == null) // if Server implementation not already registered
        {                       
            if (allowAuto) // if auto registration allowed
            {
                lid.value = 
                    internalRegisterPOA
                    (
                        ipnKey, 
                        pnKey, 
                        regInfo, 
                        0, 
                        0,
                        longID
                    );
            }
            else // auto registration not allowed so can't continue
            {
                throw new NotFound
                (
                    "Implementation not found and not auto allowed: "
                    + pnKey.toString()
                );
            }
        }
        else // server implementation is registered
        {            
            if (regServers == null)
            {
                /* Get a list of server ids registered under this name */
                regServers = (ArrayList) runningPOAs.get(pnKey);
            }

            if (regServers != null
                    && regServers.size() > 0) // at least one server is registered
            {
                if (irec.multiples) // multiples allowed so register
                {
                    /* Register using the key under which the servers are registered*/
                    lid.value =
                        internalRegisterPOA
                        (
                            ipnKey,
                            pnKey,
                            regInfo,
                            irec.lb_policy,
                            irec.mode,
                            longID
                        );
                }
                else // multiples not allowed
                {                    
                    /* If multiples are not allowed there should only be 1*/
                    rpoa = (RunningPOA) regServers.get(0);

                    /* check to see whether the server registered is actually still 
                     * active.  If it is no longer active then we
                     * can simply register this server by overwriting the inactive 
                     * server
                     */
                    boolean remap = false;

                    /* If a server is re-registering because an ImR failed and the
                     * id's are the same then it's the same server - this may need
                     * to change at the next phase when all servers are registered
                     * with all ImRs.
                     */
                    if (longID == rpoa.id)
                    {
                        remap = true;
                    }                  
                    else
                    {
                        try
                        {
                            ((RunningPOA) regServers.get(0)).reg.srv.ping(0);
                        }
                        catch (Exception ex)
                        {
                            /* This assumes that the server is not active
                             * so remap is set to true to allow a new 
                             * instance to be started.  This caters for 
                             * situations where a server is still registered
                             * but is no longer running and we want to 
                             * start another instance
                             */
                            deregisterAll(((RunningPOA) regServers.get(0)).id);
                                
                            remap = true;
                        }
                    }

                    if (remap == false)
                        // not the same server and registered server is active
                    {
                        throw new ImplActive
                        (
                            "Implementation already active: Duplicates not allowed: "
                            + pnKey.toString()
                        );
                    }

                    /* If we get this far we can safely register the server 
                     * with the new details 
                     */
                    lid.value =
                        internalRegisterPOA
                        (
                            ipnKey,
                            pnKey,
                            regInfo,
                            irec.lb_policy,
                            irec.mode,
                            longID
                        );
                }
            }
            else // no servers registered so just register this one
            {
                lid.value =
                    internalRegisterPOA
                    (
                        ipnKey,
                        pnKey,
                        regInfo,
                        irec.lb_policy,
                        irec.mode,
                        longID
                    );
            }
        }
    }

    /**
     * Deregisters POA for a given server
     * @param id the id of the server
     * @param poas the names of the poas to be deregistered
     * @throws NotFound if the server is not found
     */
    public void deregisterPOA(long lid, String[] poas) throws NotFound
    {
        Map.Entry entry;
        Iterator iter;
        ArrayList activeServers;      
        RunningPOA rpoa;
        POANameKey pnkey;
        
        RunningServer rsrv = (RunningServer) runningServers.get(new Long(lid));
        
        if (rsrv != null)
        {
           pnkey = new POANameKey(rsrv.impl, poas);
        }
        else
        {
            /* If the server is not in the running servers table then
             * it shouldn't haave any references anywhere else
             */
            throw new NotFound();
        }

        activeServers = (ArrayList) runningPOAs.get(pnkey);
        
        if (activeServers == null)
        {
            throw new NotFound();
        }    
                        
        iter = activeServers.iterator();
        
        synchronized (runningPOAs)
        {    
            while (iter.hasNext())
            {
                rpoa = (RunningPOA) iter.next();
                
                if (rpoa.id == lid)
                {
                    iter.remove();
                }
            }
        }                    

        writePending = true;
        
        repData.removePOA(lid, pnkey);
        
        try
        {
            synchronized (writeThread)
            {
                writeThread.notifyAll();
            }
        }
        catch (Exception e)
        {                
        }
    }

    /**
     * This method is used to obtain the next available ID when a server is
     * not already registered
     * @return long the next id
     */
    public long nextID()
    {
        return idGenerator.nextLong();
    }

    public AlivenessPolicy aliveness()
    {
        return aliveness;
    }

    /**
     * Sets the <code>aliveness</code> variable that indicates when
     * the IMR should check whether servers are still alive.next
     *
     * PING      - the IMR should PING active servers at the specified
     *             heartbeat
     * ON_EXPIRY - the IMR should PING active servers when the activity
     *              timeout expires.
     * ON_LOOKUP - the IMR should PING the server when a locate request
     *             is received.
     *
     * @see AlivenessPolicy
     */
    public void aliveness(AlivenessPolicy policy)
    {
        aliveness = policy;

        if (aliveness.value() == AlivenessPolicy._PING && heartbeat > 0)
        {
            if (!pinging)
            {
                timer.schedule(new PingTask(), 0, heartbeat);
                pinging = true;
            }
        }
        else
            if (aliveness.value() == AlivenessPolicy._ON_EXPIRY)
            {
                timer.schedule(new ActivityTask(), 0, activeTimeout);
            }
            else
                if (aliveness.value() == AlivenessPolicy._ON_LOOKUP)
                {
                    checkAlive(true);
                }
    }

    /* ServerOperations */
 
    /**
     * Shuts the ImR down
     * @boolean <code>wait</code> whether to wait for shutdown to complete
     */
    public void shutdown(boolean wait)
    {
        try
        {
            synchronized(writeThread)
            {
                writeThread.shutdown();
                writeThread.notifyAll();
            }           
        }
        catch (Exception ex)
        {            
        }

        timer.cancel();

        if (ms != null)
        {
            ms.removeMulticastListener(this);
            ms.close();
            ms = null;
        }
        
        persistence.backup(repData);

        if (orb != null)
        {
            orb.shutdown(wait);
        }
    }

    /**
     * Sets the <code>heartbeat</code> variable that indicates how often
     * the IMR should check whether servers are still alive.  This is
     * used in conjunction with the <code>AlivenessPolicy</code>
     * @param beat the number of seconds after which the IMR checks if
     * servers are alive.
     */
    public void heartbeat(int beat)
    {
        heartbeat = beat;
    }

    /**
     * Returns the heartbeat value
     */
    public int heartbeat()
    {
        return heartbeat;
    }

    /**
     * Returns the id of the IMR
     */
    public long id()
    {
        return id;
    }

    /**
     * Returns the id of the IMR
     */
    public void id(long id_)
    {
        id = id_;
    }

    /**
     * Method invoked by servers that are registered with the IMR to check if
     * the IMR is still running.  The method also updates the activity table
     * to indicate that the server is still active.
     * @param sid the id of the server that has pinged us.
     */
    public void ping(long sid)
    {
        if (logger.isDebugEnabled())
        {
            logger.debug("Got pinged by server " + sid);
        }

        synchronized (activityTable)
        {
            if (runningServers.containsKey(new Long(sid)))
            {
                activityTable.put
                (
                    new Long(sid),
                    new Long(System.currentTimeMillis())
                );
            }
            else
            {
                // TODO - should an exception be thrown to prompt the server to
                // register itself - cos we don't know anything about it
                if (logger.isDebugEnabled())
                {
                    logger.debug
                        ("IMR Got pinged by unknown server :: id " + sid);
                }
            }
        }
    }

    // Admin operations

    /**
     * Utility method to list POAs registered for a given server
     * @param server the id under which the server is registered
     * @return an array of POA information objects
     */
    public POAInfo[] listPOAs(long server)
    {
        POANameKey pnKey;
        POANameKey ipnKey;
        ArrayList arrayList = new ArrayList();
        POAInfo pinfo;
        POAInfo[] pinfos;
        LoadBalancingPolicy lb_policy;
        ImplRecord irec;
        ArrayList rpoas;
        Map.Entry entry;
        Iterator poaIter;
        RunningPOA rpoa;

        Iterator iter = runningPOAs.entrySet().iterator();

        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();

            pnKey = (POANameKey) entry.getKey();
            rpoas = (ArrayList) entry.getValue();

            poaIter = rpoas.iterator();

            while (poaIter.hasNext())
            {
                rpoa = (RunningPOA) poaIter.next();

                if (rpoa.id == server)
                {
                    ipnKey = getIPOANameKey(pnKey);
                            
                    irec = (ImplRecord) repData.getImpl(ipnKey);

                    if (irec != null)
                    {
                        lb_policy =
                            LoadBalancingPolicy.from_int(irec.lb_policy);
                            
                        pinfo = new POAInfo(pnKey.toString(), lb_policy);

                        arrayList.add(pinfo);
                    }
                    else
                    {
                        if (logger.isDebugEnabled())
                        {
                            logger.warn
                            (
                                "Listing POAS and implementation details not "
                                + " found for " + pnKey.toString()
                            );
                        }

                        continue;
                    }
                }
            }
        }

        pinfos = new POAInfo[arrayList.size()];

        return (POAInfo[]) arrayList.toArray(pinfos);
    }

    /**
     * Utility method to list servers registered for a given implementation
     * @param sname the name of the implementation.
     * @return an array of Server information objects
     */
    public ServerInfo[] listServers(ServerName sname) throws NotFound
    {
        ServerInfo[] sinfos = null;
        ServerInfo sinfo;
        Server srv = null;
        boolean hasImpl = false;
        boolean aliveCheck = aliveness.value() != 3;
        POANameKey[] keys = null;
        Map.Entry map;
        POANameKey key;
        RunningPOA rpoa;
        ArrayList arrayList = new ArrayList();

        Iterator iter = runningPOAs.entrySet().iterator();
        ArrayList rpoas;
        Iterator poaIter;

        if (sname.discriminator())
        {
            hasImpl = true;
        }
        else
        {
            keys = POANameKey.serverNameToKeys(sname);
        }

        while (iter.hasNext())
        {
            map = (Map.Entry) iter.next();
            key = (POANameKey) map.getKey();

            if (hasImpl)
            {
                /* There may be a number of POAs for 1 implementation so check
                 * for a match on impl name
                 */
                if (sname.impl().equals(key.getImplName()))
                {
                    rpoas = (ArrayList) map.getValue();
                    poaIter = rpoas.iterator();

                    while (poaIter.hasNext())
                    {
                        rpoa = (RunningPOA) poaIter.next();

                        sinfo =
                            new ServerInfo
                            (
                                rpoa.id,
                                getServerState(rpoa.id, aliveCheck),
                                rpoa.reg.srv
                            );

                        arrayList.add(sinfo);
                    }
                }
            }
            else
            {
                for (int i = 0; i < keys.length; i++)
                {
                    rpoas = (ArrayList) runningPOAs.get(keys[i]);

                    if (rpoas != null)
                    {
                        poaIter = rpoas.iterator();

                        while (iter.hasNext())
                        {
                            rpoa = (RunningPOA) iter.next();

                            sinfo =
                                new ServerInfo
                                (
                                    rpoa.id,
                                    getServerState(rpoa.id, aliveCheck),
                                    rpoa.reg.srv
                                );

                            arrayList.add(sinfo);
                        }
                    }
                }
            }
        }

        sinfos = new ServerInfo[arrayList.size()];

        return (ServerInfo[]) arrayList.toArray(sinfos);
    }

    public void shutdownServer(long server)
    {
    }
    
    /**
     * Gets the ServerState to be used when displaying Server details
     */
    private ServerState getServerState(long id, boolean aliveCheck)
    {
        /* The default is unknown if we don't find it active or we're
         * not checking aliveness then we don't know what state it's in!
         */
        ServerState active = ServerState.from_int(ServerState._UNKNOWN);
        ;

        /* If the server is in the activityTable we assume it's active*/
        if (activityTable.containsKey(new Long(id)))
        {
            active = ServerState.from_int(ServerState._ACTIVE);
        }
        else
            if (aliveCheck)
            {
                /* If it's not in the activity table and we're doing aliveness 
                 * checks assume it's inactive - don't know that we'll get here
                 */
                active = ServerState.from_int(ServerState._INACTIVE);
            }

        return active;
    }

    /**
     * Utility method to list all registered implementations.  This caters for
     * situations where the implementation has been registered but no POA info
     * has been registered so the implementation cannot be associated with a
     * host
     */
    public ImplInfo[] listAllImpls()
    {
        ImplInfo[] iinfos;
        ImplInfo iinfo;
        ImplRecord irec;
        Map.Entry entry;
        POANameKey pnKey;
        ArrayList alist = new ArrayList();

        Set regImpls = repData.getImpls().entrySet();

        Iterator iter = regImpls.iterator();
                
        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();
            pnKey = (POANameKey) entry.getKey();
            irec = (ImplRecord) entry.getValue();

            iinfo =
                new ImplInfo
                (
                    pnKey.keyToServerName(),
                    irec.multiples,
                    LoadBalancingPolicy.from_int(irec.lb_policy),
                    ActivationMode.from_int(irec.mode)
                );

            alist.add(iinfo);
        }

        iinfos = new ImplInfo[alist.size()];

        return (ImplInfo[]) alist.toArray(iinfos);
    }

    /**
     * Utility method to list all <code>Activators</code> registered with
     * this IMR.  This is not implemented yet as the ActivatorImpl is not 
     * a Server and does not get a ServerImpl created as it is not 
     * registered with the ImR.  So we cannot return anything other than
     * it's ID which is more or less useless as it tells the user nothing.
     * TODO
     * @return an array of ServerInfo objects
     */
    public ServerInfo[] listActivators()
    {
        //ServerInfo [] sinfos = null;
                
        throw new RuntimeException("Not yet implemented");
/*        Map.Entry entry;
        Long id;
        Activator act;
        ServerInfo sinfo;
        ServerInfo [] sinfos = new ServerInfo[activators.size()];
        int i = 0;
                
        Iterator iter = activators.entrySet().iterator();
        
        while (iter.hasNext())
        {
            entry = (Map.Entry) iter.next();
        
            sinfo = new ServerInfo();
            sinfo.id = ( (Long) entry.getKey()).longValue();
            sinfo.state = ServerState.ACTIVE;            
            sinfo.srv = (Server) entry.getValue();        
            sinfos[i] = sinfo;
            
            i++;
        }
*/        
       // return sinfos;
    }

    /*Private Methods */

    /**
     * Internal method to register implementations, servers and poas.  An 
     * ImplInfo is only created if one does not already exist for this server.  
     * If it is to be created the multiple and autoActivate flags are set to 
     * false as multiples and auto activation are only allowed if the server 
     * implementation is pre-registered via the <code>IMRManager</code>
     * @param <code>poas</code> the poa names to be registered
     * @param <code>rinfo</code> the RegistrationInfo object containing details
     * of the server to be registered
     * @param <code>createImpl</code> boolean indicating whether an ImplInfo is 
     * to be created
     * @param <code>policy</code> the load balancing policy to be applied
     * @param <code>id</code> the id number - which may already be populated 
     * if re-registration
     *
     */
    private long internalRegisterPOA
    (
        POANameKey ipnKey,
        POANameKey pnKey,
        RegistrationInfo rinfo,
        int lb_policy,
        int act_mode,
        long sID
    )
    throws NotFound
    {
        ServerRecord srecord;
        POARecord precord;
        RunningPOA rpoa;
        RunningServer rserver;
        LoadBalancingPolicy policy =
            LoadBalancingPolicy.from_int(lb_policy);
        ActivationMode mode = ActivationMode.from_int(act_mode);
        ArrayList rpoas;
        boolean replacement = false;
                
        writePending = true;
        
        /* For the first release it will  not be possible to have multiples if
         * the implementation is being automatically registered.  If multiples
         * are required the implementation must be pre-registered using the
         * command line utility.  The load balancing policy is therefore
         * not relevant as there will only ever be 1 server registered.
         */
        if (repData.getImpl(ipnKey) == null)
        {
            ImplRecord irec = new ImplRecord(false, policy.value(), mode.value());

            repData.addImpl
            (
                ipnKey, 
                irec
            );
        }

        /* Check to see if this is a new registration.  We cannot test on negatives as the
         * Random object can return negatives
         */        
        if (sID == 0 || sID == -1)
        {
            sID = nextID();
        }

        if (stateLogger.isDebugEnabled())
        {
            stateLogger.debug
            (
                "Registering POA with id "
                + sID
                + " and key "
                + pnKey.toString()
            );
        }
         
              
        rpoas = (ArrayList) runningPOAs.get(pnKey);

        /* If there are no POAs for this key then create an entry*/
        if (rpoas == null)
        { 
            precord = new POARecord(rinfo.endpoints, sID);
            rpoa = new RunningPOA(rinfo, sID);                        
            rpoas = new ArrayList();
            rpoas.add(rpoa);
            runningPOAs.put(pnKey, rpoas);
            repData.addPOA(pnKey, precord);            
        }
        else
        {
            /* If an entry already exists see if the server is recorded*/
            for (int i = 0; i < rpoas.size(); i++)
            {
                rpoa = (RunningPOA) rpoas.get(i);
                
                if (rpoa.id == sID)
                {
                    /* replace the data*/
                    rpoa.reg = rinfo;
                    precord = new POARecord(rinfo.endpoints, sID);
                    repData.replacePOA(pnKey, precord);
                    replacement = true;
                    break;
                }       
            }

            /* If it wasn't a replacement then add the server to
             * this POAs records
             */
            if (!replacement)
            {   
                precord = new POARecord(rinfo.endpoints, sID); 
                rpoa = new RunningPOA(rinfo, sID);    
                rpoas.add(rpoa);
                repData.addPOA(pnKey, precord);
            }
        }
        
        rserver = 
            new RunningServer
            (
                rinfo.srv,
                System.currentTimeMillis(),
                pnKey.getImplName()
            );
               
        /* Store the server details with the time the server was 
         * registered 
         */
        srecord =
            new ServerRecord
            (
                orb.object_to_string(rinfo.srv),
                ipnKey
            );

        synchronized (runningServers)
        {
            /* If an entry already exists for this key the value will be overwritten
             * with the new server details - HashMap
             */
            runningServers.put(new Long(sID), rserver);
        }

        synchronized (activityTable)
        {
            activityTable.put(new Long(sID), new Long(System.currentTimeMillis()));
        }

        /* Store the server details in persistence*/            
        repData.addServer
        (
            sID,
            srecord
        );
        
        try
        {
            synchronized (writeThread)
            {
                writeThread.notifyAll();
            }
        }
        catch (Exception e)
        {                
        }
        
        return sID;
    }

    /**
     * Internal method that checks other ImRs running in the domain to see if they
     * have the required object registererd.  This is only called if an internal
     * object has not been found and the initial request was not from another ImR.
     * @param <code> obj </code> the required object
     * @return the CORBA Object - can be null
     */
    private EndpointProfile[] checkOtherImRs(String impl, String[] poas)
    {
        ArrayList imrs = null;
        Repository rep = null;
        EndpointProfile[] endpoints = null;
        String thisIMR = orb.object_to_string(imrObj);
        String otherIMR = null;
        IMRLocator locator = null;
        
        if (locateLogger.isDebugEnabled())
        {
            locateLogger.debug("In checkOtherImRs");
        }

        int uport =
            Environment.getIntPropertyWithDefault("jacorb.imr.udp_port", 0);

        try
        {
            locator = new IMRLocator(orb, uport);

            locator.findIMRs();

            Thread.sleep(3000);
        }
        catch (InterruptedException ie)
        {}
        catch (Exception ex)
        {
            // put out some debug statement
        }

        imrs = locator.getIMRs();
        
        if (locateLogger.isDebugEnabled())
        {
            locateLogger.debug("Found " + imrs.size() + " other ImRs");
        }

        for (int i = 0; i < imrs.size(); i++)
        {
            try
            {
                
                /* Check that it's not this IMR - this is temporary and awful
                 * but future enhancements will involve all ImRs knowing 
                 * about all registered objects and changes so this 
                 * functionality will  no longer be necessary
                 */                
                otherIMR = orb.object_to_string((org.omg.CORBA.Object) imrs.get(i));

                if (!(thisIMR.equals(otherIMR)))
                {                                    
                    rep = (Repository) imrs.get(i);
                                                     
                    endpoints = rep.imrLocateRequest(impl, poas);

                    if (locateLogger.isDebugEnabled())
                    {
                        locateLogger.debug("Got " + endpoints + " endpoints from other IMR");
                    }
                }

                if (endpoints != null)
                {
                    break;
                }
            }
            catch (Exception ex)
            {
                locateLogger.warn
                (
                    "An exception occurred with remote ImR. ",
                    ex
                );
            }
        }

        return endpoints;
    }

    /**
     * Internal method that attempts to start a server - if all the necessary conditions
     * are met
     */
    private RunningPOA startProcess(POANameKey pnKey)
    {
        Activator activator = null;
        ArrayList activeServers;
        RunningPOA rpoa = null;
        ServerName sname = null;
        boolean startRequested = false;
        
        if (locateLogger.isDebugEnabled())
        {
            locateLogger.debug
                ("IMR :: startProcess for POA " + pnKey.toString());
        }

        /* Get the name under which Activators and impl info will be stored as
         * this will only be the implname for JacORB and not include the POA 
         * names.  If the implname is not set then it will be the POA Names 
         * (foreign objects)
         */
        POANameKey ipnKey = getIPOANameKey(pnKey);

        try
        {
            ImplRecord irec = (ImplRecord) repData.getImpl(ipnKey);

            /* Only get the Activator if we have and ImplRecord and
             * the ActivationMode is AUTO_START
             */
            if (irec != null && irec.mode == ActivationMode._AUTO_START)
            {
                /* Get the appropriate activator*/
                activator = 
                    (Activator) activators.get(new Long(irec.activatorID));                    
            }

            /* If the activator is not null then the irec must be populated
             * and the ActivationMode must be AUTO_START
             */            
            if (activator != null)
            {
                sname = pnKey.keyToServerName();
                
                startRequested = true;
                
                activator.startServer(sname, true);
                                
                /* Wait a configured amount of time for the server to start. 
                 * As it may come up on another port we can't ping to 
                 * determine liveness. Default it here to 5 seconds if not set 
                 * already.
                 */
                Thread.sleep
                    (Environment.getIntPropertyWithDefault
                        (
                            "jacorb.imr.activator.start_timeout",
                            5000)
                        );
                    
                activeServers = (ArrayList) runningPOAs.get(pnKey);

                /* There may be servers registered now - check before processing */
                if (activeServers.size() == 1)
                {
                    rpoa = (RunningPOA) activeServers.get(0);
                }
                /* need to check in case none registered */
                else if (activeServers.size() > 1)               
                {
                    rpoa = applyLoadBalancing(activeServers, pnKey);
                }
                
                updateActivator(activator, sname);
            }
            else
            {
                /* If we haven't found an Activator and the ImplRecord exists
                 * set it's ID to 0
                 */
                if (irec != null)
                {
                    irec.activatorID = 0;   
                }
            }                     
        }
        catch (StartupFailed sf)
        {            
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug(sf.reason);
            }
            
            updateActivator(activator, sname);

            
            throw new INTERNAL
                ("Activator failed to start server " + sf.reason);            
        }
        catch (UnknownServer us)
        { 
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug
                    (ipnKey.toString() + " not registered with Activator");
            }                  
            
            updateActivator(activator, sname);
                             
            throw new INTERNAL
                (ipnKey.toString() + " not registered with Activator");                           
        }
        catch (Exception ex)
        {
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug(ex.getMessage());
            }
            
            updateActivator(activator, sname);
                        
            throw new INTERNAL("Activator failed to start server " + ex);            
        }
        
        return rpoa;
    }

    private void updateActivator(Activator activator, ServerName sname)
    {
        try
        {                    
            activator.serverStarted(sname);
        }
        catch (UnknownServer us)
        {}
        catch (Exception ex)
        {}
    }              

    /**
     * Internal method to instigate the application of the appropriate
     * load balancing policy and return the selected <code>ServerInfo</code>
     * object.
     */
    private RunningPOA applyLoadBalancing
    (
        ArrayList activeServers,
        POANameKey pnKey
    )
    throws StartupFailed
    {
        RunningPOA rpoa = null;
        Long sID = null;
        RunningServer rsrv = null;
        HashMap selected = null;
        
        boolean activeServerFound = false;
        long id;
        
        POANameKey ipnKey = getIPOANameKey(pnKey);
        TreeMap serverTimes = new TreeMap();
        
        Iterator iter = activeServers.iterator();       

        while (iter.hasNext())
        {
            id = ((RunningPOA) iter.next()).id;
            rsrv = (RunningServer) runningServers.get(new Long(id));
            serverTimes.put(new Long(rsrv.regTime), new Long(id));
        }

        LoadBalancingPolicy lbpolicy =
            LoadBalancingPolicy.from_int
                (((ImplRecord) repData.getImpl(ipnKey)).lb_policy);

        LBPolicy policy = LBPolicyRegistry.lookup(lbpolicy);

        /* Get the next server according to the policy.  A server that is no longer
         * running may be returned. The Policy does not check whether the
         * Server is running as the Aliveness is configured for the ImR.  If the
         * ImR is checking and the server is no longer alive it can deregister the
         * server.  A further attempt is made to locate a running server until 
         * there are no further Servers to check.  This process does not take account
         * of any Server that may be registered during the locate process.
         */
        try
        {
            Map.Entry entry = null;
            
            int servers = 0;
            int initSize = serverTimes.size();
            
            /* This is a very ropey attempt to deal with failover of a server
             * and incorporating load balancing.  It will probably be 
             * overhauled at the next release of this code - however it works!
             */
            while (!activeServerFound && servers < initSize)
            {
                sID = null;
                 
                servers++;
                
                selected =
                    policy.getSelectedServer
                        (ipnKey.toString(), serverTimes);
               
                if (locateLogger.isDebugEnabled())
                {
                    locateLogger.debug("Applying load balancing and " 
                        + selected.size() + " servers found");
                }
                                
                if (selected.size() > 0)
                {     
                    entry = (Map.Entry) selected.entrySet().iterator().next();
            
                    sID = (Long) entry.getValue();           
                }
            
                if (locateLogger.isDebugEnabled())
                {
                    locateLogger.debug
                        ("Applying load balancing check alive is " 
                            + checkAlive + " server id is " + sID.longValue());
                }
                
                /* Check if configured to ping on lookup */
                if ((checkAlive()) && (sID != null))
                {
                    try
                    {
                        ((RunningServer) runningServers.get(sID)).srv.ping(0);
                        activeServerFound = true;
                        
                        if (locateLogger.isDebugEnabled())
                        {
                            locateLogger.debug
                                ("Apply load balancing active server found");
                        }
                    }
                    catch (Exception ex)
                    {                        
                        /* IF the server isn't alive remove it*/
                        try
                        {
                            if (locateLogger.isDebugEnabled())
                            {
                                locateLogger.debug
                                    ("Apply load balancing and deregistering server id " 
                                        + sID.longValue());
                            }
                            
                            deregisterAll(sID.longValue());
                        }
                        catch (NotFound nf)
                        {}
                        
                        serverTimes.remove( (Long) entry.getKey());                        
                    }
                }
                else if (sID != null)
                {
                    /* If we aren't pinging then we just have to go with the server returned
                     * if it's in the runningServers table
                     */
                    if (runningServers.get(sID) != null)
                    {
                        activeServerFound = true;
                    }
                }
            }
        }
        catch (Exception ex)
        {
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug
                   (
                       "Exception applying load balancing ",
                       ex
                   );
            }
        }

        if (!activeServerFound)
        {
            /* For the initial release the ability to autoActivate when there
             * are multiple registrations is not allowed.  This is because the
             * policy as to what to do if the selected server is not active
             * but other servers are active has not yet been bottomed out.
             */

            throw new StartupFailed(
               "Selected server is not active.  AutoActivate multiples NYI");
        }
        else
        {   
            /* Search through the active servers for the matching POARecord from
             * which the endpoints will be obtained
             */
            iter = null;
            iter = activeServers.iterator();

            while (iter.hasNext())
            {
                rpoa = (RunningPOA) iter.next();

                if (rpoa.id == sID.longValue())
                {
                    break;
                }
            }
        }
        
        return rpoa;
    }

    /**
     * The IMR adds itself as a listener to a MulticastServer at
     * instantiation time.  If a request is received by the
     * MulticastServer it is passed to the IMR.  The IMR then
     * calls the object looking for an imr, giving it a reference
     * to this IMR.
     */
    public byte[] multicastEvent(MulticastEvent me)
    {
        byte[] returnPacket = null;
        byte[] requestPacket = null;
                        
        if (locateLogger.isDebugEnabled())
        {
            locateLogger.debug("Got the multicast event \n");
        }

        try
        {
            requestPacket = me.getData();

            CDRInputStream clientRequest = new CDRInputStream(orb, requestPacket);            
            clientRequest.openEncapsulatedArray();
            String serviceName = clientRequest.read_string();

            clientRequest.close();

            CDROutputStream serverResponse = new CDROutputStream();
            serverResponse.beginEncapsulatedArray();
            serverResponse.write_string("ImplementationRepository");

            /* Separate steps to assist debugging*/

            RepositoryHelper.write(serverResponse, RepositoryHelper.narrow(imrObj));
            
            returnPacket = serverResponse.getBufferCopy();

            serverResponse.close();
        }
        catch (Exception ex)
        {
            if (locateLogger.isDebugEnabled())
            {
                locateLogger.debug
                (
                    "Error handling MulticastEvent: ",
                    ex
                );
                    
                locateLogger.debug(ex.getMessage());
            }
            
            returnPacket = new  byte[0];
        }

        return returnPacket;
    }

    /**
     * When an IMR is looking for other IMRs it will send a request to
     * locate other running IMRs via UDP.  
     */
    public void registerIMR(Repository rep)
    {
        // this method is not currently implemented as the federation of IMRs
        // has not yet been tackled.
        throw new RuntimeException("NYI");
    }

    /**
     * Gets the <code>Activator</code> for the name specified
     */
    private Activator getActivator(String implName)
    {
        Activator act = null;
        return act;
    }

    /**
     * Gets an Implementation type <code>POANameKey</code>.  This will be
     * either based on an impl name only or will be based on POA names only
     * but not both.  This is because implementation information can be 
     * pre-registered and at that time we will only have an impl name or
     * POA names but not both.  This all happens because some ORBs do not
     * use impl names and the ultimate aim is to allow objects from 
     * different ORBs to be registered with this ImR.
     */
    private POANameKey getIPOANameKey(POANameKey pnKey)
    {
        if (pnKey.hasImpl())
        {
            return new POANameKey(pnKey.getImplName());
        }
        else
        {
            return new POANameKey(pnKey.getPOAS());
        }
    }

    /**
     * Method to write data to the appropriate persistence object
     */
    protected void writeData()
    {
        persistence.write(repData);
        writePending = false;
    }
    
    /**
     * Main
     */
    public static void main(String args[])
    {
        Properties props = new Properties();

        System.setProperty("jacorb.implname", "theImR");
        System.setProperty("jacorb.use_imr", "off");

        checkArgs(args, props);

        orb = org.omg.CORBA.ORB.init(args, props);

        try
        {
            new RepositoryImpl();
        }
        catch (Exception ex)
        {
            System.out.println("Exception starting imr");
            ex.printStackTrace();

            // Force an exit
            System.exit(1);
        }
    }

    /**
     * Separate method to check the input arguments.  Any failure will result
     * in a System.exit()
     */
    private static void checkArgs(String args[], Properties props)
    {
        String default_host = null;
        String imrPort = null;
        String udpPort = null;
        String timeOut = null;
        String policy = null;
        String dataFile = null;
        long mseconds = 0;

        int apolicy = 0;

        String temp = null;

        try
        {
            for (int i = 0; i < args.length; i++)
            {
                if (args[i].equals("-h") || args[i].equals("--help"))
                {
                    usage(); //exits
                }

                if (args[i].equals("-n") 
                        || args[i].equals("--imr_new"))
                {
                    if (dataFile != null)
                    {
                        System.out.println("New file not appropriate "
                            + "if data file specified");
                        System.exit(0);                                       
                    }
                    else
                    {
                        newFile = true;
                    }
                }
                else if (i < (args.length - 1))
                {
                    if (args[i].equals("-p") || args[i].equals("--port"))
                    {
                        imrPort = args[++i];

                        if (imrPort != null)
                        {
                            Environment.setProperty(
                                "jacorb.imr.port_number",
                                imrPort);
                        }
                    }

                    else if (args[i].equals("-i")
                             || args[i].equals("--imr_ior"))
                    {
                        iorfile = args[++i];

                    }
                    else if (args[i].equals("-a")
                             || args[i].equals("--aliveness"))
                    {
                        policy = args[++i];

                        if (policy != null && isValidPolicy(policy))
                        {
                            Environment.setProperty
                            (
                                "jacorb.imr.aliveness_policy",
                                policy
                            );
                        }
                    }
                    else if (args[i].equals("-t")
                             || args[i].equals("--timeout"))
                    {
                        timeOut = args[++i];
                    }
                    else if (args[i].equals("-v")
                             || args[i].equals("--activator_timeout"))
                    {
                        Environment.setProperty
                        (
                            "jacorb.imr.activator.start_timeout",
                            args[++i]
                        );
                    }
                    else if (args[i].equals("-u")
                             || args[i].equals("--udp_port"))
                    {
                        udpPort = args[++i];
                    }
                    else if (args[i].equals("-s")
                             || args[i].equals("--auto"))
                    {
                        Environment.setProperty(
                            "jacorb.imr.allow_auto_register", args[++i]);
                    }
                    else if (args[i].equals("-f")
                             || args[i].equals("--imr_data"))
                    {
                        if (newFile)
                        {
                            System.out.println("Data file name not appropriate "
                               + "if new file specified");
                            System.exit(0);                               
                        }
                        else
                        {
                            dataFile = args[++i];
                            Environment.setProperty
                                ("jacorb.imr.imr_data_file", dataFile);
                        }
                    }

                    else if (args[i].equals("-b") 
                        || args[i].equals("--imr_backup"))
                    {
                        Environment.setProperty
                            ("jacorb.imr.imr_backup_file", args[++i]);
                    }
                }
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            usage();
        }

        if (timeOut != null)
        {
            if (Environment.getAlivenessPolicy() == AlivenessPolicy._ON_EXPIRY)
            {
                Environment.setProperty(
                    "jacorb.imr.active_server_timeout",
                    timeOut);
            }
            else
                if (Environment.getAlivenessPolicy() == AlivenessPolicy._PING)
                {
                    Environment.setProperty("jacorb.imr.heartbeat", timeOut);
                }
        }

        if (iorfile == null)
        {
            iorfile = Environment.getProperty("jacorb.imr.ior_file");

            if (iorfile == null)
            {
                System.out.println("ImR IOR file location not known");
                usage();
            }
        }
        else
        {
            Environment.setProperty("jacorb.imr.ior_file", iorfile);
        }

        /* Check port number for MulticastServer */
        if (udpPort != null)
        {
            Environment.setProperty("jacorb.imr.udp_port", udpPort);
        }

        temp = Environment.getProperty("jacorb.imr.port_number");

        if (temp != null && temp.length() > 0)
        {
            props.setProperty("OAPort", temp);
        }

        default_host = Environment.getProperty("jacorb.imr.host");

        if (default_host != null && default_host.length() > 0)
        {
            props.setProperty("OAIAddr", default_host);
        }
    }

    /* Private method to check that the policy entered is valid*/
    private static boolean isValidPolicy(String policy)
    {
        /** Used when setting the Aliveness policy*/
        String[] policies = new String[] { "0", "1", "2", "3" };

        return policies[Integer.parseInt(policy)] != null;
    }

    /**
     * Help menu
     */
    public static void usage()
    {
        System.out.println();
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("Usage:                                                        ");
        System.out.println();
        System.out.println
            ("[ -p | --port ] <port number>  The port number on which the   ");
        System.out.println
            ("                               IMR is to listen               ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -i | --imr_ior ] <filename>  The filename for the IOR for   ");
        System.out.println
            ("                               this Repository                ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -f | --imr_data ] <filename> The filename in which the      ");
        System.out.println
            ("                               Repository data is to be       ");
        System.out.println
            ("                               saved                          ");
        System.out.println
            ("--------------------------------------------------------------");    
        System.out.println
            ("[ -b | --imr_backup ] <filename> The filename in which the    ");
        System.out.println
            ("                               Repository data is to be       ");
        System.out.println
            ("                               backed up - if required        ");
        System.out.println
            ("--------------------------------------------------------------");             
        System.out.println
            ("[ -n | --imr_new ]            If set a new file will be       ");
        System.out.println
            ("                               created for the Repository data");
        System.out.println
            ("                               - if not set data in any       ");
        System.out.println
            ("                               existing file will be loaded   ");
        System.out.println
            ("--------------------------------------------------------------");                 
        System.out.println
            ("[ -a | --aliveness ] <policy> The aliveness policy to be      ");
        System.out.println
            ("                              applied i.e. when/if the ImR    ");
        System.out.println
            ("                              pings the server to see if it is");
        System.out.println
            ("                              alive.                          ");
        System.out.println
            ("                                                              ");
        System.out.println
            ("                                0 = PING                      ");
        System.out.println
            ("                                1 = ON_LOOKUP                 ");
        System.out.println
            ("                                2 = ON_EXPIRY                 ");
        System.out.println
            ("                                3 = NONE                      ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -t | -- timeout ] <timeout> The time interval to be applied ");
        System.out.println
            ("                              when the aliveness policy is    ");
        System.out.println
            ("                              ON_EXPIRY or PING               ");
        System.out.println
            ("                              (time in milliseconds)          ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -u | --udp_port ]           The port number on which the    ");
        System.out.println
            ("                              MulticastServer is to run for   ");
        System.out.println
            ("                              this ImR.                       ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -s | --auto ] <on/off>     Flag indicating whether automatic");
        System.out.println
            ("                             registration of servers is       ");
        System.out.println
            ("                             allowed                          ");
        System.out.println
            ("                                values on or off              ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -v | --activator_timeout ] <timeout>                        ");
        System.out.println
            ("                              The time interval that the      ");
        System.out.println
            ("                              Repository should wait for a    ");
        System.out.println
            ("                              server to be autoactivated      ");
        System.out.println
            ("                              (time in milliseconds)          ");
        System.out.println
            ("--------------------------------------------------------------");
        System.out.println
            ("[ -h | --help ]               Displays this help              ");
        System.out.println
            ("--------------------------------------------------------------");
        System.exit(0);
    }

    /**
     * This thread is started if the AlivenessPolicy is "PING" and the
     * heartbeat value is > 0.  It pings any active servers to ensure
     * they are still alive.  If an exception is not returned from the
     * ping then the server is taken to be alive and the activity table
     * is updated with the current time.
     */
    private class PingTask extends TimerTask
    {
        public void run()
        {
            Set s;
            HashMap sinfos;
            Iterator iter;
            ServerRecord server;
            Long srvID;
            RunningServer rsrv = null;
            Map.Entry entry;
            
            synchronized (activityTable)
            {
                iter = activityTable.entrySet().iterator();

                /* Iterate through the active servers*/
                while (iter.hasNext())
                {
                    entry = (Map.Entry) iter.next();
                    
                    srvID = (Long) entry.getKey();

                    /* Ping each server then update the activity table */
                    try
                    {
                       rsrv = ((RunningServer) runningServers.get(srvID));
                       rsrv.pingCount++;
                       rsrv.srv.ping(0);
                       
                       /* If the ping is successful reset the count */
                       rsrv.pingCount = 0;   
                        
                       entry.setValue(new Long(System.currentTimeMillis()));
                    }
                    catch (Exception ex)
                    {
                        if (rsrv.pingCount == maxPings)
                        {
                            iter.remove();
                            try
                            {
                                deregisterAll(srvID.longValue());
                            }
                            catch (NotFound nf)
                            { 
                                // Nothing to do - if the server isn't found we
                                // can't remove it                              
                            }                                                       
                        }           
                    }
                }
            }
        }
    }

    /**
     * This thread is used to check whether a server is still active.
     * If the server has not communicated with the imr in the specified
     * timeout it is deregistered
     */
    private class ActivityTask extends TimerTask
    {
        public void run()
        {
            Set s;
            Iterator iter;
            Long sid;
            long time;
            HashMap sinfos;
            RunningServer rsrv = null;
            Map.Entry entry;
                        
            synchronized (activityTable)
            {                
                iter = activityTable.entrySet().iterator();

                while (iter.hasNext())
                {
                    entry = (Map.Entry) iter.next();
                    
                    sid = (Long) entry.getKey();
                    time = ((Long) entry.getValue()).longValue();

                    if ((System.currentTimeMillis() - time) > (activeTimeout))
                    {
                        try
                        {
                            rsrv = ((RunningServer) runningServers.get(sid));
                            rsrv.pingCount++;
                            rsrv.srv.ping(0);
                            rsrv.pingCount = 0;                            
                        }
                        catch (Exception ex)
                        {
                            if (rsrv.pingCount == maxPings)
                            {
                                iter.remove();
                                
                                try
                                {
                                    deregisterAll(sid.longValue());
                                }
                                catch (NotFound nf)
                                {
                                    // Nothing to do if the server isn't found we can't
                                    // remove it
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * <code>Shutdown</code> is a thread that is run when receiving
     * a Ctrl-C or kill.
     */
    protected class Shutdown extends Thread
    {
        public synchronized void run()
        {
            if (logger.isDebugEnabled())
            {
                logger.debug("ImR: Shutting down");
            }

            shutdown(true);
        }
    }

    /**
     * <code>Write</code> runs as a background thread and writes the
     * repository data out as changes are made
     */
    private class Write extends Thread
    {
        boolean done;

        public Write ()
        {
        }

        public void run ()
        {
            while (true)
            {
                try
                {
                    writeData();
                }
                catch (Exception ex)
                {
                    logger.warn("Exception writing data", ex);
                }
                
                if (done)
                {
                    break;
                }

                /* Only wait if there are no further writes pending*/
                if (!writePending)
                {
                    try
                    {
                        synchronized (this)
                        {
                            this.wait();
                        }
                    }
                    catch (InterruptedException ex) 
                    {
                    }
                }
            }
        }

        /**
         * <code>shutdown</code> toggles the thread to shut itself down.
         */
        public void shutdown ()
        {
            done = true;
        }
    }
 
    /**
     * Internal class to hold server details to allow quick access
     * at runtime
     */   
    protected class RunningServer
    {
        /* The Server object*/
        Server srv;
  
        /* The time this Server was registered - needed for Load Balancing
         */      
        long regTime;
        
        /* The impl applicable to this Server*/
        String impl;
        
        /* The number of times this Server has been pinged and failed.
         * Controls when we decide the Server can be deemed INACTIVE
         */
        int pingCount = 0;
           
        /* Constructor */     
        public RunningServer(Server in_srv, long in_time, String in_impl)
        {
            srv = in_srv;
            regTime = in_time;
            impl = in_impl;
        }
    }

    /**
     * Internal class to hold POA details for Servers that are running.
     */    
    protected class RunningPOA
    {
        /* The registration information supplied when the POA was registered*/
        RegistrationInfo reg;
        
        /* The ID of the Server to which this information relates*/
        long id;
        
        /* Constructor*/
        public RunningPOA(RegistrationInfo in_reg, long in_id)
        {
            reg = in_reg;
            id = in_id;
        }        
    }
}
