package org.jacorb.imrutility.imr;

import java.util.Timer;
import java.util.TimerTask;

import org.jacorb.util.Debug;

import org.omg.CORBA.INTERNAL;
import org.omg.PortableServer.POAPackage.*;

import org.apache.avalon.framework.logger.*;

/**
 * <code>ServerImpl</code> is CORBA object singleton created within every
 * running server.  It is registered on the root POA during ORB
 * initialization. Note that the root POA must be activated if any of the server
 * operations are to be invoked. This activation may or may not be actually
 * required depending on the exact usage semantics required, for example whether
 * the ImR is configured to ping the Server.
 *
 * @author <a href="mailto:rnc@prismtechnologies.com"></a>
 * @version 1.0
 */
public class ServerImpl extends ServerPOA
{
    /**
     * <code>id</code> is the identity of a running server. This is passed as as
     * attribute to the ping operation so the sender of a ping request can be
     * identified. The ImR has a fixed id of zero.
     */
    private long id = -1;

    /**
     * <code>heartbeat</code> is the interval between hearbeat calls (the ping
     * operation) between the ImR and servers. If set to zero heartbeat is
     * disabled. If set for the ImR it heartbeats all the servers that are
     * registered with it. If set for a server it haerbeats the ImR. If a
     * heartbeat is successfully received from a server or sent by an ImR to a
     * server then the ImR marks the status of that server as ACTIVE.
     */
    private int heartbeat = 0;
    
    /**
     * The ORB
     */
    private org.jacorb.orb.ORB orb = null;

    /**
     * Timer to carry out Ping operations if configured to do so
     */
    private java.util.Timer timer = null;

    /**
     * Logs output
     */
    private Logger logger = Debug.getNamedLogger("jacorb.imr");
    
    /**
     * Constructor
     */
    public ServerImpl()
    {         
    }
    
    /**
     * Constructor
     * @param <code>orb</code> 
     */
    public ServerImpl(org.jacorb.orb.ORB orb)
    {
        this.orb = orb;
    }

    /**
     * <code>id</code> See {@link ServerImpl#id}.
     *
     * @return a <code>long</code> value
     */
    public long id()
    {
        return id;
    }

    /**
     * <code>id</code> See {@link ServerImpl#id}.
     *
     * @param value a <code>long</code> value
     */
    public void id( long value )
    {
        /* If id is -1 the timer should not be running yet */
        boolean isFirstID = (id == -1);
        
        if (value == -1)
        {
            /*
             * The timer should be stopped (if it's running)
             * because we will no longer have an id to ping with.
             */
            stopTimer();
        }
        
        /* Store the new id */
        id = value;
        
        if (isFirstID)
        {
            /* May need to start the timer again. */
            startTimer();
        }
    }

    /**
     * <code>heatbeat</code> See {@link ServerImpl#heartbeat}.
     *
     * @return an <code>int</code> value
     */
    public int heartbeat()
    {
        return heartbeat;
    }

    /**
     * <code>heatbeat</code> See {@link ServerImpl#heartbeat}.
     *
     * @param interval an <code>int</code> value
     */
    public void heartbeat( int interval )
    {
        if (interval != heartbeat)
        {
            stopTimer();
            heartbeat = interval;
            startTimer();
        }
    }
    
    /**
     * <code>ping</code> is used for one server to ping another. This can either
     * be the ImR calling to a server or a server calling the ImR or both. The
     * ImR aliveness policy and the server hearbeat property determine how this
     * operation is used.
     *
     * @param from a <code>long</code> value
     */
    public void ping( long from )
    {
        /*
         * This is a noop operation. The parameter is only is used in the ImR's
         * implementation of it.
         */
       if (logger.isDebugEnabled() )
       {
           logger.debug("Got pinged by imr " + from);
       }
    }

    /**
     * <code>shutdown</code> is an operation to shutdown the running server
     * cleanly. If this is called on any server other than the implementation
     * repository it should call the deregisterAll operation on the ImR so the
     * server state can be marked as inactive.
     *
     * @param wait a <code>boolean</code> value
     */
    public void shutdown( boolean wait )
    {
       orb.shutdown(wait);
    }
    
    /**
    * Install this servant onto the root POA and return the
    * object reference.
    */
    public Server installOnRootPOA()
    {
        org.omg.CORBA.Object obj = null;
        try
        {
            obj = orb.getRootPOA().servant_to_reference(this);
        }
        catch( ServantNotActive e )
        {
            throw new INTERNAL( "Failed to activate Server handler" + e );
        }
        catch( WrongPolicy e )
        {
            throw new INTERNAL( "Failed to activate Server handler" + e );
        }
        return ServerHelper.narrow(obj);
    }
    
    /**
    * Start the ping timer loop *iff* the heartbeat and id are set
    * and we have an ORB pointer to work with.
    */
    private void startTimer()
    {
        if (heartbeat != 0 && orb != null && id != -1)
        {
            timer = new Timer();
            timer.schedule
            (
                new PingImR(),
                heartbeat,
                heartbeat
            );
        }    
    }
    
    /**
    * Stop pinging (if we are).
    */
    private void stopTimer()
    {
        if (timer != null)
        {
            timer.cancel();
            timer = null;
        }
    }

    
    /**
     * <code>PingImR</code> is a timer task to ping a configured IMR. This call
     * will end up in RepositoryImpl::ping - which will note that the server is
     * currently active.
     */
    private class PingImR extends TimerTask
    {
        /**
         * <code>run</code> is the implementation of TimerTask::run.
         */
        public void run()
        {
            Repository imr = orb.getImR();
            try
            {
                imr.ping(id);
            }
            /*  Might throw a transient if it fails. */
            catch( Exception e )
            {
                /*
                 * Inform the ORB that its ImR ref is dead.
                 * Will prompt re-registration if an alternative can
                 * be found. 
                 */
                orb.imrDead(imr);
            }
        }
    }
}
