package org.jacorb.orb;

/*
 *        JacORB  - a free Java ORB
 *
 *   Copyright (C) 1997-2003  Gerald Brose.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;
import java.io.*;
import java.applet.Applet;
import java.lang.reflect.*;

import org.jacorb.imrutility.imr.EndpointProfile;
import org.jacorb.imrutility.imr.NotFound;
import org.jacorb.imrutility.imr.Repository;
import org.jacorb.imrutility.imr.RepositoryHelper;
import org.jacorb.imrutility.imr.Server;
import org.jacorb.imrutility.imr.util.IMRLocator;
import org.jacorb.imrutility.imr.ServerImpl;
import org.jacorb.util.*;
import org.jacorb.orb.iiop.*;
import org.jacorb.orb.policies.*;
import org.jacorb.orb.dii.Request;
import org.jacorb.orb.giop.*;
import org.jacorb.orb.portableInterceptor.*;
import org.jacorb.poa.util.POAUtil;

import org.apache.avalon.framework.logger.*;

import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.INITIALIZE;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.BAD_QOS;
import org.omg.CORBA.LongHolder;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.BooleanHolder;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.ORBPackage.InvalidName;
import org.omg.CORBA.portable.ValueFactory;
import org.omg.CORBA.portable.BoxedValueHelper;
import org.omg.CORBA.portable.StreamableValue;
import org.omg.Messaging.*;
import org.omg.PortableServer.POA;
import org.omg.PortableInterceptor.*;
import org.omg.IOP.*;
import org.omg.IIOP.*;
import org.omg.ETF.*;

/**
 * @author Gerald Brose, FU Berlin
 * @version $Id: ORB.java,v 1.25 2004/10/15 18:45:40 cj Exp $
 */

public final class ORB
    extends ORBSingleton
    implements org.jacorb.poa.POAListener
{
    private static final String versionString = org.jacorb.util.Version.version;
    private static final String dateString = org.jacorb.util.Version.date;
    private static final String nullIORString =
        "IOR:00000000000000010000000000000000";

    /** "initial" references */
    private Map initial_references = new HashMap();

    private org.jacorb.poa.POA rootpoa;
    private org.jacorb.poa.Current poaCurrent;
    private BasicAdapter basicAdapter;
    private org.omg.SecurityLevel2.Current securityCurrent = null;

    /** interceptor handling */
    private InterceptorManager interceptor_manager = null;
    private boolean hasClientInterceptors = false;
    private boolean hasServerInterceptors = false;
    private org.omg.PortableInterceptor.Current piCurrent = new PICurrent();

    /** reference caching */
    private Map knownReferences = null;

    /** connection mgmt. */
    private ClientConnectionManager clientConnectionManager;

    /** The transport manager*/
    private TransportManager transport_manager = null;

    private GIOPConnectionManager giop_connection_manager = null;

    /** buffer mgmt. */
    private BufferManager bufferManager = BufferManager.getInstance();

    /**
     * Maps repository ids (strings) to objects that implement
     * org.omg.CORBA.portable.ValueFactory.  This map is used by
     * register/unregister_value_factory() and lookup_value_factory().
     */
    protected Map valueFactories = new HashMap();

    private Map objectKeyMap = new HashMap();

    /** properties */
    private java.util.Properties _props;

    /** the ORB object's logger */
    private Logger logger;

    /** command like args */
    public String[] _args;

    public java.applet.Applet applet;

    /* for run() and shutdown()  */
    private Object orb_synch = new java.lang.Object();
    private boolean run = true;
    private boolean wait = true;
    private boolean shutdown_in_progress = false;
    private boolean destroyed = false;
    private Object shutdown_synch = new Object();

    /* for registering POAs with the ImR */
    private Repository imr;

    /**
    * A lock for updating the IMR pointer.
    */
    private byte[] imrUpdateLock = new byte[0];

    /**
    * A lock for the imrDead dead IMR reporting method;
    */
    private byte[] imrIsDeadLock = new byte[0];

    /**
     * <code>locator</code> is a cached IMRLocator - which contains an
     * IMRClient used for pinging.
     */
    private IMRLocator locator;

    /* Object ref for the ImR control server (bound to the rootPOA) */
    private Server activeServerHandler;

    /* The servant implementation of the ImR control servant */
    private ServerImpl imrServerImpl;

    private int persistentPOACount;

    public static final String orb_id = "jacorb:" + org.jacorb.util.Version.version;

    /* outstanding dii requests awaiting completion */
    private Set requests = Collections.synchronizedSet( new HashSet() );
    /* most recently completed dii request found during poll */
    private Request request = null;

    /* PolicyManagement */
    private org.jacorb.orb.policies.PolicyManager policyManager = null;

    /* policy factories, from portable interceptor spec */
    private Map policy_factories = null;

    private static org.omg.CORBA.TCKind kind;

    private static final String [] services  =
        {"RootPOA","POACurrent", "DynAnyFactory", "PICurrent", "CodecFactory"};

    private boolean bidir_giop = false;


    public ORB()
    {
        logger = Debug.getNamedLogger("jacorb.orb");
    }

    /**
     * Overrides id() in org.omg.CORBA_2_5.ORB
     */

    public String id()
    {
        return orb_id;
    }

    public boolean useBiDirGIOP()
    {
        return bidir_giop;
    }

    public void turnOnBiDirGIOP()
    {
        if( ! bidir_giop )
        {
            bidir_giop = true;

            clientConnectionManager.setRequestListener( basicAdapter.getRequestListener() );
        }
    }

    /**
     *  This  version of _getObject  is used for references  that have
     *  arrived over the network and is called from CDRInputStream. It
     *  removes stale cache entries
     */
    public synchronized org.omg.CORBA.Object _getObject( ParsedIOR pior )
    {
        String key = pior.getIORString();
        org.omg.CORBA.portable.ObjectImpl o =
            (org.omg.CORBA.portable.ObjectImpl)knownReferences.get( key );

        if( o != null )
        {
            org.jacorb.orb.Delegate del = (org.jacorb.orb.Delegate) o._get_delegate();
            if (del != null)
            {
                ParsedIOR delpior= del.getParsedIOR();
                if (delpior == null)
                {
                    knownReferences.remove(key);
                    if( logger.isDebugEnabled() )
                    {
                        logger.debug("Removing an invalid reference from cache.");
                    }
                }
                else if( pior.getEffectiveProfile()
                           .is_match( delpior.getEffectiveProfile() ))
                {
                    return o._duplicate();
                }
            }
            else
            {
                if( logger.isDebugEnabled() )
                {
                    logger.debug("Remove stale reference from cache ");
                }
                knownReferences.remove( key );
            }
        }

        org.jacorb.orb.Delegate d = new Delegate(this, pior );
        o = d.getReference( null );

        if( Environment.cacheReferences() )
        {
            knownReferences.put( key, o );
        }
        return o;
    }

    /**
     * Find a local POA for a delegate (called from is_local())
     * returns non-null only if a root POA is already activated
     * and all POAs along the path on the poa name are active, i.e.
     * returns null for POAs in the holding state
     */

    org.jacorb.poa.POA findPOA( org.jacorb.orb.Delegate d,
                                org.omg.CORBA.Object ref )
    {
        List scopes;
        String res;
        String refImplName = null;

        // if no POAs activated, we don't look further
        if( rootpoa == null || basicAdapter == null )
        {
            return null;
        }

        String orbImplName = Environment.getProperty( "jacorb.implname", "" );

        try
        {
            refImplName =
                org.jacorb.poa.util.POAUtil.extractImplName( d.getObjectKey() );
        }
        catch( org.jacorb.poa.except.POAInternalError pie )
        {
            if( logger.isDebugEnabled() )
            {
                logger.debug("findPOA: reference generated by foreign POA");
            }
            return null;
        }

        if( refImplName == null )
        {
            if( orbImplName.length() > 0 )
            {
                if( logger.isDebugEnabled() )
                {
                    logger.debug("findPOA: impl_name mismatch");
                }
                return null;
            }
        }
        else
        {
            if( !(orbImplName.equals( refImplName ) ))
            {
                if( logger.isDebugEnabled() )
                {
                    logger.debug("findPOA: impl_name mismatch");
                }
                return null;
            }
        }

        /* ImRs have the same impl name so if an ImR needs to communicate
         * with another ImR we need to work out whether the request if for
         * this ImR or for another ImR.  Without this isLocal would return
         * true and the ImR would end up talking to itself
         */
        if (Environment.isImR())
        {
            String imrKey = (String) objectKeyMap.get("ImplementationRepository");

            /* If this is the ImR and the request is NOT for this ImR */
            if (!(POAUtil.convert(d.getObjectKey()).equals(imrKey)))
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("findPOA: ImR request to ImR");
                }

                return null;
            }
        }

        try
        {
            org.jacorb.poa.POA tmp_poa = (org.jacorb.poa.POA)rootpoa;
            String poa_name =
                org.jacorb.poa.util.POAUtil.extractPOAName( d.getObjectKey() );

            /* strip scoped poa name (first part of the object key before "::",
             *  will be empty for the root poa
             */
            scopes = POAUtil.extractScopedPOANames (poa_name);

            for( int i = 0; i < scopes.size(); i++)
            {
                res = ((String)scopes.get( i ));

                if( res.equals (""))
                    break;

                /* the following is a  call to a method in the private
                   interface between the ORB  and the POA. It does the
                   necessary    synchronization    between   incoming,
                   potentially concurrent  requests to activate  a POA
                   using its  adapter activator. This  call will block
                   until  the correct  POA is  activated and  ready to
                   service    requests.    Thus,    concurrent   calls
                   originating  from a  single,  multi-threaded client
                   will be serialized  because the thread that accepts
                   incoming  requests  from   the  client  process  is
                   blocked.  Concurrent  calls from other destinations
                   are not  serialized unless they  involve activating
                   the same adapter.  */

                try
                {
                    tmp_poa = tmp_poa._getChildPOA( res );
                }
                catch ( org.jacorb.poa.except.ParentIsHolding p )
                {
                    if( logger.isDebugEnabled() )
                    {
                        logger.debug("findPOA: holding adapter");
                    }
                    return null;
                }
            }
            byte[] objectId =
                org.jacorb.poa.util.POAUtil.extractOID( ref );

            if( tmp_poa.isSystemId()
               && ! tmp_poa.previouslyGeneratedObjectId( objectId ))
            {
                if( logger.isDebugEnabled() )
                {
                    logger.debug("findPOA: not a previously generated object key.");
                }
                return null;
            }

            return tmp_poa;
        }
        catch( Exception e )
        {
            logger.debug(e.getMessage()); // TODO
        }
        if( logger.isDebugEnabled() )
        {
            logger.debug("findPOA: nothing found");
        }
        return null;
    }


    public ClientConnectionManager getClientConnectionManager()
    {
        return clientConnectionManager;
    }

    public GIOPConnectionManager getGIOPConnectionManager()
    {
        return giop_connection_manager;
    }


    /**
     * Take a string rather then a Delegate object to prevent data race
     * warning.
     */
    synchronized void _release( String iorString )
    {
        knownReferences.remove( iorString );
    }


    /**
     * This method creates a policy  with the given type and the given
     * value.
     *
     * @param type The policies type.
     * @param value The policies value.
     * @exception org.omg.CORBA.PolicyError There is no PolicyFactory for the
     * given type or the policy creation failed.
     * @see org.omg.PortableInterceptor.PolicyFactory
     */

    public org.omg.CORBA.Policy create_policy( int type, org.omg.CORBA.Any value )
        throws org.omg.CORBA.PolicyError
    {
        switch (type)
        {
            case MAX_HOPS_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.MaxHopsPolicy (value);
            case QUEUE_ORDER_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.QueueOrderPolicy (value);
            case REBIND_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.RebindPolicy (value);
            case RELATIVE_REQ_TIMEOUT_POLICY_TYPE.value:
              return new
                org.jacorb.orb.policies.RelativeRequestTimeoutPolicy (value);
            case RELATIVE_RT_TIMEOUT_POLICY_TYPE.value:
              return new
                org.jacorb.orb.policies.RelativeRoundtripTimeoutPolicy (value);
            case REPLY_END_TIME_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.ReplyEndTimePolicy (value);
            case REPLY_PRIORITY_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.ReplyPriorityPolicy (value);
            case REPLY_START_TIME_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.ReplyStartTimePolicy (value);
            case REQUEST_END_TIME_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.RequestEndTimePolicy (value);
            case REQUEST_PRIORITY_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.RequestPriorityPolicy (value);
            case REQUEST_START_TIME_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.RequestStartTimePolicy (value);
            case ROUTING_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.RoutingPolicy (value);
            case SYNC_SCOPE_POLICY_TYPE.value:
                return new
                    org.jacorb.orb.policies.SyncScopePolicy (value);
            default:
                Integer key = new Integer(type);
                if ( policy_factories == null ||
                     (! policy_factories.containsKey(key)) )
                    throw new org.omg.CORBA.PolicyError();
                PolicyFactory factory =
                    (PolicyFactory)policy_factories.get(key);
                return factory.create_policy(type, value);
        }
    }


    /**
     * Tests if a policy factory is present for the given type.
     */
    public boolean hasPolicyFactoryForType(int type)
    {
        return (policy_factories != null &&
                policy_factories.containsKey( new Integer(type)) );
    }

    public org.omg.CORBA.ContextList create_context_list()
    {
        throw new org.omg.CORBA.NO_IMPLEMENT ();
    }

    public org.omg.CORBA.Environment create_environment()
    {
        throw new org.omg.CORBA.NO_IMPLEMENT ();
    }

    public org.omg.CORBA.portable.OutputStream create_output_stream()
    {
        return new CDROutputStream(this);
    }

    org.omg.IOP.IOR createIOR (String repId,
                               byte[] objectKey,
                               boolean _transient,
                               org.jacorb.poa.POA poa,
                               Map policy_overrides)
    {
        if (!_transient && Environment.useImR() && Environment.useImREndpoint())
        {
            // Build an IOR with our object key but the ImR's endpoints.
            org.omg.IOP.IOR ior = createIOR(repId,
                                            objectKey,
                                            poa.getImREndpoints());

            if (Environment.imrProxyHost() != null || Environment.imrProxyPort() == -1)
            {
                // We need to handpatch a proxy host and port value into any IIOP profiles
                for (int i = 0; i < ior.profiles.length; i++)
                {
                    if (ior.profiles[i].tag == org.omg.IOP.TAG_INTERNET_IOP.value)
                    {
                        IIOPProfile profile = new IIOPProfile(ior.profiles[i].profile_data);
                        profile.patchPrimaryAddress(Environment.imrProxyHost(),
                                                    Environment.imrProxyPort());
                        ior.profiles[i] = profile.asTaggedProfile();
                    }
                }
            }

            return ior;
        }
        else
        {
            // Normal operation - no ImR.
            List profiles     = new ArrayList();
            Map  componentMap = new HashMap();

            // Retrieve the profiles for this object ref
            getProfiles(poa, policy_overrides, componentMap, profiles);

            // marshal the profiles into the IOR and return
            TaggedProfile[] tps = new TaggedProfile [profiles.size()];
            TaggedProfileHolder      tp = new TaggedProfileHolder();
            TaggedComponentSeqHolder tc = new TaggedComponentSeqHolder();

            for (int i=0; i < profiles.size(); i++)
            {
                Profile p = (Profile)profiles.get(i);
                p.set_object_key (objectKey);
                TaggedComponentList c =
                    (TaggedComponentList)componentMap.get (new Integer (p.tag()));
                tc.value = c.asArray();
                p.marshal (tp, tc);
                tps[i] = tp.value;
            }

            return new IOR (repId, tps);
        }
    }

    public void getProfiles
    (
        org.jacorb.poa.POA poa,
        Map policy_overrides,
        Map componentMap,
        List profiles
    )
    {
        for (Iterator i = basicAdapter.getEndpointProfiles().iterator();
             i.hasNext();)
        {
            Profile profile = (Profile)i.next();

            profiles.add (profile);

            TaggedComponentList profileComponents = new TaggedComponentList();
            profileComponents.addComponent (create_ORB_TYPE_ID());
            componentMap.put (new Integer (profile.tag()), profileComponents);

            // use proxy address if necessary
            if (profile instanceof IIOPProfile)
            {
                patchAddress ((IIOPProfile)profile);
            }
        }

        TaggedComponentList multipleComponents = new TaggedComponentList();
        componentMap.put (new Integer (TAG_MULTIPLE_COMPONENTS.value),
                          multipleComponents);

        // invoke IOR interceptors
        if ((interceptor_manager != null) &&
            interceptor_manager.hasIORInterceptors())
        {
            IORInfoImpl info = new IORInfoImpl (this,
                                                poa,
                                                componentMap,
                                                policy_overrides,
                                                profiles);
            try
            {
                interceptor_manager.getIORIterator().iterate( info );
            }
            catch (Exception e)
            {
                Debug.output(2, e);
            }
        }

        if (Environment.giopMinorVersion() == 0 || Environment.giopAdd_1_0_Profiles())
        {
            // add GIOP 1.0 profile if necessary
            IIOPProfile iiopProfile = findIIOPProfile (profiles);
            if (iiopProfile != null)
            {
                Profile profile_1_0 = iiopProfile.to_GIOP_1_0();
                profiles.add (profile_1_0);

                // shuffle all components over into the multiple components profile
                TaggedComponentList iiopComponents =
                    (TaggedComponentList)componentMap.get
                                           (new Integer (TAG_INTERNET_IOP.value));

                multipleComponents.addAll (iiopProfile.getComponents());
                multipleComponents.addAll (iiopComponents);

                // if we only want GIOP 1.0, remove the other profile
                if (Environment.giopMinorVersion() == 0)
                {
                    profiles.remove (iiopProfile);
                }
            }
        }

        if (!multipleComponents.isEmpty())
        {
            profiles.add(new org.jacorb.orb.giop.MultipleComponentProfile(multipleComponents));
        }
    }

    /**
    * Method that can create an object ref from a repository ID, object key and a sequence of
    * imr::EndpointProfile's.
    * <p>
    * Is used to create object refs for persistent objects which are running under
    * control of the IMR by combining the object's type and object key with the endpoints
    * of one or more IMR's.
    * <p>
    * Conversely, is also used by the IMR to combine registered POA endpoints with corresponding
    * object keys and type ID's to create an object reference that can be used in the LocationForward
    * exception it returns to clients.
    */
    public org.omg.IOP.IOR createIOR( String repId,
                               byte[] objectKey,
                               EndpointProfile[] endpoints)
    {
        // Allocate a TaggedProfile array to hold all the endpoint profiles
        TaggedProfile[] tps = new TaggedProfile[endpoints.length];

        // We assume we checked the encapsulation tag sometime before calling this method
        // and we can go ahead with each of these profiles encapsulating them the only way
        // we know how.
        for (int i = 0; i < endpoints.length; i++)
        {
            // Start a CDR stream for the profile_data of this tagged profile
            CDROutputStream profileDataStream = new CDROutputStream();

            switch(endpoints[i].encapsulation)
            {
                case ImRReadyProfile.ALREADY_COMPLETE:
                    // No need to open a fresh encapsulation. Just write byte by byte.
                    profileDataStream.write_octet_array(endpoints[i].addressProfile, 0, endpoints[i].addressProfile.length);
                    break;
                case ImRReadyProfile.BIG_ENDIAN_TAG_INTERNET_IOP:
                    // Write the opaque AddressProfile bytes for this profile...
                    profileDataStream.write_octet_array
                        (endpoints[i].addressProfile, 0, endpoints[i].addressProfile.length);

                    // ... then the object key
                    profileDataStream.write_long(objectKey.length);
                    profileDataStream.write_octet_array(objectKey,0,objectKey.length);

                    // Write the length of the TaggedComponent sequence.
                    profileDataStream.write_long(endpoints[i].taggedComponents.length);

                    // Write the TaggedComponents
                    for (int j = 0; j < endpoints[i].taggedComponents.length; j++)
                    {
                        TaggedComponentHelper.write(profileDataStream, endpoints[i].taggedComponents[j]);
                    }
            }

            tps[i] = new TaggedProfile(endpoints[i].tag, profileDataStream.getBufferCopy());
        }

        return new IOR (repId, tps);
    }


    private TaggedProfile createMultipleComponentsProfile
                                  (TaggedComponentList components)
    {
        CDROutputStream out = new CDROutputStream (this);
        out.beginEncapsulatedArray();
        MultipleComponentProfileHelper.write (out, components.asArray());
        return new TaggedProfile
        (
            TAG_MULTIPLE_COMPONENTS.value,
            out.getBufferCopy()
        );
    }

    /**
     * Finds the first IIOPProfile in the given List of Profiles,
     * and returns it.  If no such profile is found, this method
     * returns null.
     */
    private IIOPProfile findIIOPProfile (List profiles)
    {
        for (Iterator i = profiles.iterator(); i.hasNext();)
        {
            Profile p = (Profile)i.next();
            if (p instanceof IIOPProfile)
                return (IIOPProfile)p;
        }
        return null;
    }

    public org.omg.CORBA.Context get_default_context ()
    {
        throw new org.omg.CORBA.NO_IMPLEMENT ();
    }


    /**
     * used from the POA
     * @return the basic adapter used by this ORB instance
     */
    public org.jacorb.orb.BasicAdapter getBasicAdapter()
    {
        if( basicAdapter == null )
            throw new INITIALIZE("Adapters not initialized; resolve RootPOA.");
        return basicAdapter;
    }


    /**
     * getPOACurrent
     */

    public org.jacorb.poa.Current getPOACurrent()
    {
        if (poaCurrent == null)
        {
            poaCurrent = org.jacorb.poa.Current._Current_init();
        }
        return poaCurrent;
    }

    /**
     * called by POA to create an IOR
     *
     * @param poa the calling POA
     * @param object_key
     * @param rep_id
     * @param _transient is the new reference transient or persistent
     * @return a new CORBA Object reference
     */
    public org.omg.CORBA.Object getReference( org.jacorb.poa.POA poa,
                                              byte[] object_key,
                                              String rep_id,
                                              boolean _transient )
    {
        if( rep_id == null )
            rep_id = "IDL:omg.org/CORBA/Object:1.0";

        org.omg.IOP.IOR ior =
        createIOR( rep_id, object_key, _transient, poa, null );

        Delegate d = new Delegate( this, ior );

        return d.getReference( poa );
    }

    public org.jacorb.poa.POA getRootPOA()
        throws org.omg.CORBA.INITIALIZE
    {
        if( rootpoa == null )
        {
            rootpoa = org.jacorb.poa.POA._POA_init(this);
            rootpoa._addPOAEventListener( this );

            basicAdapter = new BasicAdapter( this,
                                             rootpoa,
                                             transport_manager,
                                             giop_connection_manager );
        }
        return rootpoa;
    }

    /**
     * <code>findImR</code> handles attempting to find other IMRs and
     * (re)register this server's POAs with them.
     * <p>
     * You should not be calling this method. In particular you must
     * not call this method without the possessing imrUpdateLock.
     */
    private boolean findImR()
    {
        Repository newImR = null;

        // First we always try to resolve initial references
        try
        {
            // Lookup the implementation repository
            newImR = RepositoryHelper.narrow
                (resolve_initial_references ("ImplementationRepository"));

        }
        catch( InvalidName e )
        {
            if (logger.isInfoEnabled())
            {
                logger.info
                    ("Failed to resolve initial references for the Implementation Repository");
            }
            // throw new INTERNAL ("Unable to resolve reference to OpenFusion IMR");
        }
        catch (Exception e)
        {
            if (logger.isInfoEnabled())
            {
                logger.info
                    ("Failed to resolve initial references for the Implementation Repository");
            }
        }

        if (newImR != null && newImR._non_existent())
        {
            newImR = null;
        }


        // If we haven't found an ImR by resolving initial references we'll try
        // UDP broadcast.

        int uport = Environment.getIntPropertyWithDefault("jacorb.imr.udp_port", 0);

        if (newImR == null && rootpoa != null && uport > 0)
        {
            try
            {
                if (locator == null)
                {
                    locator = new IMRLocator(this, uport );
                }

                locator.findIMRs();

                if (logger.isDebugEnabled())
                {
                    logger.debug("Waiting 3 seconds for IMR responses" );
                }

                Thread.sleep( 3000 );

                newImR = RepositoryHelper.narrow((org.omg.CORBA.Object)locator.getIMRs().get(0));
            }
            catch (Exception e)
            {
                if (logger.isInfoEnabled())
                {
                    logger.info("Failed in attempt to locate ImR by UDP on port: " + uport );
                }
                newImR = null;
            }
        }

        if (newImR != null)
        {
            // We found a 'new' ImR of some sort.

            // Update the imr attribute on this ORB.
            imr = newImR;

            if (rootpoa != null && ! POAUtil.isHolding(rootpoa.the_POAManager().get_state()))
            {
                try
                {
                    // There is a root poa and it's active
                    // There may be previously registered POAs to re-register.
                    rootpoa.reRegisterWithImR();
                }
                catch (Exception e)
                {

                    // Any exception here should be unexpected and will indicate that
                    // The ImR has died. Before the exception was re-thrown it will have prompted
                    // *another* imrDead(imr) call which will have attempted to locate another ImR.
                    // It must have failed or the exception wouldn't have been re-thrown.
                    // We therefore report and give up for now.
                    if (logger.isErrorEnabled())
                    {
                        logger.error("Failed to re-register persistent POAs with a new ImR");
                    }
                    return false;
                }
            }
            return true;
        }
        else
        {
            return false;
        }
    }

     /**
     * @return - true if ORB is initialized by an applet and
     * appligator use is switched on
     */

    public boolean isApplet()
    {
        return applet != null;
        //return applet != null && Environment.useAppligator();
    }


    public Applet getApplet()
    {
        return applet;
    }

    public String[] list_initial_services()
    {
        List l = new ArrayList();

        for( Iterator e = initial_references.keySet().iterator();
             e.hasNext(); l.add( e.next() ) );

        String [] initial_services =
            new String[ services.length + l.size()];

        l.toArray( initial_services );

        System.arraycopy( services, 0, initial_services, l.size(), services.length );
        return initial_services;
    }

    /**
     * An operation from the POAListener interface. Whenever a new POA is
     * created, the ORB is notified.
     *
     * @param <code>poa</code> the poa that has just been created.
     */
    public void poaCreated( org.jacorb.poa.POA poa )
    {
        /*
         * Add this orb as the child poa's event listener. This means that the
         * ORB is always a listener to all poa events!
         */
        poa._addPOAEventListener( this );

        /* If the new POA has a persistent lifetime policy, it is registered
         * with the implementation repository if there is one and the
         * use_imr policy is set via the "jacorb.orb.use_imr" property
         */

        if( poa.isPersistent() )
        {
            persistentPOACount++;
        }


    }

    /**
     * <code>getImR</code> returns the imr, initing if needed. Package access
     * so Delegate can call it. May throw an INTERNAL if the multicast fails.
     *
     * @return an <code>ImrAccess</code> value
     */
    public Repository getImR ()
    {
        synchronized(imrUpdateLock)
        {
            if (imr == null)
            {
                try
                {
                    findImR();
                }
                catch (INTERNAL e)
                {

                }
            }
            return imr;
        }
    }

    /**
    * Used to report that an ImR is dead. Will prompt a search for a new
    * one and any necessary persistent POA re-registration if one is found.
    * @return Indicates whether another ImR is available. <code>false</code>
    * if the dead ImR was, as Obi Wan would say, "our last hope";
    * <code>true</code> if, in the words of Yoda, "There is another".
    */
    public boolean imrDead(Repository corpse)
    {
        synchronized(imrUpdateLock)
        {
            if (!imrsEqual(imr, corpse))
            {
                // A new ImR has been found
                return true;
            }
            else
            {
                return findImR();
            }
        }
    }

    /**
    * We can't use any of the 'usual' equality tests here.
    * _is_equivalent will always return true because IMRs share the same
    * object key. A bitwise comparison of the addressing data will preclude
    * a replacement by a restarted IMR running on fixed address.
    * We fall back on object equality for now.
    * @todo Place an IMR numeric identity tagged component in IMR service ior's
    * and use that as a first choice check, falling back on object equality
    * for corbaloc and any other non-tagged IORs.
    */
    protected boolean imrsEqual(Repository imr1, Repository imr2)
    {
        return imr1 == imr2;
    }

    /**
    * Return an object reference for an ImR control servant for this server.
    * Will initialise and bind to the root poa if required.
    * @see #getImRServerImpl()
    */
    public Server getActiveServerHandler()
    {
        if (activeServerHandler == null)
        {
            imrServerImpl = new ServerImpl(this);

            imrServerImpl.heartbeat
                (Environment.getIntPropertyWithDefault
                    ("jacorb.imr.heartbeat", 0));

            activeServerHandler = imrServerImpl.installOnRootPOA();
        }
        return activeServerHandler;
    }

    /**
    * Return a handle to the ImR control servant implementation. Prior
    * to POA manager activation the object ref (also held by the ORB)
    * cannot be used to set attributes.
    * @see #()
    * @return The servant implementation or <code>null</code> if none
    * has been initialised.
    */
    public ServerImpl getImRServerImpl()
    {
        return imrServerImpl;
    }

    /**
     * Replace the server address in profile with a proxy address if necessary.
     */
    private void patchAddress (IIOPProfile profile)
    {
        profile.patchPrimaryAddress (Environment.iorProxyHost(),
                                     Environment.iorProxyPort());
    }

    /**
     * Creates an ORB_TYPE_ID tagged component for JacORB.
     */
    private TaggedComponent create_ORB_TYPE_ID()
    {
        CDROutputStream orbIDComponentDataStream = new CDROutputStream( this );

        orbIDComponentDataStream.beginEncapsulatedArray();
        orbIDComponentDataStream.write_long( ORBConstants.JACORB_ORB_ID );

        byte []result = orbIDComponentDataStream.getBufferCopy();

        orbIDComponentDataStream.close();

        return new TaggedComponent
        (
            TAG_ORB_TYPE.value,
            result
        );
    }


    /**
     * <code>getServerAddress</code> returns the address to use to locate the
     * server.  Note that this address will be overwritten by the ImR address in
     * the IOR of persistent servers if the use_imr and use_imr_endpoint
     * properties are switched on
     *
     * @return a <code>String</code>, the address for the server.
     */
    private String getServerAddress()
    {
        String address = Environment.getProperty( "jacorb.ior_proxy_host" );

        if( address == null )
        {
            //property not set
            address = getBasicAdapter().getAddress();
        }
        else
        {
            if (logger.isInfoEnabled())
            {
                logger.info("Using proxy host " + address + " in IOR" );
            }
        }

        return address;
    }


    /**
     * <code>getServerPort</code> returns the port to use to locate the server.
     * Note that this port will be overwritten by the ImR port in the IOR of
     * persistent servers if the use_imr and use_imr_endpoint properties are
     * switched on.
     *
     * @return an <code>int</code>, the port for the server.
     */
    private int getServerPort()
    {
        String port_str = Environment.getProperty( "jacorb.ior_proxy_port" );
        int port = -1;

        if( port_str != null )
        {
            try
            {
                port = Integer.parseInt( port_str );
            }
            catch( NumberFormatException nfe )
            {
                throw new BAD_QOS( "Unable to create integer from string >>" +
                                   port_str + "<<. " +
                                   "(check property \"jacorb.ior_proxy_port\")" );
            }

            if( port < 0 )
            {
                throw new BAD_QOS( "Negative port numbers are not allowed! " +
                                   "(check property \"jacorb.ior_proxy_port\")" );
            }

            if (logger.isInfoEnabled())
            {
                logger.info("Using proxy port " + port + " in IOR" );
            }
        }
        else
        {
            //property not set
            port = getBasicAdapter().getPort();
        }

        return port;
    }

    /**
     * Method to check if a Proxy host is configured for the IMR via the
     * <code>jacorb.imr.ior_proxy_host</code> property.
     * @param imrAddress The actual host name or IP of the IMR.
     * @return The proxy host value, if configured, else the supplied actual host.
     */
    private String getIMRAddressForIOR(String imrAddress)
    {
        String imrProxyHost = Environment.getProperty("jacorb.imr.ior_proxy_host");
        return (imrProxyHost == null ? imrAddress : imrProxyHost);
    }

    /**
     * Method to check if a Proxy port is configured for the IMR via the
     * <code>jacorb.imr.ior_proxy_port</code> property.
     * @param imrPort The port number that the IMR is really running on.
     * @return The proxy port number, if configured, else the supplied actual port.
     */
    private int getIMRPortForIOR(int imrPort)
    {
        String imrProxyPort = Environment.getProperty("jacorb.imr.ior_proxy_port");
        if (imrProxyPort != null)
        {
            try
            {
                imrPort = Integer.parseInt(imrProxyPort);
            }
            catch (NumberFormatException nfe)
            {
                if (logger.isWarnEnabled())
                {
                    logger.warn("IMR Proxy Port is configured to a none integer value. " +
                                "Check the value of property \"jacorb.imr.ior_proxy_port");
                }
            }
        }
        return imrPort;
    }

    public void poaStateChanged(org.jacorb.poa.POA poa, int new_state)
    {
        if( ( new_state == org.jacorb.poa.POAConstants.DESTROYED ||
              new_state == org.jacorb.poa.POAConstants.INACTIVE )  &&
            poa.isPersistent()
            )
        {
            --persistentPOACount;
        }
    }


    public void referenceCreated(org.omg.CORBA.Object o) {}

    public boolean get_service_information( short service_type,
                                            org.omg.CORBA.ServiceInformationHolder service_information)
    {
        //          if (( service_type == org.omg.CORBA.Security.value ) && Environment.supportSSL ())
        //          {
        //              byte options[] = new byte [5]; // ServiceOption[]
        //              options[0] = (byte)org.omg.Security.SecurityLevel1.value;
        //              options[1] = (byte)org.omg.Security.SecurityLevel2.value;
        //              options[2] = (byte)org.omg.Security.ReplaceORBServices.value;
        //              options[3] = (byte)org.omg.Security.ReplaceSecurityServices.value;
        //              options[4] = (byte)org.omg.Security.CommonInteroperabilityLevel0.value;
        //              org.omg.CORBA.ServiceDetail details[] = new org.omg.CORBA.ServiceDetail [2];
        //              details[0].service_detail_type = org.omg.Security.SecureTransportType.value;
        //              details[0].service_detail = org.jacorb.security.ssl.SSLSetup.getMechanismType().getBytes();
        //              details[1].service_detail_type = org.omg.Security.SecureTransportType.value;
        //              details[1].service_detail = "001010011".getBytes(); // AuditId, _PublicId, AccessId
        //              return true;
        //          }
        //          else return false;
        throw new org.omg.CORBA.NO_IMPLEMENT ();
    }

    /**
     * resolve_initial_references
     */

    public org.omg.CORBA.Object resolve_initial_references(String identifier)
        throws org.omg.CORBA.ORBPackage.InvalidName
    {
        if ( initial_references.containsKey(identifier) )
        {
            return (org.omg.CORBA.Object)initial_references.get(identifier);
        }
        else
        {
            org.omg.CORBA.Object obj = null;
            String url = Environment.getProperty("ORBInitRef." + identifier);

            if( url != null )
            {
                try
                {
                    obj = this.string_to_object( url );
                }
                catch( Exception e )
                {
                    if (logger.isErrorEnabled())
                    {
                        logger.error( "Could not create initial reference for \"" +
                                      identifier + "\"\n" + "Please check property \"ORBInitRef." +
                                      identifier + '\"' );
                    }
                    if (logger.isDebugEnabled())
                    {
                        logger.debug( e.getMessage() );
                    }

                    throw new org.omg.CORBA.ORBPackage.InvalidName();
                }
            }
            /* "special" behavior follows */
            else if (identifier.equals ("NameService") && isApplet ())
            {
                // try to get location of URL with ns's IOR from a file
                //     called NameService.ior at the applet's host
                String ior_str =
                    org.jacorb.util.ObjectUtil.readURL("http://"
                                                       + applet.getCodeBase().getHost()
                                                       + "/"
                                                       + "NameService.ior");

                obj = this.string_to_object (ior_str);

                if (obj != null)
                {
                    if (! obj._is_a (org.omg.CosNaming.NamingContextHelper.id ()))
                    {
                        obj = null;
                    }
                }
            }
            else if( identifier.equals("RootPOA") )
            {
                return (org.omg.CORBA.Object)getRootPOA();
            }
            else if( identifier.equals("POACurrent") )
            {
                return (org.omg.CORBA.Object)getPOACurrent();
            }
            else if( identifier.equals("SecurityCurrent") )
            {
                if( securityCurrent == null )
                {
                    try
                    {
                        Class currentClass = Environment.classForName( "org.jacorb.security.level2.CurrentImpl" );

                        Constructor constr = currentClass.getConstructor( new Class[]{
                            org.omg.CORBA.ORB.class });

                        securityCurrent = (org.omg.SecurityLevel2.Current)
                            constr.newInstance( new Object[]{ this });

                        Method init = currentClass.getDeclaredMethod( "init",
                                                                      new Class[0] );
                        init.invoke( securityCurrent, new Object[0] );
                    }
                    catch (Exception e)
                    {
                        if (logger.isWarnEnabled())
                        {
                            logger.warn(e.getMessage());
                        }
                    }
                }

                obj = securityCurrent;
            }
            else if( identifier.equals("DynAnyFactory") )
            {
                obj = new org.jacorb.orb.dynany.DynAnyFactoryImpl
                    ( this, logger );
            }
            else if( identifier.equals("PICurrent") )
            {
                return piCurrent;
            }
            else if( identifier.equals("ORBPolicyManager") )
            {
                return getPolicyManager();
            }
            else if( identifier.equals("CodecFactory") )
            {
                obj = new CodecFactoryImpl(this);
            }
            else
            {
                throw new org.omg.CORBA.ORBPackage.InvalidName ();
            }

            if (obj != null)
            {
                initial_references.put (identifier, obj);
            }

            return obj;
        }
    }


    PolicyManager getPolicyManager()
    {
        return policyManager;
    }

    /**
     * Register a reference, that will be returned on subsequent calls
     * to resove_initial_references(id). <br>
     * The references "RootPOA", "POACurrent" and "PICurrent" can be set,
     * but will not be resolved with the passed in references.
     * <p>
     * Overrides  register_initial_reference() in org.omg.CORBA_2_5.ORB
     *
     * @param id The references human-readable id, e.g. "MyService".
     * @param obj The objects reference.
     * @exception InvalidName A reference with id has already been registered.
     */

    public void register_initial_reference( String id, org.omg.CORBA.Object obj )
        throws InvalidName
    {
        if (id == null || id.length() == 0 ||
            initial_references.containsKey(id) )
        {
            throw new InvalidName();
        }
        else
        {
            if (logger.isDebugEnabled())
            {
                logger.debug( "Registering initial ref " + id );
            }
            initial_references.put(id, obj);
        }
    }

    public void run()
    {
        if (logger.isInfoEnabled())
        {
            logger.info("ORB run");
        }

        try
        {
            synchronized( orb_synch )
            {
                while( run )
                {
                    orb_synch.wait();
                }
            }
        }
        catch (InterruptedException ex)
        {
        }
        if (logger.isInfoEnabled())
        {
            logger.info("ORB run, exit");
        }
    }


    public void send_multiple_requests_oneway( org.omg.CORBA.Request[] req )
    {
        for( int i = 0; i < req.length; i++ )
        {
            req[i].send_oneway();
        }
    }

    public void send_multiple_requests_deferred( org.omg.CORBA.Request[] req )
    {
        for( int i = 0; i < req.length; i++ )
        {
            req[i].send_deferred();
        }
    }

    public boolean poll_next_response()
    {
        if( requests.size () == 0 )
        {
            throw new org.omg.CORBA.BAD_INV_ORDER
                ( 11, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        }

        synchronized( requests )
        {
            Request req;
            Iterator iter = requests.iterator();
            while( iter.hasNext() )
            {
                req = (Request)iter.next();
                if( req.poll_response() )
                {
                    request = req;
                    return true;
                }
            }
        }
        return false;
    }

    public org.omg.CORBA.Request get_next_response ()
    {
        if( requests.size () == 0 )
        {
            throw new org.omg.CORBA.BAD_INV_ORDER
                ( 11, org.omg.CORBA.CompletionStatus.COMPLETED_NO );
        }

        synchronized( requests )
        {
            Request req = null;
            if( request != null )
            {
                request.get_response();
                req = request;
                request = null;
                return req;
            }

            Iterator iter;
            while( true )
            {
                iter = requests.iterator();
                while( iter.hasNext() )
                {
                    req = (Request)iter.next();
                    if( req.poll_response() )
                    {
                        req.get_response();
                        return req;
                    }
                }
            }
        }
    }

    public void addRequest( org.omg.CORBA.Request req )
    {
        requests.add( req );
    }

       public void removeRequest( org.omg.CORBA.Request req )
    {
        requests.remove( req );
    }


    /**
     * called from ORB.init()
     */

    protected void set_parameters( String[] args, java.util.Properties props )
    {
        if( props != null )
        {
            _props = props;
            Environment.addProperties( props );
        }

        /*
         * find -ORBInitRef args and add them to Environment
         * (overwriting existing props).
         */

        if( args != null )
        {
            _args = args;
            for( int i = 0; i < args.length; i++ )
            {
                String arg = args[ i ].trim();

                if (arg.startsWith("-ORB"))
                {
                    // Strip the leading '-'.
                    arg = arg.substring( 1 );

                    // Look for an '=' in the argument
                    int equals_pos = arg.indexOf('=');

                    if (equals_pos == -1)
                    {
                        //This is the compliant form like: -ORBInitRef <name>=<val>
                        //Is there a next arg?
                        if( (args.length - 1) < (i + 1) )
                        {
                            throw new BAD_PARAM("Invalid ORBInitRef format: -ORB<option> argument without value" );
                        }
                        if (arg.indexOf('.') != -1)
                        {
                            // ORBInitRef.<name> value
                            throw new BAD_PARAM("-ORBInitRef.xxx yyy is not a valid format. Use -ORBInitRef name=value");
                        }

                        String prop = args[ ++i ].trim();

                        //find the equals char that separates prop name from
                        //prop value
                        equals_pos = prop.indexOf('=');
                        if (equals_pos == -1)
                        {
                            // Form is -ORBWhatever value
                            Environment.setProperty(arg, prop);
                        }
                        else
                        {
                            // Form is -ORBWhatever name=value
                            Environment.setProperty( arg + "." +
                                                     prop.substring( 0, equals_pos ),
                                                     prop.substring( equals_pos + 1) );
                        }
                    }
                    else
                    {
                        //This is the wrong jacorb form like: -ORBInitRef.<name>=<val>

                        //add the property to environment
                        Environment.setProperty( arg.substring( 0, equals_pos ),
                                                 arg.substring( equals_pos + 1) );
                    }
                }
            }
        }

        policyManager = new PolicyManager( this );
        transport_manager = new TransportManager( this );
        giop_connection_manager = new GIOPConnectionManager();
        clientConnectionManager =
            new ClientConnectionManager( this,
                                         transport_manager,
                                         giop_connection_manager );

        String s = Environment.getProperty( "jacorb.hashtable_class" );
        if( s == null || s.length() == 0 )
        {
            if (logger.isInfoEnabled())
            {
                logger.info("Property \"jacorb.hashtable_class\" not present. Will use default hashtable implementation" );
            }
            knownReferences = new HashMap();

        }
        else
        {
            try
            {
                knownReferences = (Map) Environment.classForName( s ).newInstance();
            }
            catch( Exception e )
            {
                if (logger.isInfoEnabled())
                {
                    logger.info(e.getMessage());
                }
                knownReferences = new HashMap();
            }
        }

        objectKeyMap =
            Environment.getProperties("jacorb.orb.objectKeyMap", true );

        String versionProperty =
            Environment.getProperty("jacorb.orb.print_version");

        if( versionProperty != null &&
            versionProperty.equals("on") )
        {
            System.out.println("\tJacORB V " + versionString +
                               ", www.jacorb.org");
            System.out.println("\t(C) Gerald Brose, XTRADYNE Technologies/FU Berlin, " +
                               dateString + "\n\t(PrismTech Build Source Distribution)");
        }

        interceptorInit();
    }


    /**
     * Initialization method, called from within the super class
     * org.omg.CORBA.ORB
     */
    protected void set_parameters
        ( java.applet.Applet app, java.util.Properties  props )
    {
        applet = app;
        _props = props;

        Environment.addProperties( props );

        transport_manager = new TransportManager( this );
        giop_connection_manager = new GIOPConnectionManager();
        policyManager = new PolicyManager( this );

        clientConnectionManager =
            new ClientConnectionManager(
                this,
                transport_manager,
                giop_connection_manager);

        String s = Environment.getProperty( "jacorb.hashtable_class" );
        if( s == null || s.length() == 0 )
        {
            if (logger.isInfoEnabled())
            {
                logger.info("Property \"jacorb.hashtable_class\" not present. Will use default hashtable implementation" );
            }
            knownReferences = new HashMap();

        }
        else
        {
            try
            {
                knownReferences = (Map)Environment.classForName( s ).newInstance();
            }
            catch( Exception e )
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug( e.getMessage());
                }
                knownReferences = new HashMap();
            }
        }


        // unproxyTable = new Hashtable();

        interceptorInit();
    }

    /**
     * This method retrieves the ORBInitializer-Names from the Environment,
     * and runs them.
     */

    private void interceptorInit()
    {
        // get instances from Environment
        Vector orb_initializers = Environment.getORBInitializers ();

        if (orb_initializers.size () > 0)
        {
            ORBInitInfoImpl info = new ORBInitInfoImpl (this);
            ORBInitializer init;

            // call pre_init in ORBInitializers
            for (int i = 0; i < orb_initializers.size(); i++)
            {
                try
                {
                    init = (ORBInitializer) orb_initializers.elementAt (i);
                    init.pre_init (info);
                }
                catch (Exception e)
                {
                    if (logger.isWarnEnabled())
                    {
                        logger.warn( e.getMessage());
                    }
                }
            }

            //call post_init on ORBInitializers
            for (int i = 0; i < orb_initializers.size (); i++)
            {
                try
                {
                    init = (ORBInitializer) orb_initializers.elementAt (i);
                    init.post_init (info);
                }
                catch (Exception e)
                {
                    if (logger.isWarnEnabled())
                    {
                        logger.warn( e.getMessage());
                    }
                }
            }

            //allow no more access to ORBInitInfo from ORBInitializers
            info.setInvalid ();

            Vector client_interceptors = info.getClientInterceptors ();
            Vector server_interceptors = info.getServerInterceptors ();
            Vector ior_intercept = info.getIORInterceptors ();

            hasClientInterceptors = (client_interceptors.size () > 0);
            hasServerInterceptors = (server_interceptors.size () > 0);

            if (hasClientInterceptors || hasServerInterceptors || (ior_intercept.size () > 0))
            {
                interceptor_manager = new InterceptorManager
                (
                    client_interceptors,
                    server_interceptors,
                    ior_intercept,
                    info.getSlotCount (),
                    this
                );
            }

            // add PolicyFactories to ORB
            policy_factories = info.getPolicyFactories();
        }
    }

    public void shutdown( boolean wait_for_completion )
    {
        if (logger.isInfoEnabled())
        {
            logger.info("prepare ORB for shutdown...");
        }

        if( ! run )
        {
            return; // ORB already shut down...
        }

        synchronized( shutdown_synch )
        {
            if (logger.isInfoEnabled())
            {
                logger.info("ORB going down...");
            }

            if( Environment.useImR() && imr != null )
            {
                try
                {
                    if (getImRServerImpl() != null)
                    {
                        // Stop any pinging.
                        getImRServerImpl().heartbeat(0);
                        // Deregister all POAs for this server.
                        imr.deregisterAll ( getImRServerImpl().id() );
                    }
                }
                catch (NotFound e )
                {
                    Debug.output
                        ( "Internal Error - while shutting down caught " + e );
                }
            }

            if( shutdown_in_progress && wait_for_completion )
            {
                synchronized( shutdown_synch )
                {
                    try
                    {
                        shutdown_synch.wait();
                    }
                    catch( InterruptedException ie )
                    {}

                    if (logger.isDebugEnabled())
                    {
                        logger.debug("ORB going shutdown complete (1)");
                    }

                    return;
                }
            }
            else if( shutdown_in_progress && !wait_for_completion )
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("ORB going shutdown complete (2)");
                }
                return;
            }

            shutdown_in_progress = true;
        }

        if( rootpoa != null )
            rootpoa.destroy(true, wait_for_completion);

        if( basicAdapter != null )
        {
            basicAdapter.stopListeners();
        }

        if (logger.isDebugEnabled())
        {
            logger.debug("ORB going shutdown (cleaning up ORB...)");
        }

        clientConnectionManager.shutdown();
        knownReferences.clear();
        bufferManager.release();

        /* notify all threads waiting for shutdown to complete */

        synchronized( shutdown_synch )
        {
            shutdown_synch.notifyAll();
        }

        /* notify all threads waiting in orb.run() */
        synchronized( orb_synch )
        {
            run = false;
            orb_synch.notifyAll();
        }
        if (logger.isInfoEnabled())
        {
            logger.info("ORB shutdown complete");
        }
    }

    public void destroy()
    {
        if( destroyed )
            throw new org.omg.CORBA.OBJECT_NOT_EXIST();

        if( run )
        {
            shutdown( true );
        }

        if( interceptor_manager != null )
        {
            interceptor_manager.destroy();
        }

        // other clean up possible here ?
        destroyed = true;
    }

    public org.omg.CORBA.Object string_to_object (String str)
    {
        if (str == null)
        {
            return null;
        }

        try
        {
            ParsedIOR pior = new ParsedIOR( str, this );

            if( pior.isNull() )
            {
                return null;
            }
            else
            {
                return _getObject(pior);
            }
        }
        catch (IllegalArgumentException iae)
        {
            if (logger.isDebugEnabled())
            {
                logger.debug(iae.getMessage());
            }
            return null;
        }
    }

    /**
     * always return a ValueDef or throw BAD_PARAM if not repid of a value
     */

    public org.omg.CORBA.Object get_value_def(String repid)
        throws org.omg.CORBA.BAD_PARAM
    {
        throw new org.omg.CORBA.NO_IMPLEMENT();
    }

    /**
     * called by org.jacorb.poa.RequestProcessor
     */

    public void set_delegate( java.lang.Object wrapper )
    {
        if( ! (wrapper instanceof org.omg.PortableServer.Servant) )
            throw new org.omg.CORBA.BAD_PARAM("Argument must be of type org.omg.PortableServer.Servant");
        else
        {
            try
            {
                ((org.omg.PortableServer.Servant)wrapper)._get_delegate();
            }
            catch( org.omg.CORBA.BAD_INV_ORDER bio )
            {
                // only set the delegate if it has not been set already
                org.jacorb.orb.ServantDelegate delegate =
                    new org.jacorb.orb.ServantDelegate( this );
                ((org.omg.PortableServer.Servant)wrapper)._set_delegate(delegate);
            }
        }
    }

    /**
     * forces the use of a proxy (by using a dummy applet)
     */

    public void useProxy()
    {
        applet = new java.applet.Applet();
    }

    public String object_to_string( org.omg.CORBA.Object obj)
    {
        if (obj == null)
        {
            return nullIORString;
        }

        if (obj instanceof org.omg.CORBA.LocalObject)
        {
           throw new MARSHAL ("Attempt to stringify a local object");
        }

        Object delegate =
            ((org.omg.CORBA.portable.ObjectImpl)obj)._get_delegate();
        if (delegate instanceof org.jacorb.orb.Delegate)
            return delegate.toString();
        else
            throw new BAD_PARAM("Argument has a delegate whose class is "
                                + delegate.getClass().getName()
                                + ", a org.jacorb.orb.Delegate was expected");
    }

    public void perform_work ()
    {
        if (! run)
        {
            throw new org.omg.CORBA.BAD_INV_ORDER
                (4, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    public boolean work_pending ()
    {
        if (! run)
        {
            throw new org.omg.CORBA.BAD_INV_ORDER
                (4, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }

        return false;
    }

    public ValueFactory register_value_factory (String id,
                                                ValueFactory factory)
    {
        return (ValueFactory)valueFactories.put (id, factory);
    }

    public void unregister_value_factory (String id)
    {
        valueFactories.remove (id);
    }

    public ValueFactory lookup_value_factory (String id)
    {
        ValueFactory result = (ValueFactory)valueFactories.get (id);

        if (result == null)
        {
            if (id.startsWith ("IDL"))
            {
                String valueName = org.jacorb.ir.RepositoryID.className (id);
                result = findValueFactory (valueName);
                valueFactories.put (id, result);
            }
        }
        return result;
    }

    /**
     * Finds a ValueFactory for class valueName by trying standard class names.
     */

    private ValueFactory findValueFactory (String valueName)
    {
        Class result = null;
        result = findClass (valueName + "DefaultFactory", true);
        if (result != null)
        {
            return (ValueFactory)instantiate (result);
        }
        else
        {
            // Extension of the standard: Handle the common case
            // when the Impl class is its own factory...
            Class c = findClass (valueName, false);
            result  = findClass (valueName + "Impl", false);

            if (result != null && c.isAssignableFrom (result))
            {
                if (ValueFactory.class.isAssignableFrom (result))
                {
                    return (ValueFactory)instantiate (result);
                }
                else
                {
                    // ... or create a factory on the fly
                    return new JacORBValueFactory (result);
                }
            }
            else
                return null;
        }
    }

    /**
     * Internal value factory class.  This can be used for any value
     * implementation that has a no-arg constructor.
     */
    private class JacORBValueFactory
        implements org.omg.CORBA.portable.ValueFactory
    {
        private Class implementationClass;

        public JacORBValueFactory (Class c)
        {
            implementationClass = c;
        }

        public java.io.Serializable read_value
          (org.omg.CORBA_2_3.portable.InputStream is)
        {
            java.lang.Object implObj = instantiate (implementationClass);

            if (implObj instanceof org.omg.CORBA.portable.Streamable)
            {
                StreamableValue value =
                (StreamableValue)instantiate (implementationClass);

                return is.read_value(value);
            }
            else if (implObj instanceof org.omg.CORBA.portable.CustomValue )
            {
                ((org.omg.CORBA.portable.CustomValue)implObj).unmarshal(
                    new DataInputStream(is));

                return ((org.omg.CORBA.portable.CustomValue)implObj);
            }
            else
            {
                throw new MARSHAL ("Unknown Value type " + implObj);
            }
        }
    }

    /**
     * Returns the class object for `name', if it exists, otherwise
     * returns null.  If `orgomg' is true, and `name' starts with "org.omg",
     * do a double-take using "omg.org" as the prefix.
     */
    private Class findClass (String name, boolean orgomg)
    {
        Class result = null;
        try
        {
            result = Environment.classForName(name);
        }
        catch (ClassNotFoundException e)
        {
            if (orgomg && name.startsWith ("org.omg"))
                try
                {
                    result = Environment.classForName("omg.org" + name.substring(7));
                }
                catch (ClassNotFoundException x)
                {
                    // nothing, result is null
                }
        }
        return result;
    }

    /**
     * Instantiates class `c' using its no-arg constructor.  Throws a
     * run-time exception if that fails.
     */
    private Object instantiate (Class c)
    {
        try
        {
            return c.newInstance();
        }
        catch (IllegalAccessException e1)
        {
            throw new RuntimeException ("cannot instantiate class "
                                        + c.getName()
                                        + " (IllegalAccessException)");
        }
        catch (InstantiationException e2)
        {
            throw new RuntimeException ("cannot instantiate class "
                                        + c.getName()
                                        + " (InstantiationException)");
        }
    }


    /**
     * Test, if the ORB has ClientRequestInterceptors <br>
     * Called by Delegate.
     */

    public boolean hasClientRequestInterceptors ()
    {
        return hasClientInterceptors;
    }

    /**
     * Test, if the ORB has ServerRequestInterceptors <br>
     * Called by poa.RequestProcessor.
     */

    public boolean hasServerRequestInterceptors ()
    {
        return hasServerInterceptors;
    }

    /**
     * Test, if the ORB has client or server side interceptors.
     */

    public boolean hasRequestInterceptors ()
    {
        return (hasServerInterceptors || hasClientInterceptors);
    }

    /**
     * Get the InterceptorManager, if present.
     *
     * @return the InterceptorManager, or null, if none is present.
     */

    public org.jacorb.orb.portableInterceptor.InterceptorManager getInterceptorManager()
    {
        return interceptor_manager;
    }

    public TransportManager getTransportManager()
    {
        if (transport_manager == null)
        {
            transport_manager = new TransportManager (this);
        }
        return transport_manager;
    }

    /* DII helper methods */

    public org.omg.CORBA.ExceptionList create_exception_list ()
    {
        return new org.jacorb.orb.dii.ExceptionList ();
    }

    public org.omg.CORBA.NVList create_list (int count)
    {
        return new org.jacorb.orb.NVList (this, count);
    }

    public org.omg.CORBA.NamedValue create_named_value
        (String name, org.omg.CORBA.Any value, int flags)
    {
        return new org.jacorb.orb.NamedValue (name, value, flags);
    }

    public org.omg.CORBA.NVList create_operation_list (org.omg.CORBA.Object obj)
    {
        org.omg.CORBA.OperationDef oper;

        if (obj instanceof org.omg.CORBA.OperationDef)
        {
            oper = (org.omg.CORBA.OperationDef) obj;
        }
        else
        {
            throw new org.omg.CORBA.BAD_PARAM ("Argument must be of type org.omg.CORBA.OperationDef");
        }
        return (create_operation_list (oper));
    }

    // This operation is under deprecation. To be replaced by one above.

    public org.omg.CORBA.NVList create_operation_list
        (org.omg.CORBA.OperationDef oper)
    {
        int no = 0;
        org.omg.CORBA.Any any;
        org.omg.CORBA.ParameterDescription[] params = null;
        org.omg.CORBA.ParameterDescription param;
        org.omg.CORBA.NVList list;

        params = oper.params ();
        if (params != null)
        {
            no = params.length;
        }
        list = new org.jacorb.orb.NVList (this, no);

        for (int i = 0; i < no; i++)
        {
            param = params[i];
            any = create_any ();
            any.type (param.type);
            switch (param.mode.value())
            {
                case org.omg.CORBA.ParameterMode._PARAM_IN:
                {
                    list.add_value (param.name, any, org.omg.CORBA.ARG_IN.value);
                    break;
                }
                case org.omg.CORBA.ParameterMode._PARAM_OUT:
                {
                    list.add_value (param.name, any, org.omg.CORBA.ARG_OUT.value);
                    break;
                }
                case org.omg.CORBA.ParameterMode._PARAM_INOUT:
                {
                    list.add_value (param.name, any, org.omg.CORBA.ARG_INOUT.value);
                    break;
                }
            }
        }

        return list;
    }


    /**
     * <code>addObjectKey </code> is a proprietary method that allows the internal
     * objectKeyMap to be altered programmatically. The objectKeyMap allows more
     * readable corbaloc URLs by mapping the actual object key to an arbitary string.
     * See the jacorb.properties file for more information.
     *
     * @param key_name a <code>String</code> value e.g. NameService
     * @param full_path an <code>String</code> value e.g. file:/home/rnc/NameSingleton.ior
     */
    public void addObjectKey(String key_name, String full_path)
    {
        objectKeyMap.put(key_name, full_path);
    }


    /**
     * Map an object key to another, as defined by the value
     * of a corresponding configuration property in the properties
     * file, e.g. map "NameService" to "StandardNS/NameServer-POA/_root"
     *
     * @param originalKey a <code>byte[]</code> value containing the original
     * key.
     * @return a <code>byte[]</code> value containing the mapped key, if a
     * mapping is defined, originalKey otherwise.
     */
    public byte[] mapObjectKey( byte[] originalKey )
    {
        BufferedReader br        = null;
        File           iorFile   = null;
        ParsedIOR      pIOR      = null;
        String         found     = null;
        String         original  = null;

        if( objectKeyMap.size() != 0 )
        {
            original = new String( originalKey );
            found    = (String)objectKeyMap.get( original );

            if( found != null )
            {
                if ( ParsedIOR.isParsableProtocol ( found ) )
                {
                    // We have found a file reference. Use ParsedIOR to get
                    // the byte key.
                    try
                    {
                        pIOR = new ParsedIOR( found, this );
                        return pIOR.get_object_key();
                    }
                    catch ( IllegalArgumentException e )
                    {
                        if (logger.isDebugEnabled())
                        {
                            logger.debug("Error - could not read protocol " + found );
                        }
                        return originalKey;
                    }
                }
                else
                {
                    return org.jacorb.orb.util.CorbaLoc.parseKey(found);
                }
            }
        }
        // else:
        return originalKey;
    }

    /**
     * Inner class that implements org.omg.PortableInterceptor.Current
     * by forwarding each invocation to a thread-dependent target.
     */
    private class PICurrent
        extends org.omg.CORBA.LocalObject
        implements org.omg.PortableInterceptor.Current
    {
        // Helper method that returns the actual
        // target of a PICurrent invocation
        private Current getTarget()
        {
            if (interceptor_manager == null)
                return InterceptorManager.EMPTY_CURRENT;
            else
                return interceptor_manager.getCurrent();
        }

        // org.omg.PortableInterceptor.Current implementation ---

        public org.omg.CORBA.Any get_slot(int id)
            throws InvalidSlot
        {
            return getTarget().get_slot(id);
        }

        public void set_slot(int id, org.omg.CORBA.Any data)
            throws InvalidSlot
        {
            getTarget().set_slot(id, data);
        }

    }

}
