package org.jacorb.orb.giop;

/*
 *        JacORB - a free Java ORB
 *
 *   Copyright (C) 1997-2003  Gerald Brose.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.io.UnsupportedEncodingException;
import org.jacorb.util.Debug;
import org.omg.CORBA.DATA_CONVERSION;
import org.omg.CORBA.INTERNAL;
import sun.io.ByteToCharConverter;
import sun.io.ConversionBufferFullException;
import sun.io.MalformedInputException;
import sun.io.UnknownCharacterException;


/**
 * <code>BTCConverter</code> is a class to convert byte arrays to char arrays
 * according to a chosen encoding. This version uses the sun.io classes to do
 * the conversion.
 *
 * @author <a href="mailto:rnc@prismtechnologies.com">Nick Cross</a>
 * @version 1.0
 */
public class BTCConverter
{
    /**
     * <code>converter</code> is the actual Sun converter.
     */
    private ByteToCharConverter converter;


    /**
     * <code>BTCConverter</code> creates a new instance with a specific encoding.
     *
     * @param encoding a <code>String</code> value
     */
    public BTCConverter(String encoding)
    {
        try
        {
            converter = ByteToCharConverter.getConverter(encoding);

            // When substitution mode is false, the converter
            // will throw exceptions when it encounters illegal
            // encodings.
            converter.setSubstitutionMode(false);
        }
        catch (UnsupportedEncodingException ex)
        {
            // Unknown encoding
            if (Debug.isDebugEnabled())
            {
                Debug.output(4, ex);
            }
            throw new DATA_CONVERSION(ex.getMessage());
        }
    }



    /**
     * <code>convert</code> returns the converted char array.
     *
     * @param source a <code>byte[]</code> value; the byte array
     * @param offset an <code>int</code> value; the offset to start at
     * @param length an <code>int</code> value; the number of bytes
     * @return a <code>byte[]</code> value
     */
    public char[] convert (byte[] value, int offset, int length)
    {
        char []result = null;

        try
        {
            int resultingNumChars = 0;

            result = new char[length * converter.getMaxCharsPerByte()];

            converter.reset();

            // WARNING:  The signature for convert is
            // value[], offset, endPosition
            // not
            // value[], offset, total length
            resultingNumChars = converter.convert(value, offset, offset + length,
                                            result, 0, result.length);

            resultingNumChars += converter.flush(result, 0, result.length);

            if (resultingNumChars < result.length)
            {
                // A short format was used:  Trim the byte array.
                char[] trimResult = new char[resultingNumChars];
                System.arraycopy(result, 0, trimResult, 0, resultingNumChars);
                result = trimResult;
            }
        }
        catch (MalformedInputException ex)
        {
            // There were illegal Unicode char pairs
            if (Debug.isDebugEnabled())
            {
                Debug.output(4, ex);
            }
            throw new DATA_CONVERSION(ex.getMessage());
        }
        catch (UnknownCharacterException ex)
        {
            // A character doesn't map to the desired code set
            // CORBA formal 00-11-03.
            if (Debug.isDebugEnabled())
            {
                Debug.output(4, ex);
            }
            throw new DATA_CONVERSION(ex.getMessage());
        }
        catch (ConversionBufferFullException ex)
        {
            // If this happens, then the CharToByteConverter was lying
            // about the maximum bytes per char.
            if (Debug.isDebugEnabled())
            {
                Debug.output(4, ex);
            }
            throw new INTERNAL(ex.getMessage());
        }
        return result;
    }
}
