package org.jacorb.orb.giop;

/*
 *        JacORB - a free Java ORB
 *
 *   Copyright (C) 1997-2003  Gerald Brose.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// Imports
import java.util.ArrayList;
import java.util.List;
import org.jacorb.orb.iiop.IIOPConnection;
import org.jacorb.util.Debug;
import org.jacorb.util.Environment;

/**
 * @author Nicolas Noffke
 * @version $Id: ServerGIOPConnection.java,v 1.4 2004/05/11 14:26:13 orbtest4 Exp $
 */

public class ServerGIOPConnection
    extends GIOPConnection
{
    private static final byte[] CLOSE_CONNECTION_MESSAGE =
    new byte[] {
        // Byte casts are for JDK1.2 compatibility.
        (byte )'G', (byte )'I', (byte )'O', (byte )'P', //magic start
        1, //GIOP major
        0, //GIOP minor
        0, //endianess, big-endian
        5, //message type, CloseConnection message
        0, 0, 0, 0 // message size, 0 because CloseConnection has no body
    };


    private GIOPConnectionManager manager = null;

    private boolean closeOnReadTimeout = false;

    private boolean delayClose = false;

    /**
     * <code>indices</code> is used to record which slots we are using in
     * the GIOPConnectionManager service context cache in order that we
     * can clear out the svcCache which this connection is closed.
     */
    private ArrayList indices;


    public ServerGIOPConnection( org.omg.ETF.Profile profile,
                                 org.omg.ETF.Connection transport,
                                 RequestListener request_listener,
                                 ReplyListener reply_listener,
                                 StatisticsProvider statistics_provider,
                                 GIOPConnectionManager manager )
    {
        super( profile, transport, request_listener, reply_listener, statistics_provider );

        this.manager = manager;

        delayClose =
            Environment.isPropertyOn( "jacorb.connection.delay_close" );
    }


    /*
     * <code>tryClose</close> called by GIOPConnectionManager::createServerGIOPConnection
     * if there are too many GIOP connections or if we have timed out with nothing pending.
     */
    boolean tryClose()
    {
        if( tryDiscard() )
        {
            sendCloseConnection();

            closeOnReadTimeout = true;

            if( connection_listener != null )
            {
                connection_listener.connectionClosed();
            }

            return true;
        }
        else
        {
            return false;
        }
    }


    /**
     * Atomically try to set this connection into discarding mode, if
     * it doesn't have any pending messages.
     *
     * @return true, if the connection has been idle and discarding
     * has been set
     */
    private boolean tryDiscard()
    {
        if( ! hasPendingMessages() )
        {
            synchronized( pendingUndecidedSync )
            {
                discard_messages = true;
            }

            return true;
        }
        else
        {
            return false;
        }
    }


    /**
     * <code>sendCloseConnection</code> sends a close connection message
     * flushing the transport.
     */
    private void sendCloseConnection()
    {
        try
        {
            getWriteLock();

            write( CLOSE_CONNECTION_MESSAGE,
                   0,
                   CLOSE_CONNECTION_MESSAGE.length );

            transport.flush();

            if (statistics_provider != null)
            {
                statistics_provider.flushed();
            }

            if( delayClose && transport instanceof IIOPConnection )
            {
                ((IIOPConnection)transport).turnOnFinalTimeout();
            }
            else
            {
                // Set do_close to true so anything waiting in waitUntilConnection
                // doesn't think there is a possibility of connecting. I think.
                do_close = true;

                transport.close();
            }
        }
        catch( org.omg.CORBA.COMM_FAILURE e )
        {
            Debug.output( 1, e );
        }
        finally
        {
            releaseWriteLock();
        }

        if( manager != null )
        {
            manager.unregisterServerGIOPConnection( this );
        }
    }


    protected void readTimedOut()
    {
        if( closeOnReadTimeout )
        {
            close();
        }
        else
        {
            /**
             * If we don't have any more pending messages, we'll send a
             * GIOP CloseConnection message (done by tryClose() ).
             */
            tryClose();
        }
    }

    /**
     * We're server side and can't reopen, therefore close completely
     * if stream closed.
     */
    protected void streamClosed()
    {
        close();
    }

    /**
     * @overrides close in GIOPConnection
     */
    public void close()
    {
        super.close();

        if (indices != null)
        {
            for (int i=0; i<indices.size(); i++)
            {
                manager.removeContextKey(((Integer)indices.get(i)).intValue());
            }
        }
        if( manager != null )
        {
            manager.unregisterServerGIOPConnection( this );
        }
    }


    /**
     * <code>addContextKey</code> caches the key, oid and scopes from ServerRequest
     * in order to speed communication between JacORB implementations.
     *
     * @param object_key a <code>byte[]</code> value
     * @param oid a <code>byte[]</code> value
     * @param scopes a <code>List</code> value
     * @return a <code>byte[]</code> value
     */
    public byte[] addContextKey(byte []object_key, byte []oid, List scopes)
    {
        byte []result = manager.addContextKey(object_key, oid, scopes);

        if (indices == null)
        {
            indices = new ArrayList(10);
        }

        indices.add (new Integer(manager.getSvcID()));

        return result;
    }


    /**
     * <code>getContextKey</code> will return a cached ContextKey value
     * for that mini-key if one exists.
     *
     * @param key a <code>byte[]</code> value
     * @return a <code>ContextKey</code> value
     */
    public GIOPConnectionManager.ContextKey getContextKey(byte []key)
    {
        return manager.getContextKey(key);
    }
}// ServerGIOPConnection
